
//
//  Bench.c - Test ram module alignment 
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 2 Jun 98 - James Leiterman
//
//  Last Modification
//
//      23 Jul 2001 - JCL - Brought up to date.
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries

#include <stdio.h>
#include <string.h>

    //  Book Development Library Modules

#include "CpuAsm.h"
#include "ram.h"                                        // Memory Module



#define RAM_TBL_SIZE        4096



int main( int argc, char *argv[] )
{
    unsigned int n, nCnt, bSet, bTst, bMsk, a, b;
    byte *pTbl[ RAM_TBL_SIZE ];
    CpuInfo cinfo;
//  char szBuf[ CPU_SZBUF_MAX ];

//      Normally we would call cpu detection but that won't be covered until Chapter 3!
//    CpuDetect( &cinfo );
//    printf( "\n%s\n", CpuInfoStr( szBuf, &cinfo ) );
    
    ramSetup( &cinfo );
          
        //  Allocate a series of increasingly different size blocks of memory

    for ( n = 0; n < RAM_TBL_SIZE; n++ )
      {
//      if ( NULL == ( pTbl[n] = ramAlloc( n+1 )) )
//      if ( !ramGet( &pTbl[n], n+1 ) )
        if ( !ramGetClr( &pTbl[n], n+1 ) )
          {
            printf( "low memory.  (continuing...)\n" );
            break;
          }
      }

    nCnt = n;                           // # of entries allocated
    
        //  Test memory for alignments

    bSet = 16;                          // Preset to 128bit (128/8)
    bTst = bSet - 1;
    bMsk = ~0 ^ bTst; 

    for ( n = 0; n < nCnt; n++ )
      {
        a = (unsigned int) pTbl[n];

        do {
            b = ( a + bTst ) & bMsk;    // round up to 'bSet' bits

            if ( a == b )
              {
                break;                  // okay
              }
                                        // Unaligned...
            bSet >>= 1;                 // reduce by a bit
            bTst = bSet - 1;
            bMsk = ~0 ^ bTst; 
          } while(1);
      }

        //  Release all of memory to cleanup

    for ( n = 0; n < nCnt; n++ )
      {
//        ramFree( pTbl[n] );
//        pTbl[n] = NULL;
        ramRelease( &pTbl[n] );
      }

    printf( "Ram is now aligned to %u bytes (%u bits).\n\n",
        bSet, (bSet<<3)  );
    printf( "Fin!\n\n" );

    return 0;
}

