
//
//  tally.c
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 14 Jul 98 - James Leiterman
//
//  Last Modification
//
//      23 Jul 2001 - JCL - Brought up to date.
//
//  stab=4
//


#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries

#include <stdio.h>
#include <assert.h>




uint Ary[] = {
   826, 998, 968, 276,  353, 127, 927, 324,
   467, 754, 345, 942,  667,  36, 299, 894,
   581, 373, 579, 325 
};

uint nAryCnt = sizeof(Ary) / sizeof(*Ary);    // =20



//
// Vector - 32bit Unsigned Word
//

typedef struct VectorUW
{
    uint x;
    uint y;
    uint z;
    uint a;
} VectorUW;



//
//  Add Unsigned 32bit Word Vector
//

void VecAddUW( VectorUW * const vDst, const VectorUW * const vSrcA, const VectorUW * const vSrcB )
{
    ASSERT_PTR4(vDst);
    ASSERT_PTR4(vSrcA);
    ASSERT_PTR4(vSrcB);

    vDst->x = vSrcA->x + vSrcB->x;
    vDst->y = vSrcA->y + vSrcB->y;
    vDst->z = vSrcA->z + vSrcB->z;
    vDst->a = vSrcA->a + vSrcB->a;
}



//
//  Ary Tally (Vector)
//
uint AryTallyV( const uint *pAry, uint nCnt )
{
    uint nVecCnt;
    VectorUW Tally, *vAry;

    ASSERT_PTR4(pAry);

    vAry = (VectorUW *)pAry;
    nVecCnt = nCnt >> 2;                    // # of 4 word vectors

    ASSERT( 2 <= nVecCnt );                 // only allow {8,12,16,20,etc.}

    VecAddUW( &Tally, vAry, vAry+1 );       // 1st+2nd 128bits
    vAry += 2;
    nVecCnt -= 2;

    if ( nVecCnt )
      {
        do {
            VecAddUW( &Tally, &Tally, vAry++ ); // next 128bits
          } while( --nVecCnt );
      }

    return Tally.x + Tally.y + Tally.z + Tally.a;
}



//
//  Ary Tally (Quad)
//  (Handle 4 32bit words at a time)

uint AryTallyQ( const uint *pAry, uint nCnt )
{
    uint nVecCnt, Tally[4];

    ASSERT_PTR4(pAry);

    nVecCnt = nCnt >> 2;                    // # of 4 word vectors

    ASSERT(2 <= nVecCnt);                   // only allow {8,12,16,20,etc.}

    Tally[0] = *(pAry+0) + *(pAry+4);       // 1st+2nd 128bits
    Tally[1] = *(pAry+1) + *(pAry+5);
    Tally[2] = *(pAry+2) + *(pAry+6);
    Tally[3] = *(pAry+3) + *(pAry+7);
    pAry += 8;
    nVecCnt -= 2;

    if ( nVecCnt )
      {
        do {
            Tally[0] += *(pAry+0);          // next 128bits
            Tally[1] += *(pAry+1);
            Tally[2] += *(pAry+2);
            Tally[3] += *(pAry+3);
            pAry += 4;
          } while( --nVecCnt );
      }

    return Tally[0] + Tally[1] + Tally[2] + Tally[3];
}



//
//  Ary Tally
//

uint AryTally( const uint *pAry, uint nCnt )
{
    uint nTally;

    ASSERT_PTR4(pAry);

    nTally = 0;

    if ( nCnt )
      {
        do {
            nTally += *pAry++;
          } while( --nCnt );
      }

    return nTally;
}



int main( int argc, char *argv[] )
{
    uint nTally, nTallyQ, nTallyV;

    printf( "Begin Tally Count\n" );

    printf( "Sum as elements\n" );
    nTally = AryTally( Ary, nAryCnt );      // Sum as elements

    printf( "Sum as quads\n" );;
    nTallyQ = AryTallyQ( Ary, nAryCnt );    // Sum as Quads

    printf( "Sum as vectors\n" );
    nTallyV = AryTallyV( Ary, nAryCnt );    // Sum as Vectors
 
    printf( "Control tally: %u vs Quad: %u vs Vec: %u\n\n",
        nTally, nTallyQ, nTallyV );

    printf( "Fin!\n\n" );

    return 0;
}

