
//
//  PBoolMMI.s - Packed (128bit Mips) Boolean Logic
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//      This module is designed for fast 3D math
//
//  Created - 26 Jan, 2002 - James Leiterman
//
//  Modifications
//
//
//  stab = 4
//

    .nolist
#include "Mips.inc"
#include "KariType.inc"
    .list

    .set    noreorder
    .set    nomips16



    .section .text                                                    



//_________________________________________________________________________________________________
//=================================================================================================
//  Logical NOT  D=(!A)
//
//  void vmp_pnot(  void * const pvD, const void * const pvA )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_pnotMMI
vmp_pnotMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    li      t2,-1
    lq      t1, 0(a1)       // pvA

    pcpyld  t2,t2,t2        // {-1 -1} <<- {0 01}
    pxor    t0, t1, t2

    sq      t0, 0(a0)       // pvD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_pnotMMI0
vmp_pnotMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    li      t2,-1
    ldl     t1, 7(a1)       // pvA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

#if 01                      // 64bit method
    xor     t0,t1,t2
    xor     t1,t3,t2
#else                       // 128bit method
    pcpyld  t2,t2,t2        // {-1 -1} <<- {0 01}
    pcpyld  t1,t3,t1        // A  {127...0}

    pxor    t0, t1, t2      // D  {127...0}

    pcpyud  t1,t0,t0        // D  {63...0  127...64}
#endif

    sdl     t0, 7(a0)       // pvD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


//_________________________________________________________________________________________________
//=================================================================================================
//  Logical AND  D=(A & B)
//
//  void vmp_pand(  void * const pvD, const void * const pvA, const void * const pvB )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_pandMMI
vmp_pandMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // pvA
    lq      t2, 0(a2)       // pvB
    LDELAY                  // nop - Load Delay Slot

    pand    t0, t1, t2

    sq      t0, 0(a0)       // pvD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_pandMMI0
vmp_pandMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // pvA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

    ldl     t2, 7(a2)       // pvB  {63...0}
    ldr     t2, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

#if 01                      // 64bit method
    and     t0,t1,t2
    and     t1,t3,t4
#else                       // 128bit method
    pcpyld  t1,t3,t1        // A  {127...0}
    pcpyld  t2,t4,t2        // B  {127...0}

    pand    t0, t1, t2      // D  {127...0}

    pcpyud  t1,t0,t0        // D  {63...0  127...64}
#endif

    sdl     t0, 7(a0)       // pvD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  Logical OR  D=(A | B)
//
//  void vmp_por(   void * const pvD, const void * const pvA, const void * const pvB )
//
//  *** MMI version *** 128bit aligned

    .align  3
    .global vmp_porMMI
vmp_porMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // pvA
    lq      t2, 0(a2)       // pvB
    LDELAY                  // mop - Load Delay Slot

    por     t0, t1, t2

    sq      t0, 0(a0)       // pvD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_porMMI0
vmp_porMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // pvA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

    ldl     t2, 7(a2)       // pvB  {63...0}
    ldr     t2, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

#if 01                      // 64bit method
    or      t0,t1,t2
    or      t1,t3,t4
#else                       // 128bit method
    pcpyld  t1,t3,t1        // A  {127...0}
    pcpyld  t2,t4,t2        // B  {127...0}

    por     t0, t1, t2      // D  {127...0}

    pcpyud  t1,t0,t0        // D  {63...0  127...64}
#endif

    sdl     t0, 7(a0)       // pvD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  Logical XOR (Exclusive OR)  D=(A^B)
//
//  void vmp_pxor(  void * const pvD, const void * const pvA, const void * const pvB )
//
//  *** MMI version *** 128bit aligned

    .align  3
    .global vmp_pxorMMI
vmp_pxorMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // pvA
    lq      t2, 0(a2)       // pvB
    LDELAY                  // mop - Load Delay Slot

    pxor    t0, t1, t2

    sq      t0, 0(a0)       // pvD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_pxorMMI0
vmp_pxorMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // pvA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

    ldl     t2, 7(a2)       // pvB  {63...0}
    ldr     t2, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

#if 01                      // 64bit method
    xor     t0,t1,t2
    xor     t1,t3,t4
#else                       // 128bit method
    pcpyld  t1,t3,t1        // A  {127...0}
    pcpyld  t2,t4,t2        // B  {127...0}

    pxor    t0, t1, t2      // D  {127...0}

    pcpyud  t1,t0,t0        // D  {63...0  127...64}
#endif

    sdl     t0, 7(a0)       // pvD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  Logical And w/1's Complement  D=(A & !B)
//
//  void vmp_pandc( void * const pvD, const void * const pvA, const void * const pvB )
//
//  *** MMI version *** 128bit aligned

    .align  3
    .global vmp_pandcMMI
vmp_pandcMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    li      t3,-1
    lq      t2, 0(a2)       // pvB
    lq      t1, 0(a1)       // pvA

    pcpyld  t3,t3,t3        // {-1 -1} <<- {0 01}
    pxor    t2,t2,t3        // Flip B
    pand    t0, t1, t2

    sq      t0, 0(a0)       // pvD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_pandcMMI0
vmp_pandcMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    li      t5,-1
    ldl     t1, 7(a1)       // pvA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

    ldl     t2, 7(a2)       // pvB  {63...0}
    ldr     t2, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

#if 01                      // 64bit method
    xor     t2,t2,t5
    xor     t4,t4,t5
    and     t0,t1,t2
    and     t1,t3,t4
#else                       // 128bit method
    pcpyld  t5,t5,t5        // {-1 -1} <<- {0 01}
    pcpyld  t1,t3,t1        // A  {127...0}
    pcpyld  t2,t4,t2        // B  {127...0}

    pxor    t2,t2,t5        // Flip B
    pand    t0, t1, t2      // D  {127...0}

    pcpyud  t1,t0,t0        // D  {63...0  127...64}
#endif

    sdl     t0, 7(a0)       // pvD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  Logical NOR  1's complement of OR  D=!(A | B)
//
//  void vmp_pnor( void * const pvD, const void * const pvA, const void * const pvB )
//
//  *** MMI version *** 128bit aligned

    .align  3
    .global vmp_pnorMMI
vmp_pnorMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // pvA
    lq      t2, 0(a2)       // pvB
    LDELAY                  // mop - Load Delay Slot

    pnor    t0, t1, t2

    sq      t0, 0(a0)       // pvD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_pnorMMI0
vmp_pnorMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // pvA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

    ldl     t2, 7(a2)       // pvB  {63...0}
    ldr     t2, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

#if 01                      // 64bit method
    nor t0,t1,t2
    nor t1,t3,t4
#else                       // 128bit method
    pcpyld  t1,t3,t1        // A  {127...0}
    pcpyld  t2,t4,t2        // B  {127...0}

    pnor    t0, t1, t2      // D  {127...0}

    pcpyud  t1,t0,t0        // D  {63...0  127...64}
#endif

    sdl     t0, 7(a0)       // pvD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



