
//
//  psraMMI.s - Packed (128bit Mips) Shift Right Arithmetic
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//      This module is designed for fast 3D math
//
//  Created - 26 Jan, 2002 - James Leiterman
//
//  Modifications
//
//
//  stab = 4
//

    .nolist
#include "Mips.inc"
#include "KariType.inc"
    .list

 

    .set    noreorder
    .set    nomips16

    .rdata


    .align  4

    .extern llMaskBO        //  Left Shift 32bit (unsigned) 8bit mask
    .extern rrMaskBO        //  Right Shift 32bit (unsigned) 8bit mask
    .extern llMaskHO        //  Left Shift 32bit (unsigned) 16bit mask
    .extern rrMaskHO        //  Right Shift 32bit (unsigned) 16bit mask



    .section .text                                                    



//_________________________________________________________________________________________________
//=================================================================================================
//  Packed Shift Right Arithmetic 4x32bit by 1
//
//  void vmp_psra1W( int32 * const pwD, const int32 * const pwA )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_psra1WMMI
vmp_psra1WMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // pwA
    LDELAY                  // nop - Load Delay Slot

    psraw   t0,t1,1         // D = A >> 1

    sq      t0, 0(a0)       // pwD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_psra1WMMI0
vmp_psra1WMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // pwA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)
    LDELAY                  // nop - Load Delay Slot

    pcpyld  t1,t3,t1        // A  {127...0}

    psraw   t0,t1,1         // D = A >> 1

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // pwD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  Packed Shift Right Arithmetic 4x32bit by n:{0...31}
//
//  void vmp_psraW( uint32 * const pwD, const uint32 * const pwA, uint count )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_psraWMMI
vmp_psraWMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // {3 2 1 0} pwA
    and     t2,a2,32-1      // 00011111b Clip count to data bit size (32-1)
    pcpyld  t2,t2,t2        // {0 count 0 count}

    psravw  t0,t1,t2        // {# 2 # 0}  D = A >> count

    pexcw   t1,t1           // {3 1 2 0}
    pexcw   t0,t0           // {# # 2 0}

    pextuw  t1,t1,t1        // {3 3 1 1}
    psravw  t1,t1,t2        // {# 3 # 1}  D = A >> count
    pexcw   t1,t1           // {# # 3 1}

    pextlw  t0,t1,t0        // {3 2 1 0}

    sq      t0, 0(a0)       // pwD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_psraWMMI0
vmp_psraWMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // pwA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)
    LDELAY                  // nop - Load Delay Slot

    pcpyld  t1,t3,t1        // {3 2 1 0} pwA

    and     t2,a2,32-1      // 00011111b Clip count to data bit size (32-1)
    pcpyld  t2,t2,t2        // {0 count 0 count}

    psravw  t0,t1,t2        // {# 2 # 0}  D = A >> count

    pexcw   t1,t1           // {3 1 2 0}
    pexcw   t0,t0           // {# # 2 0}

    pextuw  t1,t1,t1        // {3 3 1 1}
    psravw  t1,t1,t2        // {# 3 # 1}  D = A >> count
    pexcw   t1,t1           // {# # 3 1}

    pextlw  t0,t1,t0        // {3 2 1 0}

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // pwD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot




//_________________________________________________________________________________________________
//=================================================================================================
//  Packed Shift Right Arithmetic 8x16bit by 1
//
//  void vmp_psra1H( int16 * const phD, const int16 * const phA )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_psra1HMMI
vmp_psra1HMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // phA
    LDELAY                  // nop - Load Delay Slot

    psrah   t0,t1,1         // D = A >> 1

    sq      t0, 0(a0)       // phD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_psra1HMMI0
vmp_psra1HMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // phA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)
    LDELAY                  // nop - Load Delay Slot

    pcpyld  t1,t3,t1        // A  {127...0}

    psrah   t0,t1,1         // D = A >> 1

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // phD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  Packed Shift Right Arithmetic 8x16bit by n:{0...15}
//
//  void vmp_psraH( int16 * const phD, const int16 * const phA, uint count )
//
//  *** MMI version *** 128bit aligned
//
//  Handle 4x  2x16bit
    .align 3
    .global vmp_psraHMMI
vmp_psraHMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // {7...0} phA
    and     t2,a2,16-1      // 00001111b Clip count to data bit size (16-1)

    pcpyld  t2,t2,t2        // {0 count 0 count}

        //  [7 6 * 5 4 * 3 2 * 1 0]
    psllw   t4,t1,16        // {6_ 4_ 2_ 0_}
    psravw  t0,t1,t2        // {# 5   # 1}  A >> count
    psravw  t5,t4,t2        // {# 4   # 0}

    pexcw   t1,t1           // {7,6  3,2  5,4  1,0}
    pexcw   t0,t0           // {#  #  5  1}
    pexcw   t4,t4           // {6_ 2_ 4_ 0_}
    pexcw   t5,t5           // {#  #  4  0}

    pextuw  t1,t1,t1        // {7,6  7,6  3,2  3,2}
    pextuw  t4,t4,t4        // {6#   6#   2#   2#}
    psravw  t1,t1,t2        // {# 7  # 3}  A >> count
    psravw  t4,t4,t2        // {# 6  # 2}  A >> count
    pexcw   t1,t1           // {#  #  7  3}
    pexcw   t4,t4           // {#  #  6  2}
    pextlw  t0,t1,t0        // {7# 5# 3# 1#}
    pextlw  t5,t4,t5        // {6# 4# 2# 0#}

    psrlw   t0,t0,16        // {7 5 3 1}
    psrlw   t5,t5,16        // {6 4 2 0}
    pinteh  t0,t0,t5        // {7 6 5 4 3 2 1 0}

    sq      t0, 0(a0)       // phD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_psraHMMI0
vmp_psraHMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // phA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)
    LDELAY                  // nop - Load Delay Slot

    pcpyld  t1,t3,t1        // {3 2 1 0} pwA

        //  Aligned data section

    and     t2,a2,16-1      // 00001111b Clip count to data bit size (16-1)

    pcpyld  t2,t2,t2        // {0 count 0 count}

        //  [7 6 * 5 4 * 3 2 * 1 0]
    psllw   t4,t1,16        // {6_ 4_ 2_ 0_}
    psravw  t0,t1,t2        // {# 5   # 1}  A >> count
    psravw  t5,t4,t2        // {# 4   # 0}

    pexcw   t1,t1           // {7,6  3,2  5,4  1,0}
    pexcw   t0,t0           // {#  #  5  1}
    pexcw   t4,t4           // {6_ 2_ 4_ 0_}
    pexcw   t5,t5           // {#  #  4  0}

    pextuw  t1,t1,t1        // {7,6  7,6  3,2  3,2}
    pextuw  t4,t4,t4        // {6#   6#   2#   2#}
    psravw  t1,t1,t2        // {# 7  # 3}  A >> count
    psravw  t4,t4,t2        // {# 6  # 2}  A >> count
    pexcw   t1,t1           // {#  #  7  3}
    pexcw   t4,t4           // {#  #  6  2}
    pextlw  t0,t1,t0        // {7# 5# 3# 1#}
    pextlw  t5,t4,t5        // {6# 4# 2# 0#}

    psrlw   t0,t0,16        // {7 5 3 1}
    psrlw   t5,t5,16        // {6 4 2 0}
    pinteh  t0,t0,t5        // {7 6 5 4 3 2 1 0}

        //  Unaligned data section

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // phD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  Packed Shift Right Arithmetic 16x8bit by 1
//
//  void vmp_psra1B( int8 * const phD, const int8 * const phA )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_psra1BMMI
vmp_psra1BMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a0)

    la      t3,llMaskBO+(16*1)  // FE
    la      t2,llMaskBO+(16*7)  // 80 (Sticky Bit)

    lq      t1, 0(a1)       // phA
    lq      t3, 0(t3)       // get FE mask
    lq      t2, 0(t2)       // get 80 mask

    pand    t3,t3,t1
    pand    t2,t2,t1        // Save sign bits
    psrlh   t0,t3,1         // D = A >> 1
    por     t0,t0,t2        // sticky sign bit

    sq      t0, 0(a0)       // phD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_psra1BMMI0
vmp_psra1BMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a0)

    la      t4,llMaskBO+(16*1)  // FE
    la      t2,llMaskBO+(16*7)  // 80 (Sticky Bit)

    ldl     t1, 7(a1)       // phA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

    pcpyld  t1,t3,t1        // A  {127...0}

    lq      t4, 0(t4)       // get FE mask
    lq      t2, 0(t2)       // get 80 mask

    pand    t4,t4,t1
    pand    t2,t2,t1        // Save sign bits
    psrlh   t0,t4,1         // D = A >> 1
    por     t0,t0,t2        // sticky sign bit

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // phD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot




//_________________________________________________________________________________________________
//=================================================================================================
//  Packed Shift Right Arithmetic 16x8bit by n:{0...7}
//
//  void vmp_psraB( int8 * const phD, const int8 * const phA, uint count )
//
//  *** MMI version *** 128bit aligned
//
//  Handle 4x  4x8bit
    .align 3
    .global vmp_psraBMMI
vmp_psraBMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // {15...0} pbA
    and     t2,a2,8-1       // 00000111b Clip count to data bit size (8-1)

    pcpyld  t2,t2,t2        // {0 count 0 count}

        //t1=                   FEDC [BA98] 7654 [3210]

    psravw  t3,t1,t2        // {####  B###  ####  3###}

    psllw   t4,t1,8         // {EDC_  A98_  654_  210_}
    psravw  t4,t4,t2        // {####  A###  ####  2###}

    psllw   t5,t1,16        // {DC__  98__  54__  10__}
    psravw  t5,t5,t2        // {####  9###  ####  1###}

    psllw   t6,t1,24        // {C___  8___  4___  0___}
    psravw  t6,t6,t2        // {####  8###  ####  0###}

    pextub  t7,t3,t4        // {## ##  ## ##  BA ##  ## ##}
    pextub  t8,t5,t6        // {## ##  ## ##  98 ##  ## ##}
    pextlb  t3,t3,t4        // {## ##  ## ##  32 ##  ## ##}
    pextlb  t5,t5,t6        // {## ##  ## ##  10 ##  ## ##}

    pextlh  t7,t7,t8        // {BA 98  ## ##  ## ##  ## ##}
    pextlh  t3,t3,t5        // {32 10  ## ##  ## ##  ## ##]

    pcpyud  t9,t3,t7        // {BA98####   3210####}

        // Second Bank     t1= [FEDC] BA98 [7654] 3210

    pexcw   t1,t1           // {FEDC  7654  BA98  3210}
    pextuw  t1,t1,t1        // {FEDC  FEDC  7654  7654}
    psravw  t3,t1,t2        // {####  F###  ####  7###}

    psllw   t4,t1,8         // {EDC_  EDC_  654_  654_}
    psravw  t4,t4,t2        // {####  E###  ####  6###}

    psllw   t5,t1,16        // {DC__  DC__  54__  54__}
    psravw  t5,t5,t2        // {####  D###  ####  5###}

    psllw   t6,t1,24        // {C___  C___  4___  4___}
    psravw  t6,t6,t2        // {####  C###  ####  4###}

    pextub  t7,t3,t4        // {#,# #,#   #,# #,#  F,E #,#  #,# #,#}
    pextub  t8,t5,t6        // {#,# #,#   #,# #,#  D,C #,#  #,# #,#}
    pextlb  t3,t3,t4        // {#,# #,#   #,# #,#  7,6 #,#  #,# #,#}
    pextlb  t5,t5,t6        // {#,# #,#   #,# #,#  5,4 #,#  #,# #,#}

    pextlh  t7,t7,t8        // {FE DC  ## ##   ## ##  ## ##}
    pextlh  t3,t3,t5        // {76 54  ## ##   ## ##  ## ##]

    pcpyud  t8,t3,t7        // {FEDC####   7654####}

    pexcw   t9,t9           // {BA98  3210  ####  ####}
    pexcw   t8,t8           // {FEDC  7654  ####  ####}

    pextuw  t0,t8,t9        // {FEDC  BA98  7654  3210}

    sq      t0, 0(a0)       // pbD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_psraBMMI0
vmp_psraBMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // pbA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)
    LDELAY                  // nop - Load Delay Slot

    pcpyld  t1,t3,t1        // {3 2 1 0} pwA

        //  Aligned data section

    and     t2,a2,8-1       // 00000111b Clip count to data bit size (8-1)

    pcpyld  t2,t2,t2        // {0 count 0 count}

        //t1=                   FEDC [BA98] 7654 [3210]

    psravw  t3,t1,t2        // {####  B###  ####  3###}

    psllw   t4,t1,8         // {EDC_  A98_  654_  210_}
    psravw  t4,t4,t2        // {####  A###  ####  2###}

    psllw   t5,t1,16        // {DC__  98__  54__  10__}
    psravw  t5,t5,t2        // {####  9###  ####  1###}

    psllw   t6,t1,24        // {C___  8___  4___  0___}
    psravw  t6,t6,t2        // {####  8###  ####  0###}

    pextub  t7,t3,t4        // {## ##  ## ##  BA ##  ## ##}
    pextub  t8,t5,t6        // {## ##  ## ##  98 ##  ## ##}
    pextlb  t3,t3,t4        // {## ##  ## ##  32 ##  ## ##}
    pextlb  t5,t5,t6        // {## ##  ## ##  10 ##  ## ##}

    pextlh  t7,t7,t8        // {BA 98  ## ##  ## ##  ## ##}
    pextlh  t3,t3,t5        // {32 10  ## ##  ## ##  ## ##]

    pcpyud  t9,t3,t7        // {BA98####   3210####}

        // Second Bank     t1= [FEDC] BA98 [7654] 3210

    pexcw   t1,t1           // {FEDC  7654  BA98  3210}
    pextuw  t1,t1,t1        // {FEDC  FEDC  7654  7654}
    psravw  t3,t1,t2        // {####  F###  ####  7###}

    psllw   t4,t1,8         // {EDC_  EDC_  654_  654_}
    psravw  t4,t4,t2        // {####  E###  ####  6###}

    psllw   t5,t1,16        // {DC__  DC__  54__  54__}
    psravw  t5,t5,t2        // {####  D###  ####  5###}

    psllw   t6,t1,24        // {C___  C___  4___  4___}
    psravw  t6,t6,t2        // {####  C###  ####  4###}

    pextub  t7,t3,t4        // {#,# #,#   #,# #,#  F,E #,#  #,# #,#}
    pextub  t8,t5,t6        // {#,# #,#   #,# #,#  D,C #,#  #,# #,#}
    pextlb  t3,t3,t4        // {#,# #,#   #,# #,#  7,6 #,#  #,# #,#}
    pextlb  t5,t5,t6        // {#,# #,#   #,# #,#  5,4 #,#  #,# #,#}

    pextlh  t7,t7,t8        // {FE DC  ## ##   ## ##  ## ##}
    pextlh  t3,t3,t5        // {76 54  ## ##   ## ##  ## ##]

    pcpyud  t8,t3,t7        // {FEDC####   7654####}

    pexcw   t9,t9           // {BA98  3210  ####  ####}
    pexcw   t8,t8           // {FEDC  7654  ####  ####}

    pextuw  t0,t8,t9        // {FEDC  BA98  7654  3210}

        //  Unaligned data section

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // pbD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot




