
//
//  psrlMMI.s - Packed (128bit Mips) Shift Right Logical
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//      This module is designed for fast 3D math
//
//  Created - 26 Jan, 2002 - James Leiterman
//
//  Modifications
//
//
//  stab = 4
//

    .nolist
#include "Mips.inc"
#include "KariType.inc"
    .list

 

    .set    noreorder
    .set    nomips16

    .rdata


    .align  4


    .extern llMaskBO        //  Left Shift 32bit (unsigned) 8bit mask
    .extern rrMaskBO        //  Right Shift 32bit (unsigned) 8bit mask
    .extern llMaskHO        //  Left Shift 32bit (unsigned) 16bit mask
    .extern rrMaskHO        //  Right Shift 32bit (unsigned) 16bit mask


    .section .text                                                    



//_________________________________________________________________________________________________
//=================================================================================================
//  Packed Shift Right Logical 4x32bit by 1
//
//  void vmp_psrl1W( uint32 * const pwD, const uint32 * const pwA )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_psrl1WMMI
vmp_psrl1WMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // pwA
    LDELAY                  // nop - Load Delay Slot

    psrlw   t0,t1,1         // D = A >> 1

    sq      t0, 0(a0)       // pwD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_psrl1WMMI0
vmp_psrl1WMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // pwA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)
    LDELAY                  // nop - Load Delay Slot

    pcpyld  t1,t3,t1        // A  {127...0}

    psrlw   t0,t1,1         // D = A >> 1

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // pwD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  Packed Shift Right Logical 4x32bit by n:{0...31}
//
//  void vmp_psrlW( uint32 * const pwD, const uint32 * const pwA, uint count )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_psrlWMMI
vmp_psrlWMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // {3 2 1 0} pwA
    and     t2,a2,32-1      // 00011111b Clip count to data bit size (32-1)

    pcpyld  t2,t2,t2        // {0 count 0 count}
    psrlvw  t0,t1,t2        // {# 2 # 0}  A >> count

    pexcw   t1,t1           // {3 1 2 0}
    pexcw   t0,t0           // {# # 2 0}

    pextuw  t1,t1,t1        // {3 3 1 1}
    psrlvw  t1,t1,t2        // {# 3 # 1}  A >> count
    pexcw   t1,t1           // {# # 3 1}

    pextlw  t0,t1,t0        // {3 2 1 0}

    sq      t0, 0(a0)       // pwD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_psrlWMMI0
vmp_psrlWMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // pwA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)
    LDELAY                  // nop - Load Delay Slot

    pcpyld  t1,t3,t1        // {3 2 1 0} pwA

    and     t2,a2,32-1      // 00011111b Clip count to data bit size (32-1)
    pcpyld  t2,t2,t2        // {0 count 0 count}

    psrlvw  t0,t1,t2        // {# 2 # 0}  D = A >> count

    pexcw   t1,t1           // {3 1 2 0}
    pexcw   t0,t0           // {# # 2 0}

    pextuw  t1,t1,t1        // {3 3 1 1}
    psrlvw  t1,t1,t2        // {# 3 # 1}  D = A >> count
    pexcw   t1,t1           // {# # 3 1}

    pextlw  t0,t1,t0        // {3 2 1 0}

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // pwD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot




//_________________________________________________________________________________________________
//=================================================================================================
//  Packed Shift Right Logical 8x16bit by 1
//
//  void vmp_psrl1H( uint16 * const phD, const uint16 * const phA )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_psrl1HMMI
vmp_psrl1HMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // phA
    LDELAY                  // nop - Load Delay Slot

    psrlh   t0,t1,1         // D = A >> 1

    sq      t0, 0(a0)       // phD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_psrl1HMMI0
vmp_psrl1HMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // phA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)
    LDELAY                  // nop - Load Delay Slot

    pcpyld  t1,t3,t1        // A  {127...0}

    psrlh   t0,t1,1         // D = A >> 1

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // phD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  Packed Shift Right Logical 8x16bit by n:{0...15}
//
//  void vmp_psrlH( uint16 * const phD, const uint16 * const phA, uint count )
//
//  *** MMI version *** 128bit aligned
//
//  Handle 4x  2x16bit
    .align 3
    .global vmp_psrlHMMI
vmp_psrlHMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // {15...0} phA
    and     t2,a2,16-1      // 00001111b Clip count to data bit size (16-1)

    pcpyld  t2,t2,t2        // {0 count 0 count}
    psrlvw  t0,t1,t2        // {# 2 # 0}  A >> count

    pexcw   t1,t1           // {3 1 2 0}
    pexcw   t0,t0           // {# # 2 0}

    pextuw  t1,t1,t1        // {3 3 1 1}
    la      t3,rrMaskHO     // Right 8x16bit mask base

    psrlvw  t1,t1,t2        // {# 3 # 1}  A >> count
    sll     t2,t2,4         // x16 Table index

    pexcw   t1,t1           // {# # 3 1}
    add     t3,t3,t2        // Adjust for index

    pextlw  t0,t1,t0        // {3 2 1 0}
    lq      t3,0(t3)        // Right 8x16bit mask

    pand    t0,t0,t3        // Strip left (unwanted) bits

    sq      t0, 0(a0)       // phD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_psrlHMMI0
vmp_psrlHMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // phA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)
    LDELAY                  // nop - Load Delay Slot

    pcpyld  t1,t3,t1        // {3 2 1 0} pwA

        //  Aligned data section

    and     t2,a2,16-1      // 00001111b Clip count to data bit size (16-1)

    pcpyld  t2,t2,t2        // {0 count 0 count}
    psrlvw  t0,t1,t2        // {# 2 # 0}  A >> count

    pexcw   t1,t1           // {3 1 2 0}
    pexcw   t0,t0           // {# # 2 0}

    pextuw  t1,t1,t1        // {3 3 1 1}
    la      t3,rrMaskHO     // Right 8x16bit mask base

    psrlvw  t1,t1,t2        // {# 3 # 1}  A >> count
    sll     t2,t2,4         // x16 Table index

    pexcw   t1,t1           // {# # 3 1}
    add     t3,t3,t2        // Adjust for index

    pextlw  t0,t1,t0        // {3 2 1 0}
    lq      t3,0(t3)        // Right 8x16bit mask

    pand    t0,t0,t3        // Strip left (unwanted) bits

        //  Unaligned data section

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // phD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


//_________________________________________________________________________________________________
//=================================================================================================
//  Packed Shift Right Logical 16x8bit by 1
//
//  void vmp_psrl1B( uint8 * const phD, const uint8 * const phA )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_psrl1BMMI
vmp_psrl1BMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a0)

    la      t2,rrMaskBO+16  // 7F

    lq      t1, 0(a1)       // phA
    lq      t2, 0(t2)       // get mask

    psrlh   t0,t1,1         // D = A >> 1
    pand    t0,t0,t2

    sq      t0, 0(a0)       // phD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_psrl1BMMI0
vmp_psrl1BMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a0)

    la      t2,rrMaskBO+16  // 7F...7F mask

    ldl     t1, 7(a1)       // phA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

    lq      t2, 0(t2)       // get mask

    pcpyld  t1,t3,t1        // A  {127...0}

    psrlh   t0,t1,1         // D = A >> 1
    pand    t0,t0,t2        // 7F...7F mask

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // phD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  Packed Shift Right Logical 8x16bit by n:{0...15}
//
//  void vmp_psrlB( uint8 * const pbD, const uint8 * const pbA, uint count )
//
//  *** MMI version *** 128bit aligned
//
//  Handle 4x  2x16bit
    .align 3
    .global vmp_psrlBMMI
vmp_psrlBMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // {15...0} pbA
    and     t2,a2,8-1       // 00000111b Clip count to data bit size (8-1)

    pcpyld  t2,t2,t2        // {0 count 0 count}
    psrlvw  t0,t1,t2        // {# 2 # 0}  A >> count

    pexcw   t1,t1           // {3 1 2 0}
    pexcw   t0,t0           // {# # 2 0}

    pextuw  t1,t1,t1        // {3 3 1 1}
    la      t3,rrMaskBO     // Right 16x8bit mask base

    psrlvw  t1,t1,t2        // {# 3 # 1}  A >> count
    sll     t2,t2,4         // x16 Table index

    pexcw   t1,t1           // {# # 3 1}
    add     t3,t3,t2        // Adjust for index

    pextlw  t0,t1,t0        // {3 2 1 0}
    lq      t3,0(t3)        // Right 16x8bit mask

    pand    t0,t0,t3        // Strip left (unwanted) bits

    sq      t0, 0(a0)       // pbD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_psrlBMMI0
vmp_psrlBMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // pbA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)
    LDELAY                  // nop - Load Delay Slot

    pcpyld  t1,t3,t1        // {3 2 1 0} pwA

        //  Aligned data section

    and     t2,a2,8-1       // 00000111b Clip count to data bit size (8-1)

    pcpyld  t2,t2,t2        // {0 count 0 count}
    psrlvw  t0,t1,t2        // {# 2 # 0}  A >> count

    pexcw   t1,t1           // {3 1 2 0}
    pexcw   t0,t0           // {# # 2 0}

    pextuw  t1,t1,t1        // {3 3 1 1}
    la      t3,rrMaskBO     // Right 16x8bit mask base

    psrlvw  t1,t1,t2        // {# 3 # 1}  A >> count
    sll     t2,t2,4         // x16 Table index

    pexcw   t1,t1           // {# # 3 1}
    add     t3,t3,t2        // Adjust for index

    pextlw  t0,t1,t0        // {3 2 1 0}
    lq      t3,0(t3)        // Right 16x8bit mask

    pand    t0,t0,t3        // Strip left (unwanted) bits

        //  Unaligned data section

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // pbD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


