
//
//  FasMIPS.c - (Scalar) Addition & Subtraction
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 3 Oct 2001 - James Leiterman
//
//  Last Modification
//
//  stab=4
//


#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include    <assert.h>
#include    <stdio.h>
#include    <math.h>

#ifdef CC_VMP_PS2
#ifndef HOMEBREW
#include    <libvu0.h>
#endif
#endif

    //  Book Development Library Modules

#include    "CpuAsm.h"                                  // CPU
#include    "vmp3D.h"
#include    "vmp3DEx.h"
#include    "Fas.h"                                     // Addition & Subtraction
#include    "FasEx.h"                                   //  "           "           Extra



#ifdef HOMEBREW
typedef float       sceVu0FVECTOR[4] __attribute__((aligned (16)));
#endif



//
//  (float) Addition
//

void vmp_FAddVU( float * const pfD, float fA, float fB )
{
    sceVu0FVECTOR vD;
    
    ASSERT_PTR4(pfD);

    asm __volatile__("

    mfc1    $8,%1               #fA
    mfc1    $9,%2               #fB
    
    qmtc2    $8,vf4             #vf4=  xfer FP from EE to VU
    qmtc2    $9,vf5             #vf5=  xfer FP from EE to VU
        
    vadd.x vf4,vf4,vf5          # =a+b

    sqc2    vf4,0x0(%0)

    ":
     : "r" (vD) , "f" (fA), "f" (fB) 
     : "$8","$9");
    
    *pfD = vD[0];
}


//
//  (float) Subtraction
//

void vmp_FSubVU( float * const pfD, float fA, float fB )
{
    sceVu0FVECTOR vD;
    
    ASSERT_PTR4(pfD);

    asm __volatile__("

    mfc1    $8,%1               #fA
    mfc1    $9,%2               #fB
    
    qmtc2    $8,vf4             #vf4=  xfer FP from EE to VU
    qmtc2    $9,vf5             #vf5=  xfer FP from EE to VU
        
    vsub.x vf4,vf4,vf5          # =a+b

    sqc2    vf4,0x0(%0)         # pfD=

    ":
     : "r" (vD) , "f" (fA), "f" (fB) 
     : "$8","$9");
    
    *pfD = vD[0];
}



//
//  Setup module
//
//  NOTE: This function sets up the function vectors so the best function
//      is attached to the vector for the processor running.
//

void FasSetup( const CpuInfo * const pcinfo )
{
    ASSERT_PTR4(pcinfo);

    if ( CPUBITS_VU0 & pcinfo->nBits )
      {
        vmp_FAdd =              vmp_FAddVU;
        vmp_FSub =              vmp_FSubVU;
      }

    else
      {
        vmp_FAdd =              vmp_FAddGeneric;
        vmp_FSub =              vmp_FSubGeneric;
      }
}





