
//
//  Bench.c - Packed Addition/Subtraction
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 31 Dec 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include <stdio.h>
#include <assert.h>

    //  Book Development Library Modules
    
#include "CpuAsm.h"                                     // CPU Module
#include "vmp3D.h"
#include "Pas.h"
#include "PasEx.h"



typedef enum
{
    iTEST_PADDB,            // add 8bit
    iTEST_PADDH,            // add 16bit
    iTEST_PADDW,            // add 32bit

    iTEST_PSUBB,            // sub 8bit
    iTEST_PSUBH,            // sub 16bit
    iTEST_PSUBW,            // sub 32bit

    iTEST_PADDSB,           // add 8bit signed w/Saturation
    iTEST_PADDSH,           // add 16bit

    iTEST_PSUBSB,           // sub 8bit
    iTEST_PSUBSH,           // sub 16bit

    iTEST_PADDUSB,          // add 8bit unsigned w/Saturation
    iTEST_PADDUSH,          // add 16bit

    iTEST_PSUBUSB,          // sub 8bit
    iTEST_PSUBUSH           // sub 16bit
} iTEST;




//
//  Compare two 128bit bit blocks
//

bool vmp_IsPVEqual( void * const pvA, void * const pvB )
{
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);
    
    if ( memcmp( pvA, pvB, 16 ) )
      {
        return false;
      }

    return true;
}                                               



//
//  Test Packed Shift 128bit
//
static void test_pV( bool bAlign, iTEST iFunc )
{
    int8    v;
    uint8   buf[4*16 +15], *bt, *bu, *bv, *bw;

    if (bAlign)
      {
        bt = (uint8*) ALIGN16((int)buf);
      }
    else
      {
        bt = buf;
        if (0== (((uint)bt)&0x3)) bt += 4;
      }

    bu = bt + (128/8);
    bv = bu + (128/8);
    bw = bv + (128/8);

    v = 0;
    do {
        bt[0]=v;     bt[1]=v+1;   bt[2]=v+2;   bt[3]=v+3;
        bt[4]=v+4;   bt[5]=v+5;   bt[6]=v+6;   bt[7]=v+7;
        bt[8]=v+8;   bt[9]=v+9;   bt[10]=v+10; bt[11]=v+11;
        bt[12]=v+12; bt[13]=v+13; bt[14]=v+14; bt[15]=v+15;

        bw[0]=v-1;   bw[1]=v-3;   bw[2]=v-5;   bw[3]=v-7;
        bw[4]=v-9;   bw[5]=v-11;  bw[6]=v-13;  bw[7]=v-15;
        bw[8]=v-17;  bw[9]=v-19;  bw[10]=v-23; bw[11]=v-29;
        bw[12]=v-31; bw[13]=v-37; bw[14]=v-41; bw[15]=v-43;

        switch (iFunc)
          {
            case iTEST_PADDB:                       // add 8bit
                vmp_paddBGeneric(bu, bt, bw);
                if ( bAlign )
                    vmp_paddB(   bv, bt, bw);
                else
                    vmp_paddB0(  bv, bt, bw);
                break;

            case iTEST_PADDH:                       // add 16bit
                vmp_paddHGeneric((uint16*)bu, (uint16*)bt, (uint16*)bw);
                if ( bAlign )
                    vmp_paddH(   (uint16*)bv, (uint16*)bt, (uint16*)bw);
                else
                    vmp_paddH0(  (uint16*)bv, (uint16*)bt, (uint16*)bw);
                break;

            case iTEST_PADDW:                       // add 32bit
                vmp_paddWGeneric((uint32*)bu, (uint32*)bt, (uint32*)bw);
                if ( bAlign )
                    vmp_paddW(   (uint32*)bv, (uint32*)bt, (uint32*)bw);
                else
                    vmp_paddW0(  (uint32*)bv, (uint32*)bt, (uint32*)bw);
                break;

            case iTEST_PADDUSB:                     // add 8bit unsigned w/Saturation
                vmp_paddusBGeneric(bu, bt, bw);
                if ( bAlign )
                    vmp_paddusB(   bv, bt, bw);
                else
                    vmp_paddusB0(  bv, bt, bw);
                break;

            case iTEST_PADDUSH:                     // add 16bit unsigned w/Saturation
                vmp_paddusHGeneric((uint16*)bu, (uint16*)bt, (uint16*)bw);
                if ( bAlign )
                    vmp_paddusH(   (uint16*)bv, (uint16*)bt, (uint16*)bw);
                else
                    vmp_paddusH0(  (uint16*)bv, (uint16*)bt, (uint16*)bw);
                break;

            case iTEST_PADDSB:                      // add 8bit signed w/Saturation
                vmp_paddsBGeneric((int8*)bu, (int8*)bt, (int8*)bw);
                if ( bAlign )
                    vmp_paddsB(   (int8*)bv, (int8*)bt, (int8*)bw);
                else
                    vmp_paddsB0(  (int8*)bv, (int8*)bt, (int8*)bw);
                break;

            case iTEST_PADDSH:                      // add 16bit signed w/Saturation
                vmp_paddsHGeneric((int16*)bu, (int16*)bt, (int16*)bw);
                if ( bAlign )
                    vmp_paddsH(   (int16*)bv, (int16*)bt, (int16*)bw);
                else
                    vmp_paddsH0(  (int16*)bv, (int16*)bt, (int16*)bw);
                break;

            case iTEST_PSUBB:                       // sub 8bit
                vmp_psubBGeneric(bu, bt, bw);
                if ( bAlign )
                    vmp_psubB(   bv, bt, bw);
                else
                    vmp_psubB0(  bv, bt, bw);
                break;

            case iTEST_PSUBH:                       // sub 16bit
                vmp_psubHGeneric((uint16*)bu, (uint16*)bt, (uint16*)bw);
                if ( bAlign )
                    vmp_psubH(   (uint16*)bv, (uint16*)bt, (uint16*)bw);
                else
                    vmp_psubH0(  (uint16*)bv, (uint16*)bt, (uint16*)bw);
                break;

            case iTEST_PSUBW:                       // sub 32bit
                vmp_psubWGeneric((uint32*)bu, (uint32*)bt, (uint32*)bw);
                if ( bAlign )
                    vmp_psubW(   (uint32*)bv, (uint32*)bt, (uint32*)bw);
                else
                    vmp_psubW0(  (uint32*)bv, (uint32*)bt, (uint32*)bw);
                break;

            case iTEST_PSUBUSB:                     // sub 8bit unsigned w/Saturation
                vmp_psubusBGeneric(bu, bt, bw);
                if ( bAlign )
                    vmp_psubusB(   bv, bt, bw);
                else
                    vmp_psubusB0(  bv, bt, bw);
                break;

            case iTEST_PSUBUSH:                     // sub 16bit unsigned w/Saturation
                vmp_psubusHGeneric((uint16*)bu, (uint16*)bt, (uint16*)bw);
                if ( bAlign )
                    vmp_psubusH(   (uint16*)bv, (uint16*)bt, (uint16*)bw);
                else
                    vmp_psubusH0(  (uint16*)bv, (uint16*)bt, (uint16*)bw);
                break;

            case iTEST_PSUBSB:                      // sub 8bit signed w/Saturation
                vmp_psubsBGeneric((int8*)bu, (int8*)bt, (int8*)bw);
                if ( bAlign )
                    vmp_psubsB(   (int8*)bv, (int8*)bt, (int8*)bw);
                else
                    vmp_psubsB0(  (int8*)bv, (int8*)bt, (int8*)bw);
                break;

            case iTEST_PSUBSH:                      // sub 16bit signed w/Saturation
                vmp_psubsHGeneric((int16*)bu, (int16*)bt, (int16*)bw);
                if ( bAlign )
                    vmp_psubsH(   (int16*)bv, (int16*)bt, (int16*)bw);
                else
                    vmp_psubsH0(  (int16*)bv, (int16*)bt, (int16*)bw);
                break;

          }

        if ( !vmp_IsPVEqual( bu, bv ) )
          {
            printf( "Mismatch on %d\n",  v );
          }
      } while( 0 != ++v);
}



int main( int argc, char *argv[] )
{
    CpuInfo cinfo;
    char szBuf[ CPU_SZBUF_MAX ];

    CpuDetect( &cinfo );                                // Detect CPU
    
    printf( "\nPacked (Vector) Addition/Subtraction Test\n\n" );
    printf( "%s\n", CpuInfoStr( szBuf, &cinfo ));
    
    PasSetup( &cinfo );                             // Setup our 3D Vector stuff

    printf( "Testing (Vector) Addition 16x8bit\n" );
    test_pV( false, iTEST_PADDB );

    printf( "Testing (Vector) Addition 16x8bit (Aligned)\n" );
    test_pV( true,  iTEST_PADDB );

    printf( "Testing (Vector) Addition 8x16bit\n" );
    test_pV( false, iTEST_PADDH );

    printf( "Testing (Vector) Addition 8x16bit (Aligned)\n" );
    test_pV( true,  iTEST_PADDH );

    printf( "Testing (Vector) Addition 4x32bit\n" );
    test_pV( false, iTEST_PADDW );

    printf( "Testing (Vector) Addition 4x32bit (Aligned)\n" );
    test_pV( true,  iTEST_PADDW );
    printf( "\n" );


    printf( "Testing (Vector) Subtraction 16x8bit\n" );
    test_pV( false, iTEST_PSUBB );

    printf( "Testing (Vector) Subtraction 16x8bit (Aligned)\n" );
    test_pV( true,  iTEST_PSUBB );

    printf( "Testing (Vector) Subtraction 8x16bit\n" );
    test_pV( false, iTEST_PSUBH );

    printf( "Testing (Vector) Subtraction 8x16bit (Aligned)\n" );
    test_pV( true,  iTEST_PSUBH );

    printf( "Testing (Vector) Subtraction 4x32bit\n" );
    test_pV( false, iTEST_PSUBW );

    printf( "Testing (Vector) Subtraction 4x32bit (Aligned)\n" );
    test_pV( true,  iTEST_PSUBW );
    printf( "\n" );


    printf( "Testing (Vector) Addition 16x8bit unsigned w/Saturation\n" );
    test_pV( false, iTEST_PADDUSB );

    printf( "Testing (Vector) Addition 16x8bit unsigned w/Saturation (Aligned)\n" );
    test_pV( true,  iTEST_PADDUSB );

    printf( "Testing (Vector) Addition 8x16bit unsigned w/Saturation\n" );
    test_pV( false, iTEST_PADDUSH );

    printf( "Testing (Vector) Addition 8x16bit unsigned w/Saturation (Aligned)\n" );
    test_pV( true,  iTEST_PADDUSH );
    printf( "\n" );

    printf( "Testing (Vector) Subtraction 16x8bit unsigned w/Saturation\n" );
    test_pV( false, iTEST_PSUBUSB );

    printf( "Testing (Vector) Subtraction 16x8bit unsigned w/Saturation (Aligned)\n" );
    test_pV( true,  iTEST_PSUBUSB );

    printf( "Testing (Vector) Subtraction 8x16bit unsigned w/Saturation\n" );
    test_pV( false, iTEST_PSUBUSH );

    printf( "Testing (Vector) Subtraction 8x16bit unsigned w/Saturation (Aligned)\n" );
    test_pV( true,  iTEST_PSUBUSH );
    printf( "\n" );


    printf( "Testing (Vector) Addition 16x8bit signed w/Saturation\n" );
    test_pV( false, iTEST_PADDSB );

    printf( "Testing (Vector) Addition 16x8bit signed w/Saturation (Aligned)\n" );
    test_pV( true,  iTEST_PADDSB );

    printf( "Testing (Vector) Addition 8x16bit signed w/Saturation\n" );
    test_pV( false, iTEST_PADDSH );

    printf( "Testing (Vector) Addition 8x16bit signed w/Saturation (Aligned)\n" );
    test_pV( true,  iTEST_PADDSH );
    printf( "\n" );

    printf( "Testing (Vector) Subtraction 16x8bit signed w/Saturation\n" );
    test_pV( false, iTEST_PSUBSB );

    printf( "Testing (Vector) Subtraction 16x8bit signed w/Saturation (Aligned)\n" );
    test_pV( true,  iTEST_PSUBSB );

    printf( "Testing (Vector) Subtraction 8x16bit signed w/Saturation\n" );
    test_pV( false, iTEST_PSUBSH );

    printf( "Testing (Vector) Subtraction 8x16bit signed w/Saturation (Aligned)\n" );
    test_pV( true,  iTEST_PSUBSH );
    printf( "\n" );

    printf( "Fin!\n\n" );

    return 0;
}

