
//
//  PAddMMI.s - Packed (128bit Mips) Addition
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//      This module is designed for fast 3D math
//
//  Created - 24 Jan, 2002 - James Leiterman
//
//  Modifications
//
//
//  stab = 4
//

    .nolist
#include "Mips.inc"
#include "KariType.inc"
    .list

    .set    noreorder
    .set    nomips16

    .section .text                                                    



//_________________________________________________________________________________________________
//=================================================================================================
//  (Quad) Vector 16x8bit Summation (Addition)
//
//  void vmp_paddBMMI( uint8  * const pbD, const uint8  * const pbA, const uint8  * const pbB );
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_paddBMMI
vmp_paddBMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // pbA
    lq      t2, 0(a2)       // pbB
    LDELAY                  // mop - Load Delay Slot

    paddb   t0, t1, t2

    sq      t0, 0(a0)       // pbD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_paddBMMI0
vmp_paddBMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // pbA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

    ldl     t2, 7(a2)       // pbB  {63...0}
    ldr     t2, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t3,t1        // A  {127...0}
    pcpyld  t2,t4,t2        // B  {127...0}

    paddb   t0, t1, t2      // D  {127...0}

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // pbD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


//_________________________________________________________________________________________________
//=================================================================================================
//  (Quad) Vector Signed 16x8 Summation (Addition) w/Saturation  {-128...127}
//
//  void vmp_paddsBMMI( int8  * const pbD, const int8  * const pbA, const int8  * const pbB );
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_paddsBMMI
vmp_paddsBMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // pbA
    lq      t2, 0(a2)       // pbB
    LDELAY                  // mop - Load Delay Slot

    paddsb  t0, t1, t2

    sq      t0, 0(a0)       // pbD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_paddsBMMI0
vmp_paddsBMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // pbA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

    ldl     t2, 7(a2)       // pbB  {63...0}
    ldr     t2, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t3,t1        // A  {127...0}
    pcpyld  t2,t4,t2        // B  {127...0}

    paddsb  t0, t1, t2      // D  {127...0}

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // pbD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  (Quad) Vector Unsigned 16x8 Summation (Addition) w/Saturation  {0...255}
//
//  void vmp_paddusBMMI(      uint8  * const pbD, const uint8  * const pbA, const uint8  * const pbB );
//
//  *** MMI version *** 128bit aligned

    .align  3
    .global vmp_paddusBMMI
vmp_paddusBMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)
                                
    lq      t1, 0(a1)       // pbA
    lq      t2, 0(a2)       // pbB
    LDELAY                  // mop - Load Delay Slot

    paddub  t0, t1, t2

    sq      t0, 0(a0)       // pbD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_paddusBMMI0
vmp_paddusBMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // pbA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

    ldl     t2, 7(a2)       // pbB  {63...0}
    ldr     t2, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t3,t1        // A  {127...0}
    pcpyld  t2,t4,t2        // B  {127...0}

    paddub  t0, t1, t2      // D  {127...0}

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // pbD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  Packed (un)signed Addition 8x16bit
//
//  void vmp_paddHMMI( uint16 * const phD, const uint16 * const phA, const uint16 * const phB );
//
//  *** MMI version *** 128bit aligned

    .align  3
    .global vmp_paddHMMI
vmp_paddHMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // phA
    lq      t2, 0(a2)       // phB
    LDELAY                  // mop - Load Delay Slot

    paddh   t0, t1, t2

    sq      t0, 0(a0)       // phD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_paddHMMI0
vmp_paddHMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // phA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

    ldl     t2, 7(a2)       // phB  {63...0}
    ldr     t2, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t3,t1        // A  {127...0}
    pcpyld  t2,t4,t2        // B  {127...0}

    paddh   t0, t1, t2      // D  {127...0}

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // phD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  (Quad) Vector Signed 8x16 Summation (Addition) w/Saturation {-32768...32767}
//
//  void vmp_paddsHMMI(      int16 * const phD, const int16 * const phA, const int16 * const phB );
//
//  *** MMI version *** 128bit aligned

    .align  3
    .global vmp_paddsHMMI
vmp_paddsHMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // phA
    lq      t2, 0(a2)       // phB
    LDELAY                  // mop - Load Delay Slot

    paddsh  t0, t1, t2

    sq      t0, 0(a0)       // phD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_paddsHMMI0
vmp_paddsHMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // phA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

    ldl     t2, 7(a2)       // phB  {63...0}
    ldr     t2, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t3,t1        // A  {127...0}
    pcpyld  t2,t4,t2        // B  {127...0}

    paddsh  t0, t1, t2      // D  {127...0}

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // phD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  (Quad) Vector Unsigned 16x8 Summation (Addition) w/Saturation  {0...255}
//
//  void vmp_paddusHMMI(      uint16 * const phD, const uint16 * const phA, const uint16 * const phB );
//
//  *** MMI version *** 128bit aligned

    .align  3
    .global vmp_paddusHMMI
vmp_paddusHMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // phA
    lq      t2, 0(a2)       // phB
    LDELAY                  // mop - Load Delay Slot

    padduh  t0, t1, t2

    sq      t0, 0(a0)       // phD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_paddusHMMI0
vmp_paddusHMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // phA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

    ldl     t2, 7(a2)       // phB  {63...0}
    ldr     t2, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t3,t1        // A  {127...0}
    pcpyld  t2,t4,t2        // B  {127...0}

    padduh  t0, t1, t2      // D  {127...0}

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // phD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  Packed (un)signed Addition 4x32bit
//
//  void vmp_paddW4MMI( uint32 * const pwD, const uint32 * const pwA, const uint32 * const pwB );
//
//  *** MMI version *** 128bit aligned

    .align  3
    .global vmp_paddWMMI
vmp_paddWMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // pwA
    lq      t2, 0(a2)       // pwB
    LDELAY                  // mop - Load Delay Slot

    nop
    paddw   t0, t1, t2
    nop

    sq      t0, 0(a0)       // pwD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_paddWMMI0
vmp_paddWMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // pwA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

    ldl     t2, 7(a2)       // pwB  {63...0}
    ldr     t2, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t3,t1        // A  {127...0}
    pcpyld  t2,t4,t2        // B  {127...0}

    paddw   t0, t1, t2      // D  {127...0}

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // pwD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


