
//
//  PSubMMI.s - Packed (128bit Mips) Subtraction
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//      This module is designed for fast 3D math
//
//  Created - 24 Jan, 2002 - James Leiterman
//
//  Modifications
//
//
//  stab = 4
//

    .nolist
#include "Mips.inc"
#include "KariType.inc"
    .list

    .set    noreorder
    .set    nomips16

    .section .text                                                    



//_________________________________________________________________________________________________
//=================================================================================================
//  (Quad) Vector 16x8bit Subtraction (Addition)
//
//  void vmp_psubBMMI( uint8  * const pbD, const uint8  * const pbA, const uint8  * const pbB );
//
//  *** MMI version *** 128bit aligned

    .align  3
    .global vmp_psubBMMI
vmp_psubBMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // phA
    lq      t2, 0(a2)       // phB
    LDELAY                  // mop - Load Delay Slot

    psubb   t0, t1, t2

    sq      t0, 0(a0)       // phD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_psubBMMI0
vmp_psubBMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // phA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

    ldl     t2, 7(a2)       // phB  {63...0}
    ldr     t2, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t3,t1        // A  {127...0}
    pcpyld  t2,t4,t2        // B  {127...0}

    psubb   t0, t1, t2      // D  {127...0}

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // phD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot




//_________________________________________________________________________________________________
//=================================================================================================
//  (Quad) Vector Signed 16x8 Subtraction (Addition) w/Saturation  {-128...127}
//
//  void vmp_psubsBMMI( int8  * const pbD, const int8  * const pbA, const int8  * const pbB );
//
//  *** MMI version *** 128bit aligned

    .align  3
    .global vmp_psubsBMMI
vmp_psubsBMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // phA
    lq      t2, 0(a2)       // phB
    LDELAY                  // mop - Load Delay Slot

    psubsb  t0, t1, t2

    sq      t0, 0(a0)       // phD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_psubsBMMI0
vmp_psubsBMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // phA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

    ldl     t2, 7(a2)       // phB  {63...0}
    ldr     t2, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t3,t1        // A  {127...0}
    pcpyld  t2,t4,t2        // B  {127...0}

    psubsb  t0, t1, t2      // D  {127...0}

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // phD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  (Quad) Vector Unsigned 16x8 Subtraction (Addition) w/Saturation  {0...255}
//
//  void vmp_psubusBMMI(      uint8  * const pbD, const uint8  * const pbA, const uint8  * const pbB );
//
//  *** MMI version *** 128bit aligned

    .align  3
    .global vmp_psubusBMMI
vmp_psubusBMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // phA
    lq      t2, 0(a2)       // phB
    LDELAY                  // mop - Load Delay Slot

    psubub  t0, t1, t2

    sq      t0, 0(a0)       // phD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_psubusBMMI0
vmp_psubusBMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // phA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

    ldl     t2, 7(a2)       // phB  {63...0}
    ldr     t2, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t3,t1        // A  {127...0}
    pcpyld  t2,t4,t2        // B  {127...0}

    psubub  t0, t1, t2      // D  {127...0}

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // phD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  Packed (un)signed Subtraction 8x16bit
//
//  void vmp_psubHMMI( uint16 * const phD, const uint16 * const phA, const uint16 * const phB );
//
//  *** MMI version *** 128bit aligned

    .align  3
    .global vmp_psubHMMI
vmp_psubHMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // phA
    lq      t2, 0(a2)       // phB
    LDELAY                  // mop - Load Delay Slot

    psubh   t0, t1, t2

    sq      t0, 0(a0)       // phD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_psubHMMI0
vmp_psubHMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // phA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

    ldl     t2, 7(a2)       // phB  {63...0}
    ldr     t2, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t3,t1        // A  {127...0}
    pcpyld  t2,t4,t2        // B  {127...0}

    psubh   t0, t1, t2      // D  {127...0}

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // phD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  (Quad) Vector Signed 8x16 Subtraction (Addition) w/Saturation {-32768...32767}
//
//  void vmp_psubsHMMI(      int16 * const phD, const int16 * const phA, const int16 * const phB );
//
//  *** MMI version *** 128bit aligned

    .align  3
    .global vmp_psubsHMMI
vmp_psubsHMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // phA
    lq      t2, 0(a2)       // phB
    LDELAY                  // mop - Load Delay Slot

    psubsh  t0, t1, t2

    sq      t0, 0(a0)       // phD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot

//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_psubsHMMI0
vmp_psubsHMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // phA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

    ldl     t2, 7(a2)       // phB  {63...0}
    ldr     t2, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t3,t1        // A  {127...0}
    pcpyld  t2,t4,t2        // B  {127...0}

    psubsh  t0, t1, t2      // D  {127...0}

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // phD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  (Quad) Vector Unsigned 16x8 Subtraction (Addition) w/Saturation  {0...255}
//
//  void vmp_psubusHMMI(      uint16 * const phD, const uint16 * const phA, const uint16 * const phB );
//
//  *** MMI version *** 128bit aligned

    .align  3
    .global vmp_psubusHMMI
vmp_psubusHMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // phA
    lq      t2, 0(a2)       // phB
    LDELAY                  // mop - Load Delay Slot

    psubuh  t0, t1, t2

    sq      t0, 0(a0)       // phD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_psubusHMMI0
vmp_psubusHMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // phA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

    ldl     t2, 7(a2)       // phB  {63...0}
    ldr     t2, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t3,t1        // A  {127...0}
    pcpyld  t2,t4,t2        // B  {127...0}

    psubuh  t0, t1, t2      // D  {127...0}

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // phD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  Packed (un)signed Subtraction 4x32bit
//
//  void vmp_psubWMMI( uint32 * const pwD, const uint32 * const pwA, const uint32 * const pwB );
//
//  *** MMI version *** 128bit aligned

    .align  3
    .global vmp_psubWMMI
vmp_psubWMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // phA
    lq      t2, 0(a2)       // phB
    LDELAY                  // mop - Load Delay Slot

    psubw   t0, t1, t2

    sq      t0, 0(a0)       // phD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_psubWMMI0
vmp_psubWMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // phA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

    ldl     t2, 7(a2)       // phB  {63...0}
    ldr     t2, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t3,t1        // A  {127...0}
    pcpyld  t2,t4,t2        // B  {127...0}

    psubw   t0, t1, t2      // D  {127...0}

    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // phD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot

















