
//
//  Pas.c - Packed Addition/Subtraction
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 31 Dec 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include    <assert.h>
#include    <stdio.h>

    //  Book Development Library Modules
    
#include "Pas.h"
#include "PasEx.h"



    //  Addition (Un)signed

vmp_paddBProc vmp_paddB;
vmp_paddBProc vmp_paddB0;
vmp_paddHProc vmp_paddH;
vmp_paddHProc vmp_paddH0;
vmp_paddWProc vmp_paddW;
vmp_paddWProc vmp_paddW0;

    //  Subtraction (Un)signed

vmp_psubBProc vmp_psubB;
vmp_psubBProc vmp_psubB0;
vmp_psubHProc vmp_psubH;
vmp_psubHProc vmp_psubH0;
vmp_psubWProc vmp_psubW;
vmp_psubWProc vmp_psubW0;

    //  Addition Unsigned w/Saturation

vmp_paddusBProc vmp_paddusB;
vmp_paddusBProc vmp_paddusB0;
vmp_paddusHProc vmp_paddusH;
vmp_paddusHProc vmp_paddusH0;

    //  Subtraction Unsigned w/Saturation

vmp_psubusBProc vmp_psubusB;
vmp_psubusBProc vmp_psubusB0;
vmp_psubusHProc vmp_psubusH;
vmp_psubusHProc vmp_psubusH0;

    //  Addition signed w/Saturation

vmp_paddsBProc vmp_paddsB;
vmp_paddsBProc vmp_paddsB0;
vmp_paddsHProc vmp_paddsH;
vmp_paddsHProc vmp_paddsH0;

    //  Subtraction signed w/Saturation

vmp_psubsBProc vmp_psubsB;
vmp_psubsBProc vmp_psubsB0;
vmp_psubsHProc vmp_psubsH;
vmp_psubsHProc vmp_psubsH0;



//
//  Packed (un)signed Addition 16x8bit
//
void vmp_paddBGeneric( uint8 * const pbD, const uint8 * const pbA, const uint8 * const pbB )
{
    ASSERT_PTR4(pbD);
    ASSERT_PTR4(pbA);
    ASSERT_PTR4(pbB);

    *(pbD+0) = *(pbA+0) + *(pbB+0);
    *(pbD+1) = *(pbA+1) + *(pbB+1);
    *(pbD+2) = *(pbA+2) + *(pbB+2);
    *(pbD+3) = *(pbA+3) + *(pbB+3);
    *(pbD+4) = *(pbA+4) + *(pbB+4);
    *(pbD+5) = *(pbA+5) + *(pbB+5);
    *(pbD+6) = *(pbA+6) + *(pbB+6);
    *(pbD+7) = *(pbA+7) + *(pbB+7);
    *(pbD+8) = *(pbA+8) + *(pbB+8);
    *(pbD+9) = *(pbA+9) + *(pbB+9);
    *(pbD+10) = *(pbA+10) + *(pbB+10);
    *(pbD+11) = *(pbA+11) + *(pbB+11);
    *(pbD+12) = *(pbA+12) + *(pbB+12);
    *(pbD+13) = *(pbA+13) + *(pbB+13);
    *(pbD+14) = *(pbA+14) + *(pbB+14);
    *(pbD+15) = *(pbA+15) + *(pbB+15);
}                                   



//
//  Packed (un)signed Addition 8x16bit
//
void vmp_paddHGeneric( uint16 * const phD, const uint16 * const phA, const uint16 * const phB )
{
    ASSERT_PTR4(phD);
    ASSERT_PTR4(phA);
    ASSERT_PTR4(phB);

    *(phD+0) = *(phA+0) + *(phB+0);
    *(phD+1) = *(phA+1) + *(phB+1);
    *(phD+2) = *(phA+2) + *(phB+2);
    *(phD+3) = *(phA+3) + *(phB+3);
    *(phD+4) = *(phA+4) + *(phB+4);
    *(phD+5) = *(phA+5) + *(phB+5);
    *(phD+6) = *(phA+6) + *(phB+6);
    *(phD+7) = *(phA+7) + *(phB+7);
}                                   



//
//  Packed (un)signed Addition 4x32bit
//
void vmp_paddWGeneric( uint32 * const pwD, const uint32 * const pwA, const uint32 * const pwB )
{
    ASSERT_PTR4(pwD);
    ASSERT_PTR4(pwA);
    ASSERT_PTR4(pwB);

    *(pwD+0) = *(pwA+0) + *(pwB+0);
    *(pwD+1) = *(pwA+1) + *(pwB+1);
    *(pwD+2) = *(pwA+2) + *(pwB+2);
    *(pwD+3) = *(pwA+3) + *(pwB+3);
}                                   



//______________________________________________________



//
//  Packed unsigned Addition 16x8bit w/Saturation {0...255}
//
void vmp_paddusBGeneric( uint8 * const pbD, const uint8 * const pbA, const uint8 * const pbB )
{
    int v, cnt;
    uint8 *pD, *pA, *pB;

    ASSERT_PTR4(pbD);
    ASSERT_PTR4(pbA);
    ASSERT_PTR4(pbB);

    pD = (uint8*)pbD;
    pA = (uint8*)pbA;
    pB = (uint8*)pbB;

    cnt = 16;
    do {
        v =  (int)(((uint)*pA++) + ((uint)*pB++));
        *pD++ = (uint8)(0xff & (v  | ((0xff-v) >> 31)));            // (v<256) ? v : 255;
     } while (--cnt);
}                                   



//
//  Packed unsigned Addition 8x16bit w/Saturation {0...65535}
//
void vmp_paddusHGeneric( uint16 * const phD, const uint16 * const phA, const uint16 * const phB )
{
    int v, cnt;
    uint16 *pD, *pA, *pB;

    ASSERT_PTR4(phD);
    ASSERT_PTR4(phA);
    ASSERT_PTR4(phB);

    pD = (uint16*)phD;
    pA = (uint16*)phA;
    pB = (uint16*)phB;

    cnt = 8;
    do {
        v =  (int)(((uint)*pA++) + ((uint)*pB++));
        *pD++ = (uint16)(0xffff & (v  | ((0xffff-v) >> 31)));           // (v<65536) ? v : 65535;
     } while (--cnt);
}                                   





//______________________________________________________



//
//  Packed signed Addition 16x8bit w/Saturation {-128...127}
//
void vmp_paddsBGeneric( int8 * const pbD, const int8 * const pbA, const int8 * const pbB )
{
    int v, cnt;
    int8 *pD, *pA, *pB;

    ASSERT_PTR4(pbD);
    ASSERT_PTR4(pbA);
    ASSERT_PTR4(pbB);

    pD = (int8*)pbD;
    pA = (int8*)pbA;
    pB = (int8*)pbB;

    cnt = 16;
    do {
        v =  128 + ((int)*pA++) + ((int)*pB++);
        v &= (-1^(v>>31));                                  // (v<0) ? 0 : v;
        *pD++ = (int8)(-128 + (0xff & (v  | ((0xff-v) >> 31))));            // (v<256) ? v : 255;
     } while (--cnt);
}                                   



//
//  Packed signed Addition 8x16bit w/Saturation {-32768...32767}
//
void vmp_paddsHGeneric( int16 * const phD, const int16 * const phA, const int16 * const phB )
{
    int v, cnt;
    int16 *pD, *pA, *pB;

    ASSERT_PTR4(phD);
    ASSERT_PTR4(phA);
    ASSERT_PTR4(phB);

    pD = (int16*)phD;
    pA = (int16*)phA;
    pB = (int16*)phB;

    cnt = 8;
    do {
        v =  32768 + ((int)*pA++) + ((int)*pB++);
        v &= (-1^(v>>31));                                  // (v<0) ? 0 : v;
        *pD++ = (uint16)(-32768 + (0xffff & (v  | ((0xffff-v) >> 31))));    // (v<65536) ? v : 65535;
     } while (--cnt);
}                                   




//______________________________________________________



//
//  Packed (un)signed Subtraction 16x8bit
//
void vmp_psubBGeneric( uint8 * const pbD, const uint8 * const pbA, const uint8 * const pbB )
{
    ASSERT_PTR4(pbD);
    ASSERT_PTR4(pbA);
    ASSERT_PTR4(pbB);

    *(pbD+0) = *(pbA+0) - *(pbB+0);
    *(pbD+1) = *(pbA+1) - *(pbB+1);
    *(pbD+2) = *(pbA+2) - *(pbB+2);
    *(pbD+3) = *(pbA+3) - *(pbB+3);
    *(pbD+4) = *(pbA+4) - *(pbB+4);
    *(pbD+5) = *(pbA+5) - *(pbB+5);
    *(pbD+6) = *(pbA+6) - *(pbB+6);
    *(pbD+7) = *(pbA+7) - *(pbB+7);
    *(pbD+8) = *(pbA+8) - *(pbB+8);
    *(pbD+9) = *(pbA+9) - *(pbB+9);
    *(pbD+10) = *(pbA+10) - *(pbB+10);
    *(pbD+11) = *(pbA+11) - *(pbB+11);
    *(pbD+12) = *(pbA+12) - *(pbB+12);
    *(pbD+13) = *(pbA+13) - *(pbB+13);
    *(pbD+14) = *(pbA+14) - *(pbB+14);
    *(pbD+15) = *(pbA+15) - *(pbB+15);
}                                   



//
//  Packed (un)signed Subraction 8x16bit
//
void vmp_psubHGeneric( uint16 * const phD, const uint16 * const phA, const uint16 * const phB )
{
    ASSERT_PTR4(phD);
    ASSERT_PTR4(phA);
    ASSERT_PTR4(phB);

    *(phD+0) = *(phA+0) - *(phB+0);
    *(phD+1) = *(phA+1) - *(phB+1);
    *(phD+2) = *(phA+2) - *(phB+2);
    *(phD+3) = *(phA+3) - *(phB+3);
    *(phD+4) = *(phA+4) - *(phB+4);
    *(phD+5) = *(phA+5) - *(phB+5);
    *(phD+6) = *(phA+6) - *(phB+6);
    *(phD+7) = *(phA+7) - *(phB+7);
}                                   



//
//  Packed (un)signed Subtraction 4x32bit
//
void vmp_psubWGeneric( uint32 * const pwD, const uint32 * const pwA, const uint32 * const pwB )
{
    ASSERT_PTR4(pwD);
    ASSERT_PTR4(pwA);
    ASSERT_PTR4(pwB);

    *(pwD+0) = *(pwA+0) - *(pwB+0);
    *(pwD+1) = *(pwA+1) - *(pwB+1);
    *(pwD+2) = *(pwA+2) - *(pwB+2);
    *(pwD+3) = *(pwA+3) - *(pwB+3);
}                                   



//______________________________________________________



//
//  Packed unsigned Subtraction 16x8bit w/Saturation {0...255}
//
void vmp_psubusBGeneric( uint8 * const pbD, const uint8 * const pbA, const uint8 * const pbB )
{
    int v, cnt;
    uint8 *pD, *pA, *pB;

    ASSERT_PTR4(pbD);
    ASSERT_PTR4(pbA);
    ASSERT_PTR4(pbB);

    pD = (uint8*)pbD;
    pA = (uint8*)pbA;
    pB = (uint8*)pbB;

    cnt = 16;
    do {
        v =  (int)(((uint)*pA++) - ((uint)*pB++));
        v &= (-1^(v>>31));                                  // (v<0) ? 0 : v;
        *pD++ = (uint8)(0xff & (v  | ((0xff-v) >> 31)));    // (v<256) ? v : 255;
     } while (--cnt);
}                                   



//
//  Packed unsigned Subtraction 8x16bit w/Saturation {0...65535}
//
void vmp_psubusHGeneric( uint16 * const phD, const uint16 * const phA, const uint16 * const phB )
{
    int v, cnt;
    uint16 *pD, *pA, *pB;

    ASSERT_PTR4(phD);
    ASSERT_PTR4(phA);
    ASSERT_PTR4(phB);

    pD = (uint16*)phD;
    pA = (uint16*)phA;
    pB = (uint16*)phB;

    cnt = 8;
    do {
        v =  (int)(((uint)*pA++) - ((uint)*pB++));
        v &= (-1^(v>>31));                                  // (v<0) ? 0 : v;
        *pD++ = (uint16)(0xffff & (v  | ((0xffff-v) >> 31)));           // (v<65536) ? v : 65535;
     } while (--cnt);
}                                   





//______________________________________________________



//
//  Packed signed Subtraction 16x8bit w/Saturation {-128...127}
//
void vmp_psubsBGeneric( int8 * const pbD, const int8 * const pbA, const int8 * const pbB )
{
    int v, cnt;
    int8 *pD, *pA, *pB;

    ASSERT_PTR4(pbD);
    ASSERT_PTR4(pbA);
    ASSERT_PTR4(pbB);

    pD = (int8*)pbD;
    pA = (int8*)pbA;
    pB = (int8*)pbB;

    cnt = 16;
    do {
        v =  128 + ((int)*pA++) - ((int)*pB++);
        v &= (-1^(v>>31));                                  // (v<0) ? 0 : v;
        *pD++ = (int8)(-128 + (0xff & (v  | ((0xff-v) >> 31))));            // (v<256) ? v : 255;
     } while (--cnt);
}                                   



//
//  Packed signed Subtraction 8x16bit w/Saturation {-32768...32767}
//
void vmp_psubsHGeneric( int16 * const phD, const int16 * const phA, const int16 * const phB )
{
    int v, cnt;
    int16 *pD, *pA, *pB;

    ASSERT_PTR4(phD);
    ASSERT_PTR4(phA);
    ASSERT_PTR4(phB);

    pD = (int16*)phD;
    pA = (int16*)phA;
    pB = (int16*)phB;

    cnt = 8;
    do {
        v =  32768 + ((int)*pA++) - ((int)*pB++);
        v &= (-1^(v>>31));                                  // (v<0) ? 0 : v;
        *pD++ = (uint16)(-32768 + (0xffff & (v  | ((0xffff-v) >> 31))));    // (v<65536) ? v : 65535;
     } while (--cnt);
}                                   






