
//
//  Bench.c - (Quad) Vector 3D Addition & Subtraction
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 5 Oct 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include <stdio.h>

    //  Book Development Library Modules
    
#include "CpuAsm.h"                                     // CPU Module
#include "vmp3D.h"
#include "QVas3D.h"                                     // Vector Addition & Subtraction
#include "QVas3DEx.h"                                   //  "   " (Private)



typedef enum
{
    iTEST_ADD,
    iTEST_SUB,
    iTEST_ADD_SCALAR,
    iTEST_SUB_SCALAR
} iTEST;    





//
//  Test Vector Math
//

static void testQVec( bool bAlign, iTEST iFunc )
{
    vmp3DQVector *pvA, *pvB, *pvTmp, *pvTmp2;
    int x, y, z, w, dif;
    float fPrecision, fScalar;
    float buf[4*4  +3];         // enough space + 12 bytes (3 floats) extra

        // Force proper alignment (Crude just for this demo)

    if ( bAlign )
      {
        pvA    = (vmp3DQVector*) ALIGN16((int)(buf));
        pvB    = (vmp3DQVector*) ALIGN16((int)(pvA+1));
        pvTmp  = (vmp3DQVector*) ALIGN16((int)(pvB+1));
        pvTmp2 = (vmp3DQVector*) ALIGN16((int)(pvTmp+1));
      }
    else
      {
        pvA    = (vmp3DQVector*) buf;
        if (!(0x000f & (int)pvA))
          {
            pvA = (vmp3DQVector*) (buf+1);          // Make sure misaligned!
          }

        pvB    = pvA + 1;
        pvTmp  = pvB + 1;
        pvTmp2 = pvTmp + 1;
      }

    dif = 4;
    fPrecision = SINGLE_PRECISION;
    
    for ( w = -dif; w <= dif; w++ )
      {
        pvA->w = 0.123f *(float)w;
        pvB->z = 0.123f *(float)w;
                    
        for ( z = -dif; z <= dif; z++ )
          {
            pvA->z = 0.123f *(float)z;
            pvB->y = 0.123f *(float)z;

            for ( y = -dif; y <= dif; y++ )
              {
                pvA->y = 0.123f *(float)y;
                pvB->x = 0.123f *(float)y;

                for ( x = -dif; x <= dif; x++ )
                  {
                    pvB->w = 0.123f *(float)x;
                    pvA->x = 0.123f *(float)x;

                    fScalar = pvA->x;

                    switch (iFunc)
                      {
                        case iTEST_ADD:
                            vmp_QVecAddGeneric( pvTmp2, pvA, pvB );
                            if ( bAlign )
                                vmp_QVecAdd( pvTmp, pvA, pvB ); // Aligned
                            else
                                vmp_QVecAdd0( pvTmp, pvA, pvB );    // Aligned
                            break;

                        case iTEST_SUB:
                            vmp_QVecSubGeneric( pvTmp2, pvA, pvB );
                            if ( bAlign )
                                vmp_QVecSub( pvTmp, pvA, pvB ); // Aligned
                            else
                                vmp_QVecSub0( pvTmp, pvA, pvB );
                            break;

                        case iTEST_ADD_SCALAR:
                            vmp_QVecAddScalarGeneric( pvTmp2, pvA, fScalar );
                            if ( bAlign )
                                vmp_QVecAddScalar( pvTmp, pvA, fScalar );   // Aligned
                            else
                                vmp_QVecAddScalar0( pvTmp, pvA, fScalar );
                            break;

                        case iTEST_SUB_SCALAR:
                            vmp_QVecSubScalarGeneric( pvTmp2, pvA, fScalar );
                            if ( bAlign )
                                vmp_QVecSubScalar( pvTmp, pvA, fScalar );   // Aligned
                            else
                                vmp_QVecSubScalar0( pvTmp, pvA, fScalar );  // Aligned
                            break;
                      }

                    if (!vmp_IsQVEqual( pvTmp, pvTmp2, fPrecision)) 
                      {
                        printf( " vA.x:%f vA.y:%f vA.z:%f vA.w:%f\n",
                                 pvA->x, pvA->y, pvA->z, pvA->w );
                        printf( " vB.x:%f vB.y:%f vB.z:%f vB.w:%f\n",
                                 pvB->x, pvB->y, pvB->z, pvB->w );
                        printf( "    x:%f y:%f z:%f w:%f  Good vs Bad!!!\n",
                                 pvTmp2->x, pvTmp2->y, pvTmp2->z, pvTmp2->w );
                        printf( "    x:%f y:%f z:%f w:%f   *** MISMATCH!!! ***\n\n",
                                 pvTmp->x, pvTmp->y, pvTmp->z, pvTmp->w );
                      }
                  }
              }
          }
      }
}




int main( int argc, char *argv[] )
{
    CpuInfo cinfo;
    char szBuf[ CPU_SZBUF_MAX ];

    CpuDetect( &cinfo );                                // Detect CPU
    
    printf( "\n(Quad) Vector Addition & Subtraction Test\n\n" );
    printf( "%s\n", CpuInfoStr( szBuf, &cinfo ) );
    
    QVas3DSetup( &cinfo );                              // Setup our 3D Vector stuff

    printf( "Testing (Quad) Addition Scalar\n" );
    testQVec( false, iTEST_ADD_SCALAR );

    printf( "Testing (Quad) Addition Scalar (Aligned)\n" );
    testQVec( true,  iTEST_ADD_SCALAR );

    printf( "Testing (Quad) Subtraction Scalar\n" );
    testQVec( false, iTEST_SUB_SCALAR );

    printf( "Testing (Quad) Subtraction Scalar (Aligned)\n" );
    testQVec( true,  iTEST_SUB_SCALAR );

    printf( "Testing (Quad) Addition\n" );
    testQVec( false, iTEST_ADD );

    printf( "Testing (Quad) Addition (Aligned)\n" );
    testQVec( true,  iTEST_ADD );

    printf( "Testing (Quad) Subtraction\n" );
    testQVec( false, iTEST_SUB );

    printf( "Testing (Quad) Subtraction (Aligned)\n" );
    testQVec( true,  iTEST_SUB );

    printf( "Fin\n\n" );

    return 0;
}

