
//
//  Bench.c - Vector 3D Addition & Subtraction
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 5 Oct 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include <stdio.h>

    //  Book Development Library Modules
    
#include "CpuAsm.h"                                     // CPU Module
#include "vmp3D.h"
#include "Vas3D.h"                                      // Vector Addition & Subtraction
#include "Vas3DEx.h"                                    //  "   " (Private)



typedef enum
{
    iTEST_ADD,
    iTEST_SUB,
    iTEST_ADD_SCALAR,
    iTEST_SUB_SCALAR
} iTEST;    





//
//  Test Vector Math
//

static void testVec( bool bAlign, iTEST iFunc )
{
    vmp3DVector *pvA, *pvB, *pvTmp, *pvTmp2;
    int x, y, z, dif;
    float fPrecision, fScalar, fWTst;
    float buf[4*4  +3];         // enough space + 12 bytes (3 floats) extra

        // Force proper alignment (Crude just for this demo)

    if ( bAlign )
      {
        pvA    = (vmp3DVector*) ALIGN16((int)(buf));
        pvB    = (vmp3DVector*) ALIGN16((int)(pvA+1));
        pvTmp  = (vmp3DVector*) ALIGN16((int)(pvB+1));
        pvTmp2 = (vmp3DVector*) ALIGN16((int)(pvTmp+1));
      }
    else
      {
        pvA    = (vmp3DVector*) buf;
        if (!(0x000f & (int)pvA))
          {
            pvA = (vmp3DVector*) (buf+1);           // Make sure misaligned!
          }

        pvB    = pvA + 1;
        pvTmp2 = pvB + 1;
        pvTmp  = pvTmp2 + 1;    // no tail, for .W testing
      }

    dif = 4;
    fPrecision = SINGLE_PRECISION;
    
    for ( z = -dif; z <= dif; z++ )
      {
        pvA->z = 0.123f *(float)z;
        pvB->y = 0.123f *(float)z;

        for ( y = -dif; y <= dif; y++ )
          {
            pvA->y = 0.123f *(float)y;
            pvB->x = 0.123f *(float)y;

            for ( x = -dif; x <= dif; x++ )
              { 
                pvA->x = 0.123f *(float)x;
                pvB->z = 0.123f *(float)y;
                fScalar = pvA->x;

                fWTst = ((vmp3DQVector *) pvTmp)->w = fScalar;  // Get untouchable value!

                switch (iFunc)
                  {
                    case iTEST_ADD:
                        vmp_VecAddGeneric( pvTmp2, pvA, pvB );
                        if ( bAlign )
                            vmp_VecAdd( pvTmp, pvA, pvB );  // Aligned
                        else
                            vmp_VecAdd0( pvTmp, pvA, pvB ); // Aligned
                        break;

                    case iTEST_SUB:
                        vmp_VecSubGeneric( pvTmp2, pvA, pvB );
                        if ( bAlign )
                            vmp_VecSub( pvTmp, pvA, pvB );  // Aligned
                        else
                            vmp_VecSub0( pvTmp, pvA, pvB );
                        break;

                    case iTEST_ADD_SCALAR:
                        vmp_VecAddScalarGeneric( pvTmp2, pvA, fScalar );
                        if ( bAlign )
                            vmp_VecAddScalar( pvTmp, pvA, fScalar );    // Aligned
                        else
                            vmp_VecAddScalar0( pvTmp, pvA, fScalar );
                        break;

                    case iTEST_SUB_SCALAR:
                        vmp_VecSubScalarGeneric( pvTmp2, pvA, fScalar );
                        if ( bAlign )
                            vmp_VecSubScalar( pvTmp, pvA, fScalar );    // Aligned
                        else
                            vmp_VecSubScalar0( pvTmp, pvA, fScalar );   // Aligned
                        break;
                  }

                if (!vmp_IsVEqual(pvTmp, pvTmp2, fPrecision)) 
                  {
                    printf( " vA.x:%f vA.y:%f vA.z:%f\n",
                             pvA->x, pvA->y, pvA->z );
                    printf( " vB.x:%f vB.y:%f vB.z:%f\n",
                             pvB->x, pvB->y, pvB->z );
                    printf( "    x:%f y:%f z:%f  Good vs Bad!!!\n",
                             pvTmp2->x, pvTmp2->y, pvTmp2->z );
                    printf( "    x:%f y:%f z:%f    *** MISMATCH!!! ***\n\n",
                             pvTmp->x, pvTmp->y, pvTmp->z );
                  }

                    //  Check for '.W' contamination
                if (!vmp_IsFEqual(fWTst, ((vmp3DQVector *) pvTmp)->w, fPrecision ))
                  {
                    printf( ".W is contaminated! should be: %f not %f\n", fWTst, ((vmp3DQVector *) pvTmp)->w );
                  }
              }
          }
      }
}




int main( int argc, char *argv[] )
{
    CpuInfo cinfo;
    char szBuf[ CPU_SZBUF_MAX ];

    CpuDetect( &cinfo );                                // Detect CPU
 
    printf( "\nVector Addition & Subtraction Test\n\n" );
    printf( "%s\n", CpuInfoStr( szBuf, &cinfo ) );
    
    Vas3DSetup( &cinfo );                               // Setup our 3D Vector stuff

    printf( "Testing Addition Scalar\n" );
    testVec( false, iTEST_ADD_SCALAR );

    printf( "Testing Addition Scalar (Aligned)\n" );
    testVec( true,  iTEST_ADD_SCALAR );

    printf( "Testing Subtraction Scalar\n" );
    testVec( false, iTEST_SUB_SCALAR );

    printf( "Testing Subtraction Scalar (Aligned)\n" );
    testVec( true,  iTEST_SUB_SCALAR );

    printf( "Testing Addition\n" );
    testVec( false, iTEST_ADD );

    printf( "Testing Addition (Aligned)\n" );
    testVec( true,  iTEST_ADD );

    printf( "Testing Subtraction\n" );
    testVec( false, iTEST_SUB );

    printf( "Testing Subtraction (Aligned)\n" );
    testVec( true,  iTEST_SUB );

    printf( "Fin\n\n" );

    return 0;
}

