
//
//  Bench.c - (Scalar) Multiplication & Division
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 5 Oct 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//

#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include <stdio.h>

    //  Book Development Library Modules

#include "CpuAsm.h"                                     // CPU Module
#include "vmp3D.h"
#include "Fmd.h"
#include "FmdEx.h"



typedef enum
{
    iTEST_FLT_MUL,
    iTEST_FLT_DIV,
    iTEST_FLT_DIV_FAST
} iTEST;



//
//  Test 
//

static void testVal( iTEST iFunc )
{
    int x, z, dif;
    float f, f2, fPrecision, fA, fB;

    dif = 4;

    fPrecision = SINGLE_PRECISION;
    
    for ( z = -dif; z <= dif; z++ )
      {
        fB = 0.123f *(float)z;

        for ( x = -dif; x <= dif; x++ )
          { 
            fA = 0.123f *(float)x;

            if ( 0.0f == fA ) continue;
            if ( 0.0f == fB ) continue;

            switch(iFunc)
              {
                case iTEST_FLT_MUL:
                    fPrecision = SINGLE_PRECISION;
                    vmp_FMulGeneric( &f2, fA, fB );
                    vmp_FMul( &f, fA, fB );
                    break;

                case iTEST_FLT_DIV:
                    fPrecision = SINGLE_PRECISION;
                    vmp_FDivGeneric( &f2, fA, fB );
                    vmp_FDiv( &f, fA, fB );
                    break;

                case iTEST_FLT_DIV_FAST:
                    fPrecision = FAST_PRECISION;
                    vmp_FDivGeneric( &f2, fA, fB );
                    vmp_FDivFast( &f, fA, fB );
                    break;
              }

            if (!vmp_IsFEqual(f, f2, fPrecision))
              {
                printf( " %f vs %f    fA:%f fB:%f  *** MISMATCH!!! ***\n", f2, f, fA, fB );
              }
          }
      }
}



int main( int argc, char *argv[] )
{
    CpuInfo cinfo;
    char szBuf[ CPU_SZBUF_MAX ];

    CpuDetect( &cinfo );                                // Detect CPU

    printf( "\nScalar Multiplication & Division\n\n" );
    printf( "%s\n", CpuInfoStr( szBuf, &cinfo ) );

    FmdSetup( &cinfo );                             // Setup (float) stuff

    printf( "Testing (Scalar) Multiplication\n" );
    testVal( iTEST_FLT_MUL );

    printf( "Testing (Scalar) Division\n" );
    testVal( iTEST_FLT_DIV );

    printf( "Testing (Scalar) Division *FAST (Low Precision)*\n" );
    testVal( iTEST_FLT_DIV_FAST );

    printf( "Fin!\n\n" );

    return 0;
}

