
//
//  FmdMIPS.c - (Scalar) Multiplication & Division
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 3 Oct 2001 - James Leiterman
//
//  Last Modification
//
//  stab=4
//


#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include    <assert.h>
#include    <stdio.h>
#include    <math.h>

    //  Book Development Library Modules

#include    "CpuAsm.h"                                  // CPU
#include    "vmp3D.h"
#include    "vmp3DEx.h"
#include    "Fmd.h"                                 // (float) Math
#include    "FmdEx.h"


typedef float       sceVu0FVECTOR[4] __attribute__((aligned (16)));


//
//  (float) Multiplication
//

void vmp_FMulVU( float * const pfD, float fA, float fB )
{
    sceVu0FVECTOR vD;
    
    ASSERT_PTR4(pfD);

    asm __volatile__("

    mfc1    $8,%1               #fA
    mfc1    $9,%2               #fB
    
    qmtc2    $8,vf4             #vf4=  xfer FP from EE to VU
    qmtc2    $9,vf5             #vf5=  xfer FP from EE to VU
        
    vmul.x vf4,vf4,vf5          # =a*b

    sqc2    vf4,0x0(%0)         #pfD=

    ":
     : "r" (vD) , "f" (fA), "f" (fB) 
     : "$8","$9");
    
    *pfD = vD[0];
}



//
//  (float) Division
//

void vmp_FDivVU( float * const pfD, float fA, float fB )
{
    sceVu0FVECTOR vD;
    
    ASSERT_PTR4(pfD);
    ASSERT_ZERO(fB);
    
    asm __volatile__("

    mfc1    $8,%1               #fA
    mfc1    $9,%2               #fB
    
    qmtc2    $8,vf4             #vf4=  xfer FP from EE to VU
    qmtc2    $9,vf5             #vf5=  xfer FP from EE to VU
        
    vdiv     Q,vf4x,vf5x        # q=a/b
    vwaitq      
    vaddq.x vf4x, vf0x, Q       #  =0+q

    sqc2    vf4,0x0(%0)         #pfD=

    ":
     : "r" (vD) , "f" (fA), "f" (fB) 
     : "$8","$9");
    
    *pfD = vD[0];
}



//
//  Setup Vec3D module
//
//  NOTE: This function sets up the function vectors so the best function
//      is attached to the vector for the processor running.
//

void FmdSetup( const CpuInfo * const pcinfo )
{
    ASSERT_PTR4(pcinfo);
    
    if ( CPUBITS_VU0 & pcinfo->nBits )
      {
        vmp_FMul =              vmp_FMulVU;
        vmp_FDiv =              vmp_FDivVU;
        vmp_FDivFast =          vmp_FDivVU;
      }

    else
      {
        vmp_FMul =              vmp_FMulGeneric;
        vmp_FDiv =              vmp_FDivGeneric;
        vmp_FDivFast =          vmp_FDivGeneric;
      }
}





