
//
//  Bench.c - Packed Multiplication/Division
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 31 Dec 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include <stdio.h>
#include <assert.h>

    //  Book Development Library Modules
    
#include "CpuAsm.h"                                     // CPU Module
#include "vmp3D.h"
#include "Pmd.h"
#include "PmdEx.h"



typedef enum
{
    iTEST_PMULHB,           // mul     signed High  8bit 16x8bit
    iTEST_PMULHUB,          // mul   unsigned High  8bit 16x8bit
    iTEST_PMULLUB,          // mul (un)signed Low   8bit 16x8bit
    iTEST_PMULLB,           // mul     signed Low   8bit 16x8bit

    iTEST_PMULHH,           // mul     signed High 16bit 8x16bit
    iTEST_PMULHUH,          // mul   unsigned High 16bit 8x16bit
    iTEST_PMULLUH,          // mul (un)signed Low  16bit 8x16bit    
    iTEST_PMULLH,           // mul     signed Low  16bit 8x16bit

    iTEST_PMULLUW,          // mul (un)signed Low  32bit 4x32
    iTEST_PMULLW,           // mul     signed Low  32bit 4x32

    iTEST_PMULHLB,          // mul     signed hi/lo 2x8bit 16x8bit
    iTEST_PMULHLUB,         // mul   unsigned hi/lo 2x8bit 16x8bit
    iTEST_PMULHLH,          // mul     signed hi/lo 2x16bit 8x16bit
    iTEST_PMULHLUH,         // mul   unsigned hi/lo 2x16bit 8x16bit

    iTEST_PMULLUEB,         // mul   unsigned even 16bit 16x8bit
    iTEST_PMULLUOB,         // mul   unsigned odd  16bit 16x8bit
    iTEST_PMULLEB,          // mul     signed even 16bit 16x8bit
    iTEST_PMULLOB,          // mul     signed odd  16bit 16x8bit

    iTEST_PMULLUEH,         // mul   unsigned even 32bit 8x16bit
    iTEST_PMULLUOH,         // mul   unsigned odd  32bit 8x16bit
    iTEST_PMULLEH,          // mul     signed even 32bit 8x16bit
    iTEST_PMULLOH           // mul     signed odd  32bit 8x16bit

} iTEST;




//
//  Compare two 128bit bit blocks
//

bool vmp_IsPVEqual( void * const pvA, void * const pvB )
{
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);
    
    if ( memcmp( pvA, pvB, 16 ) )
      {
        return false;
      }

    return true;
}                                               



//
//  Test Packed Shift 128bit
//
static void test_pV( bool bAlign, iTEST iFunc )
{
    int8    v;
    uint8   buf[6*16 +15], *bt, *bu, *bv, *bw, *bx, *by;
    bool    bHL;

    if (bAlign)
      {
        bt = (uint8*) ALIGN16((int)buf);
      }
    else
      {
        bt = buf;
        if (0== (((uint)bt)&0x3)) bt += 4;
      }

    bu = bt + (128/8);
    bv = bu + (128/8);
    bw = bv + (128/8);
    bx = bw + (128/8);
    by = bx + (128/8);

    bHL = false;
    v = 0;
    do {
        bt[0]=v;     bt[1]=v+1;   bt[2]=v+2;   bt[3]=v+3;
        bt[4]=v+4;   bt[5]=v+5;   bt[6]=v+6;   bt[7]=v+7;
        bt[8]=v+8;   bt[9]=v+9;   bt[10]=v+10; bt[11]=v+11;
        bt[12]=v+12; bt[13]=v+13; bt[14]=v+14; bt[15]=v+15;

        bw[0]=v+15;  bw[1]=v+14;  bw[2]=v+13;  bw[3]=v+12;
        bw[4]=v+11;  bw[5]=v+10;  bw[6]=v+9;   bw[7]=v+8;
        bw[8]=v+7;   bw[9]=v+6;   bw[10]=v+5;  bw[11]=v+4;
        bw[12]=v+3;  bw[13]=v+2;  bw[14]=v+1;  bw[15]=v+0;

        switch (iFunc)
          {
            case iTEST_PMULHB:                      // mul     signed High  8bit 16x8bit
                vmp_pmulhBGeneric((int8*)bu, (int8*)bt, (int8*)bw);
                if ( bAlign )
                    vmp_pmulhB(   (int8*)bv, (int8*)bt, (int8*)bw);
                else
                    vmp_pmulhB0(  (int8*)bv, (int8*)bt, (int8*)bw);
                break;

            case iTEST_PMULHUB:                     // mul   unsigned High  8bit 16x8bit
                vmp_pmulhuBGeneric((uint8*)bu, (uint8*)bt, (uint8*)bw);
                if ( bAlign )
                    vmp_pmulhuB(   (uint8*)bv, (uint8*)bt, (uint8*)bw);
                else
                    vmp_pmulhuB0(  (uint8*)bv, (uint8*)bt, (uint8*)bw);
                break;

            case iTEST_PMULLUB:                     // mul (un)signed Low   8bit 16x8bit
                vmp_pmulluBGeneric((uint8*)bu, (uint8*)bt, (uint8*)bw);
                if ( bAlign )
                    vmp_pmulluB(   (uint8*)bv, (uint8*)bt, (uint8*)bw);
                else
                    vmp_pmulluB0(  (uint8*)bv, (uint8*)bt, (uint8*)bw);
                break;

            case iTEST_PMULLB:                      // mul     signed Low   8bit 16x8bit
                vmp_pmullBGeneric((int8*)bu, (int8*)bt, (int8*)bw);
                if ( bAlign )
                    vmp_pmullB(   (int8*)bv, (int8*)bt, (int8*)bw);
                else
                    vmp_pmullB0(  (int8*)bv, (int8*)bt, (int8*)bw);
                break;

            case iTEST_PMULHH:                      // mul     signed High 16bit 8x16bit
                vmp_pmulhHGeneric((int16*)bu, (int16*)bt, (int16*)bw);
                if ( bAlign )
                    vmp_pmulhH(   (int16*)bv, (int16*)bt, (int16*)bw);
                else
                    vmp_pmulhH0(  (int16*)bv, (int16*)bt, (int16*)bw);
                break;

            case iTEST_PMULHUH:                     // mul   unsigned High 16bit 8x16bit
                vmp_pmulhuHGeneric((uint16*)bu, (uint16*)bt, (uint16*)bw);
                if ( bAlign )
                    vmp_pmulhuH(   (uint16*)bv, (uint16*)bt, (uint16*)bw);
                else
                    vmp_pmulhuH0(  (uint16*)bv, (uint16*)bt, (uint16*)bw);
                break;

            case iTEST_PMULLUH:                     // mul (un)signed Low  16bit 8x16bit    
                vmp_pmulluHGeneric((uint16*)bu, (uint16*)bt, (uint16*)bw);
                if ( bAlign )
                    vmp_pmulluH(   (uint16*)bv, (uint16*)bt, (uint16*)bw);
                else
                    vmp_pmulluH0(  (uint16*)bv, (uint16*)bt, (uint16*)bw);
                break;

            case iTEST_PMULLH:                      // mul signed Low  16bit 8x16bit    
                vmp_pmullHGeneric((int16*)bu, (int16*)bt, (int16*)bw);
                if ( bAlign )                    
                    vmp_pmullH(   (int16*)bv, (int16*)bt, (int16*)bw);
                else
                    vmp_pmullH0(  (int16*)bv, (int16*)bt, (int16*)bw);
                break;

            case iTEST_PMULLUW:                     // mul (un)signed Low  32bit 4x32
                vmp_pmulluWGeneric((uint32*)bu, (uint32*)bt, (uint32*)bw);
                if ( bAlign )
                    vmp_pmulluW(   (uint32*)bv, (uint32*)bt, (uint32*)bw);
                else
                    vmp_pmulluW0(  (uint32*)bv, (uint32*)bt, (uint32*)bw);
                break;

            case iTEST_PMULLW:                      // mul     signed Low  32bit 4x32
                vmp_pmullWGeneric((int32*)bu, (int32*)bt, (int32*)bw);
                if ( bAlign )
                    vmp_pmullW(   (int32*)bv, (int32*)bt, (int32*)bw);
                else                            
                    vmp_pmullW0(  (int32*)bv, (int32*)bt, (int32*)bw);
                break;

            case iTEST_PMULHLB:                     // mul     signed hi/lo 2x8bit 16x8bit
                bHL = true;
                vmp_pmulhlBGeneric((int8*)bx, (int8*)bu, (int8*)bt, (int8*)bw);
                if ( bAlign )
                    vmp_pmulhlB(   (int8*)by, (int8*)bv, (int8*)bt, (int8*)bw);
                else
                    vmp_pmulhlB0(  (int8*)by, (int8*)bv, (int8*)bt, (int8*)bw);
                break;

            case iTEST_PMULHLUB:                    // mul   unsigned hi/lo 2x8bit 16x8bit
                bHL = true;
                vmp_pmulhluBGeneric((uint8*)bx, (uint8*)bu, (uint8*)bt, (uint8*)bw);
                if ( bAlign )
                    vmp_pmulhluB(   (uint8*)by, (uint8*)bv, (uint8*)bt, (uint8*)bw);
                else
                    vmp_pmulhluB0(  (uint8*)by, (uint8*)bv, (uint8*)bt, (uint8*)bw);
                break;

            case iTEST_PMULHLH:                 // mul     signed hi/lo 2x16bit 8x16bit
                bHL = true;
                vmp_pmulhlHGeneric((int16*)bx, (int16*)bu, (int16*)bt, (int16*)bw);
                if ( bAlign )
                    vmp_pmulhlH(   (int16*)by, (int16*)bv, (int16*)bt, (int16*)bw);
                else
                    vmp_pmulhlH0(  (int16*)by, (int16*)bv, (int16*)bt, (int16*)bw);
                break;

            case iTEST_PMULHLUH:                // mul   unsigned hi/lo 2x16bit 8x16bit
                bHL = true;
                vmp_pmulhluHGeneric((uint16*)bx, (uint16*)bu, (uint16*)bt, (uint16*)bw);
                if ( bAlign )
                    vmp_pmulhluH(   (uint16*)by, (uint16*)bv, (uint16*)bt, (uint16*)bw);
                else
                    vmp_pmulhluH0(  (uint16*)by, (uint16*)bv, (uint16*)bt, (uint16*)bw);
                break;

            case iTEST_PMULLUEB:                // mul   unsigned even 16bit 16x8bit
                vmp_pmullueBGeneric((uint16*)bu, (uint8*)bt, (uint8*)bw);
                if ( bAlign )
                    vmp_pmullueB(   (uint16*)bv, (uint8*)bt, (uint8*)bw);
                else
                    vmp_pmullueB0(  (uint16*)bv, (uint8*)bt, (uint8*)bw);
                break;

            case iTEST_PMULLUOB:                // mul   unsigned odd  16bit 16x8bit
                vmp_pmulluoBGeneric((uint16*)bu, (uint8*)bt, (uint8*)bw);
                if ( bAlign )
                    vmp_pmulluoB(   (uint16*)bv, (uint8*)bt, (uint8*)bw);
                else
                    vmp_pmulluoB0(  (uint16*)bv, (uint8*)bt, (uint8*)bw);
                break;

            case iTEST_PMULLEB:                 // mul     signed even 16bit 16x8bit
                vmp_pmulleBGeneric((int16*)bu, (int8*)bt, (int8*)bw);
                if ( bAlign )
                    vmp_pmulleB(   (int16*)bv, (int8*)bt, (int8*)bw);
                else
                    vmp_pmulleB0(  (int16*)bv, (int8*)bt, (int8*)bw);
                break;

            case iTEST_PMULLOB:                 // mul     signed odd  16bit 16x8bit
                vmp_pmulloBGeneric((int16*)bu, (int8*)bt, (int8*)bw);
                if ( bAlign )
                    vmp_pmulloB(   (int16*)bv, (int8*)bt, (int8*)bw);
                else
                    vmp_pmulloB0(  (int16*)bv, (int8*)bt, (int8*)bw);
                break;

            case iTEST_PMULLUEH:                // mul   unsigned even 32bit 8x16bit
                vmp_pmullueHGeneric((uint32*)bu, (uint16*)bt, (uint16*)bw);
                if ( bAlign )
                    vmp_pmullueH(   (uint32*)bv, (uint16*)bt, (uint16*)bw);
                else
                    vmp_pmullueH0(  (uint32*)bv, (uint16*)bt, (uint16*)bw);
                break;

            case iTEST_PMULLUOH:                // mul   unsigned odd  32bit 8x16bit
                vmp_pmulluoHGeneric((uint32*)bu, (uint16*)bt, (uint16*)bw);
                if ( bAlign )
                    vmp_pmulluoH(   (uint32*)bv, (uint16*)bt, (uint16*)bw);
                else
                    vmp_pmulluoH0(  (uint32*)bv, (uint16*)bt, (uint16*)bw);
                break;

            case iTEST_PMULLEH:                 // mul     signed even 32bit 8x16bit
                vmp_pmulleHGeneric((int32*)bu, (int16*)bt, (int16*)bw);
                if ( bAlign )
                    vmp_pmulleH(   (int32*)bv, (int16*)bt, (int16*)bw);
                else
                    vmp_pmulleH0(  (int32*)bv, (int16*)bt, (int16*)bw);
                break;

            case iTEST_PMULLOH:                 // mul     signed odd  32bit 8x16bit
                vmp_pmulloHGeneric((int32*)bu, (int16*)bt, (int16*)bw);
                if ( bAlign )
                    vmp_pmulloH(   (int32*)bv, (int16*)bt, (int16*)bw);
                else
                    vmp_pmulloH0(  (int32*)bv, (int16*)bt, (int16*)bw);
                break;

          }

        if ( !vmp_IsPVEqual( bu, bv ) )
          {
            printf( "Mismatch on %d\n",  v );
//break;
          }

        if ( bHL )
          {
            if ( !vmp_IsPVEqual( bx, by ) )
              {
                printf( "Mismatch on %d\n",  v );
//break;
              }
          }
      } while( 0 != ++v);
}



int main( int argc, char *argv[] )
{
    CpuInfo cinfo;
    char szBuf[ CPU_SZBUF_MAX ];

    CpuDetect( &cinfo );                                // Detect CPU
    
    printf( "\nPacked (Vector) Multiplicaion/Division Test\n\n" );
    printf( "%s\n", CpuInfoStr( szBuf, &cinfo ));
    
    PmdSetup( &cinfo );                             // Setup our 3D Vector stuff

    printf( "Testing (Vector) Multiplication     signed High 8bit = 16x8bit\n" );
    test_pV( false, iTEST_PMULHB );

    printf( "Testing (Vector) Multiplication     signed High 8bit = 16x8bit (Aligned)\n" );
    test_pV( true,  iTEST_PMULHB );

    printf( "Testing (Vector) Multiplication   unsigned High 8bit = 16x8bit\n" );
    test_pV( false, iTEST_PMULHUB );

    printf( "Testing (Vector) Multiplication   unsigned High 8bit = 16x8bit (Aligned)\n" );
    test_pV( true,  iTEST_PMULHUB );

    printf( "Testing (Vector) Multiplication (un)signed Low  8bit = 16x8bit\n" );
    test_pV( false, iTEST_PMULLUB );

    printf( "Testing (Vector) Multiplication (un)signed Low  8bit = 16x8bit (Aligned)\n" );
    test_pV( true,  iTEST_PMULLUB );

    printf( "Testing (Vector) Multiplication     signed Low  8bit = 16x8bit\n" );
    test_pV( false, iTEST_PMULLB );

    printf( "Testing (Vector) Multiplication     signed Low  8bit = 16x8bit (Aligned)\n" );
    test_pV( true,  iTEST_PMULLB );
    printf( "\n" );

    printf( "Testing (Vector) Multiplication     signed High 16bit = 8x16bit\n" );
    test_pV( false, iTEST_PMULHH );

    printf( "Testing (Vector) Multiplication     signed High 16bit = 8x16bit (Aligned)\n" );
    test_pV( true,  iTEST_PMULHH );

    printf( "Testing (Vector) Multiplication   unsigned High 16bit = 8x16bit\n" );
    test_pV( false, iTEST_PMULHUH );

    printf( "Testing (Vector) Multiplication   unsigned High 16bit = 8x16bit (Aligned)\n" );
    test_pV( true,  iTEST_PMULHUH );

    printf( "Testing (Vector) Multiplication (un)signed Low  16bit = 8x16bit\n" );
    test_pV( false, iTEST_PMULLUH );

    printf( "Testing (Vector) Multiplication (un)signed Low  16bit = 8x16bit (Aligned)\n" );
    test_pV( true,  iTEST_PMULLUH );

    printf( "Testing (Vector) Multiplication     signed Low  16bit = 8x16bit\n" );
    test_pV( false, iTEST_PMULLH );

    printf( "Testing (Vector) Multiplication     signed Low  16bit = 8x16bit (Aligned)\n" );
    test_pV( true,  iTEST_PMULLH );
    printf( "\n" );


    printf( "Testing (Vector) Multiplication (un)signed Low  32bit = 4x32bit\n" );
    test_pV( false, iTEST_PMULLUW );

    printf( "Testing (Vector) Multiplication (un)signed Low  32bit = 4x32bit (Aligned)\n" );
    test_pV( true,  iTEST_PMULLUW );

    printf( "Testing (Vector) Multiplication     signed Low  32bit = 4x32bit\n" );
    test_pV( false, iTEST_PMULLW );

    printf( "Testing (Vector) Multiplication     signed Low  32bit = 4x32bit (Aligned)\n" );
    test_pV( true,  iTEST_PMULLW );
    printf( "\n" );


    printf( "Testing (Vector) Multiplication   signed High/Low 2x8bit = 16x8bit\n" );
    test_pV( false, iTEST_PMULHLB );

    printf( "Testing (Vector) Multiplication   signed High/Low 2x8bit = 16x8bit (Aligned)\n" );
    test_pV( true,  iTEST_PMULHLB );

    printf( "Testing (Vector) Multiplication unsigned High/Low 2x8bit = 16x8bit\n" );
    test_pV( false, iTEST_PMULHLUB );

    printf( "Testing (Vector) Multiplication unsigned High/Low 2x8bit = 16x8bit (Aligned)\n" );
    test_pV( true,  iTEST_PMULHLUB );

    printf( "Testing (Vector) Multiplication   signed High/Low 2x16bit = 8x16bit\n" );
    test_pV( false, iTEST_PMULHLH );

    printf( "Testing (Vector) Multiplication   signed High/Low 2x16bit = 8x16bit (Aligned)\n" );
    test_pV( true,  iTEST_PMULHLH );

    printf( "Testing (Vector) Multiplication unsigned High/Low 2x16bit = 8x16bit\n" );
    test_pV( false, iTEST_PMULHLUH );

    printf( "Testing (Vector) Multiplication unsigned High/Low 2x16bit = 8x16bit (Aligned)\n" );
    test_pV( true,  iTEST_PMULHLUH );
    printf( "\n" );


    printf( "Testing (Vector) Multiplication unsigned even 16bit = 16x8bit\n" );
    test_pV( false, iTEST_PMULLUEB );

    printf( "Testing (Vector) Multiplication unsigned even 16bit = 16x8bit (Aligned)\n" );
    test_pV( true,  iTEST_PMULLUEB );

    printf( "Testing (Vector) Multiplication unsigned odd  16bit = 16x8bit\n" );
    test_pV( false, iTEST_PMULLUOB );

    printf( "Testing (Vector) Multiplication unsigned odd  16bit = 16x8bit (Aligned)\n" );
    test_pV( true,  iTEST_PMULLUOB );

    printf( "Testing (Vector) Multiplication   signed even 16bit = 16x8bit\n" );
    test_pV( false, iTEST_PMULLEB );

    printf( "Testing (Vector) Multiplication   signed even 16bit = 16x8bit (Aligned)\n" );
    test_pV( true,  iTEST_PMULLEB );

    printf( "Testing (Vector) Multiplication   signed odd  16bit = 16x8bit\n" );
    test_pV( false, iTEST_PMULLOB );

    printf( "Testing (Vector) Multiplication   signed odd  16bit = 16x8bit (Aligned)\n" );
    test_pV( true,  iTEST_PMULLOB );
    printf( "\n" );


    printf( "Testing (Vector) Multiplication unsigned even 32bit = 8x16bit\n" );
    test_pV( false, iTEST_PMULLUEH );

    printf( "Testing (Vector) Multiplication unsigned even 32bit = 8x16bit (Aligned)\n" );
    test_pV( true,  iTEST_PMULLUEH );

    printf( "Testing (Vector) Multiplication unsigned odd  32bit = 8x16bit\n" );
    test_pV( false, iTEST_PMULLUOH );

    printf( "Testing (Vector) Multiplication unsigned odd  32bit = 8x16bit (Aligned)\n" );
    test_pV( true,  iTEST_PMULLUOH );

    printf( "Testing (Vector) Multiplication   signed even 32bit = 8x16bit\n" );
    test_pV( false, iTEST_PMULLEH );

    printf( "Testing (Vector) Multiplication   signed even 32bit = 8x16bit (Aligned)\n" );
    test_pV( true,  iTEST_PMULLEH );

    printf( "Testing (Vector) Multiplication   signed odd  32bit = 8x16bit\n" );
    test_pV( false, iTEST_PMULLOH );

    printf( "Testing (Vector) Multiplication   signed odd  32bit = 8x16bit (Aligned)\n" );
    test_pV( true,  iTEST_PMULLOH );
    printf( "\n" );

    printf( "Fin!\n\n" );

    return 0;
}

