
//
//  PMulMMI.s - Packed (128bit Mips) Multiplication
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//      This module is designed for fast 3D math
//
//  Created - 18 Feb, 2002 - James Leiterman
//
//  Modifications
//
//
//  stab = 4
//

    .nolist
#include "Mips.inc"
#include "KariType.inc"
    .list

    .set    noreorder
    .set    nomips16

    .section .text                                                    



//_________________________________________________________________________________________________
//=================================================================================================
//  Packed signed Multiplication High(8bit) = 16x8bit
//
//  void vmp_pmulhB( int8 * const pbD, const int8 * const pbA, const int8 * const pbB )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_pmulhBMMI
vmp_pmulhBMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // pbA
    lq      t3, 0(a2)       // pbB

        //  Convert 16x8bit to 2x 8x16bit
    pextlb  t0, t1, t1      // {A7  A7  ... A0 A0}
    pextub  t1, t1, t1      // {A15 A15 ... A8 A8}
    pextlb  t2, t3, t3      // {B7  B7  ... B0 B0}
    pextub  t3, t3, t3      // {B15 B15 ... B8 B8}

    psrah   t0,t0,8         // +- {A7...A0}
    psrah   t1,t1,8
    psrah   t2,t2,8
    psrah   t3,t3,8

        //  2x 16bit multiplies
    pmulth  t0, t0, t2      // {D6 D4 D2 D0} A7...A0 * B7...B0
    pmfhl.uw    t2          // {D7 D5 D3 D1} LO:HI

    pmulth  t1, t1, t3      // {D14 D12 D10 D8} A15...A8 * B15...B8
    pmfhl.uw    t3          // {D15 D13 D11 D9} LO:HI

        //  Pack lower 16bits of 32
    pinteh  t2,t2,t0        // +-{D7 ... D0} 16bit
    pinteh  t3,t3,t1        // +-{D15... D8} 16bit

        //  Only want upper 8bit results
    psrlh   t2,t2,8         // 0 {D7 ... D0}
    psrlh   t3,t3,8         // 0 {D15... D8}

        //  Pack 16x8bit
    ppacb   t0,t3,t2        // {D15...D0} Upper 8bits

    sq      t0, 0(a0)       // pbD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_pmulhBMMI0
vmp_pmulhBMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // pbA  {63...0}
    ldr     t1, 0(a1)
    ldl     t2, 15(a1)      //     {127...64}
    ldr     t2, 8(a1)

    ldl     t3, 7(a2)       // pbB  {63...0}
    ldr     t3, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t2,t1        // A  {127...0}
    pcpyld  t3,t4,t3        // B  {127...0}
//  ________________

        //  Convert 16x8bit to 2x 8x16bit
    pextlb  t0, t1, t1      // {A7  A7  ... A0 A0}
    pextub  t1, t1, t1      // {A15 A15 ... A8 A8}
    pextlb  t2, t3, t3      // {B7  B7  ... B0 B0}
    pextub  t3, t3, t3      // {B15 B15 ... B8 B8}

    psrah   t0,t0,8         // +- {A7...A0}
    psrah   t1,t1,8
    psrah   t2,t2,8
    psrah   t3,t3,8

        //  2x 16bit multiplies
    pmulth  t0, t0, t2      // {D6 D4 D2 D0} A7...A0 * B7...B0
    pmfhl.uw    t2          // {D7 D5 D3 D1} LO:HI

    pmulth  t1, t1, t3      // {D14 D12 D10 D8} A15...A8 * B15...B8
    pmfhl.uw    t3          // {D15 D13 D11 D9} LO:HI

        //  Pack lower 16bits of 32
    pinteh  t2,t2,t0        // +-{D7 ... D0} 16bit
    pinteh  t3,t3,t1        // +-{D15... D8} 16bit

        //  Only want upper 8bit results
    psrlh   t2,t2,8         // 0 {D7 ... D0}
    psrlh   t3,t3,8         // 0 {D15... D8}

        //  Pack 16x8bit
    ppacb   t0,t3,t2        // {D15...D0} Upper 8bits

//  ________________
    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // pbD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  Packed unsigned Multiplication High(8bit) = 16x8bit
//
//  void vmp_pmulhuB( uint8 * const pbD, const uint8 * const pbA, const uint8 * const pbB )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_pmulhuBMMI
vmp_pmulhuBMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // pbA
    lq      t3, 0(a2)       // pbB

        //  Convert 16x8bit to 2x 8x16bit
    pextlb  t0, t1, t1      // {A7  A7  ... A0 A0}
    pextub  t1, t1, t1      // {A15 A15 ... A8 A8}
    pextlb  t2, t3, t3      // {B7  B7  ... B0 B0}
    pextub  t3, t3, t3      // {B15 B15 ... B8 B8}

    psrlh   t0,t0,8         // 0 {A7...A0}
    psrlh   t1,t1,8
    psrlh   t2,t2,8
    psrlh   t3,t3,8

        //  2x 16bit multiplies
    pmulth  t0, t0, t2      // {D6 D4 D2 D0} A7...A0 * B7...B0
    pmfhl.uw    t2          // {D7 D5 D3 D1} LO:HI

    pmulth  t1, t1, t3      // {D14 D12 D10 D8} A15...A8 * B15...B8
    pmfhl.uw    t3          // {D15 D13 D11 D9} LO:HI

        //  Pack lower 16bits of 32
    pinteh  t2,t2,t0        // +-{D7 ... D0} 16bit
    pinteh  t3,t3,t1        // +-{D15... D8} 16bit

        //  Only want upper 8bit results
    psrlh   t2,t2,8         // 0 {D7 ... D0}
    psrlh   t3,t3,8         // 0 {D15... D8}

        //  Pack 16x8bit
    ppacb   t0,t3,t2        // {D15...D0} Upper 8bits

    sq      t0, 0(a0)       // pbD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_pmulhuBMMI0
vmp_pmulhuBMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // pbA  {63...0}
    ldr     t1, 0(a1)
    ldl     t2, 15(a1)      //     {127...64}
    ldr     t2, 8(a1)

    ldl     t3, 7(a2)       // pbB  {63...0}
    ldr     t3, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t2,t1        // A  {127...0}
    pcpyld  t3,t4,t3        // B  {127...0}
//  ________________

        //  Convert 16x8bit to 2x 8x16bit
    pextlb  t0, t1, t1      // {A7  A7  ... A0 A0}
    pextub  t1, t1, t1      // {A15 A15 ... A8 A8}
    pextlb  t2, t3, t3      // {B7  B7  ... B0 B0}
    pextub  t3, t3, t3      // {B15 B15 ... B8 B8}

    psrlh   t0,t0,8         // 0 {A7...A0}
    psrlh   t1,t1,8
    psrlh   t2,t2,8
    psrlh   t3,t3,8

        //  2x 16bit multiplies
    pmulth  t0, t0, t2      // {D6 D4 D2 D0} A7...A0 * B7...B0
    pmfhl.uw    t2          // {D7 D5 D3 D1} LO:HI

    pmulth  t1, t1, t3      // {D14 D12 D10 D8} A15...A8 * B15...B8
    pmfhl.uw    t3          // {D15 D13 D11 D9} LO:HI

        //  Pack lower 16bits of 32
    pinteh  t2,t2,t0        // +-{D7 ... D0} 16bit
    pinteh  t3,t3,t1        // +-{D15... D8} 16bit

        //  Only want upper 8bit results
    psrlh   t2,t2,8         // 0 {D7 ... D0}
    psrlh   t3,t3,8         // 0 {D15... D8}

        //  Pack 16x8bit
    ppacb   t0,t3,t2        // {D15...D0} Upper 8bits

//  ________________
    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // pbD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  Packed (un)signed Multiplication Low(8bit) = 16x8bit
//
//  void vmp_pmulluB( uint8 * const pbD, const uint8 * const pbA, const uint8 * const pbB )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_pmulluBMMI
vmp_pmulluBMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // pbA
    lq      t3, 0(a2)       // pbB

        //  Convert 16x8bit to 2x 8x16bit
    pextlb  t0, t1, t1      // {A7  A7  ... A0 A0}
    pextub  t1, t1, t1      // {A15 A15 ... A8 A8}
    pextlb  t2, t3, t3      // {B7  B7  ... B0 B0}
    pextub  t3, t3, t3      // {B15 B15 ... B8 B8}

    psrah   t0,t0,8         // +- {A7...A0}
    psrah   t1,t1,8
    psrah   t2,t2,8
    psrah   t3,t3,8

        //  2x 16bit multiplies
    pmulth  t0, t0, t2      // {D6 D4 D2 D0} A7...A0 * B7...B0
    pmfhl.uw    t2          // {D7 D5 D3 D1} LO:HI

    pmulth  t1, t1, t3      // {D14 D12 D10 D8} A15...A8 * B15...B8
    pmfhl.uw    t3          // {D15 D13 D11 D9} LO:HI

        //  Pack lower 16bits of 32
    pinteh  t2,t2,t0        // +-{D7 ... D0} 16bit
    pinteh  t3,t3,t1        // +-{D15... D8} 16bit

        //  Pack 16x8bit
    ppacb   t0,t3,t2        // {D15...D0} Upper 8bits

    sq      t0, 0(a0)       // pbD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_pmulluBMMI0
vmp_pmulluBMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // pbA  {63...0}
    ldr     t1, 0(a1)
    ldl     t2, 15(a1)      //     {127...64}
    ldr     t2, 8(a1)

    ldl     t3, 7(a2)       // pbB  {63...0}
    ldr     t3, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t2,t1        // A  {127...0}
    pcpyld  t3,t4,t3        // B  {127...0}
//  ________________

        //  Convert 16x8bit to 2x 8x16bit
    pextlb  t0, t1, t1      // {A7  A7  ... A0 A0}
    pextub  t1, t1, t1      // {A15 A15 ... A8 A8}
    pextlb  t2, t3, t3      // {B7  B7  ... B0 B0}
    pextub  t3, t3, t3      // {B15 B15 ... B8 B8}

    psrah   t0,t0,8         // +- {A7...A0}
    psrah   t1,t1,8
    psrah   t2,t2,8
    psrah   t3,t3,8

        //  2x 16bit multiplies
    pmulth  t0, t0, t2      // {D6 D4 D2 D0} A7...A0 * B7...B0
    pmfhl.uw    t2          // {D7 D5 D3 D1} LO:HI

    pmulth  t1, t1, t3      // {D14 D12 D10 D8} A15...A8 * B15...B8
    pmfhl.uw    t3          // {D15 D13 D11 D9} LO:HI

        //  Pack lower 16bits of 32
    pinteh  t2,t2,t0        // +-{D7 ... D0} 16bit
    pinteh  t3,t3,t1        // +-{D15... D8} 16bit

        //  Pack 16x8bit
    ppacb   t0,t3,t2        // {D15...D0} Upper 8bits

//  ________________
    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // pbD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//=================================================================================================
//  Packed signed Multiplication High(16bit) = 8x16bit
//
//  void vmp_pmulhH( int16 * const phD, const int16 * const phA, const int16 * const phB )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_pmulhHMMI
vmp_pmulhHMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // phA
    lq      t2, 0(a2)       // phB
    LDELAY                  // nop - Load Delay Slot

    pmulth  t0, t1, t2      // Dst = {D6 D4 D2 D0}

    pmfhi   t1              // HI  = {D7 D6 D3 D2}
    pmflo   t2              // LO  = {D5 D4 D1 D0}

    psrlw   t1,t1,16        // {_D7 _D6 _D3 _D2}
    psrlw   t2,t2,16        // {_D5 _D4 _D1 _D0}

    ppach   t1,t1,t2        // {D7 D6 D3 D2 D5 D4 D1 D0}
    pexcw   t0,t1           // {D7 D6 D5_D4 D3_D2 D1 D0}

    sq      t0, 0(a0)       // phD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_pmulhHMMI0
vmp_pmulhHMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // phA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

    ldl     t2, 7(a2)       // phB  {63...0}
    ldr     t2, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t3,t1        // A  {127...0}
    pcpyld  t2,t4,t2        // B  {127...0}
//  ________________

    pmulth  t0, t1, t2      // Dst = {D6 D4 D2 D0}
    pmfhi   t1              // HI  = {D7 D6 D3 D2}
    pmflo   t2              // LO  = {D5 D4 D1 D0}

    psrlw   t1,t1,16        // {_D7 _D6 _D3 _D2}
    psrlw   t2,t2,16        // {_D5 _D4 _D1 _D0}

    ppach   t1,t1,t2        // {D7 D6 D3 D2 D5 D4 D1 D0}
    pexcw   t0,t1           // {D7 D6 D5_D4 D3_D2 D1 D0}

//  ________________
    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // phD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  Packed unsigned Multiplication High(16bit) = 8x16bit
//
//  void vmp_pmulhuH( uint16 * const phD, const uint16 * const phA, const uint16 * const phB )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_pmulhuHMMI
vmp_pmulhuHMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // pwA
    lq      t2, 0(a2)       // pwB

    //  Expand unsigned 16bit into 32bit
    pinteh  t0,zero,t1      // {0 A6 0 A4 0 A2 0 A0}
    pinteh  t3,zero,t2      // {0 B6 0 B4 0 B2 0 B0}
    psrlw   t1,t1,16        // {0 A7 0 A5 0 A3 0 A1}
    psrlw   t2,t2,16        // {0 B7 0 B5 0 B3 0 B1}

    //  32bit unsigned multiplication
    pmultuw t4, t0, t3      // {A4*B4  A0*B0} even
    pexeh   t0, t0          // {0 A4 0 A6 0 A0 0 A2}
    pexeh   t3, t3          // {0 B4 0 B6 0 B0 0 B2}
    
    pmultuw t5, t1, t2      // {A5*B5  A1*B1} odd
    pexeh   t1, t1          // {0 A5 0 A7 0 A1 0 A3}
    pexeh   t2, t2          // {0 B5 0 B7 0 B1 0 B3}

    pmultuw t0, t0, t3      // {A6*B6 A2*B2}
    pmultuw t1, t1, t2      // {A7*B7 A3*B3}

    //  Repackage 32bit into upper 16bit
    ppacw   t0, t0, t4      // {D6 D2 D4 D0}
    ppacw   t1, t1, t5      // {D7 D3 D5 D1}

    psrlw   t0,t0,16        // Shift upper 16 Bits into position
    psrlw   t1,t1,16

    pexcw   t0,t0           // {D6 D4 D2 D0}
    pexcw   t1,t1           // {D7 D5 D3 D1}

    pinteh  t0,t1,t0        // {D7 D6 D5 D4 D3 D2 D1 D0}

    sq      t0, 0(a0)       // pwD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_pmulhuHMMI0
vmp_pmulhuHMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // pwA  {63...0}
    ldr     t1, 0(a1)
    ldl     t2, 15(a1)      //     {127...64}
    ldr     t2, 8(a1)

    ldl     t3, 7(a2)       // pwB  {63...0}
    ldr     t3, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t2,t1        // A  {127...0}
    pcpyld  t2,t4,t3        // B  {127...0}
//  ________________

    //  Expand unsigned 16bit into 32bit
    pinteh  t0,zero,t1      // {0 A6 0 A4 0 A2 0 A0}
    pinteh  t3,zero,t2      // {0 B6 0 B4 0 B2 0 B0}
    psrlw   t1,t1,16        // {0 A7 0 A5 0 A3 0 A1}
    psrlw   t2,t2,16        // {0 B7 0 B5 0 B3 0 B1}

    //  32bit unsigned multiplication
    pmultuw t4, t0, t3      // {A4*B4  A0*B0} even
    pexeh   t0, t0          // {0 A4 0 A6 0 A0 0 A2}
    pexeh   t3, t3          // {0 B4 0 B6 0 B0 0 B2}
    
    pmultuw t5, t1, t2      // {A5*B5  A1*B1} odd
    pexeh   t1, t1          // {0 A5 0 A7 0 A1 0 A3}
    pexeh   t2, t2          // {0 B5 0 B7 0 B1 0 B3}

    pmultuw t0, t0, t3      // {A6*B6 A2*B2}
    pmultuw t1, t1, t2      // {A7*B7 A3*B3}

    //  Repackage 32bit into upper 16bit
    ppacw   t0, t0, t4      // {D6 D2 D4 D0}
    ppacw   t1, t1, t5      // {D7 D3 D5 D1}

    psrlw   t0,t0,16        // Shift upper 16 Bits into position
    psrlw   t1,t1,16

    pexcw   t0,t0           // {D6 D4 D2 D0}
    pexcw   t1,t1           // {D7 D5 D3 D1}

    pinteh  t0,t1,t0        // {D7 D6 D5 D4 D3 D2 D1 D0}

//  ________________
    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // pwD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//=================================================================================================
//  Packed (un)signed Multiplication Low(16bit) = 8x16bit
//
//  void vmp_pmulluH( uint16 * const phD, const uint16 * const phA, const uint16 * const phB )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_pmulluHMMI
vmp_pmulluHMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // phA
    lq      t2, 0(a2)       // phB
    LDELAY                  // nop - Load Delay Slot

    pmulth  t0, t1, t2      // Dst = {D6 D4 D2 D0}
    pmfhi   t1              // HI  = {D7 D6 D3 D2}
    pmflo   t2              // LO  = {D5 D4 D1 D0}

    ppach   t1,t1,t2        // {D7 D6 D3 D2 D5 D4 D1 D0}
    pexcw   t0,t1           // {D7 D6 D5_D4 D3_D2 D1 D0}

    sq      t0, 0(a0)       // phD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_pmulluHMMI0
vmp_pmulluHMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // phA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

    ldl     t2, 7(a2)       // phB  {63...0}
    ldr     t2, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t3,t1        // A  {127...0}
    pcpyld  t2,t4,t2        // B  {127...0}
//  ________________

    pmulth  t0, t1, t2      // Dst = {D6 D4 D2 D0}
    pmfhi   t1              // HI  = {D7 D6 D3 D2}
    pmflo   t2              // LO  = {D5 D4 D1 D0}

    ppach   t1,t1,t2        // {D7 D6 D3 D2 D5 D4 D1 D0}
    pexcw   t0,t1           // {D7 D6 D5_D4 D3_D2 D1 D0}

//  ________________
    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // phD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  Packed (un)signed Multiplication Low(32bit) = 4x32bit
//
//  void vmp_pmulluW( uint32 * const pwD, const uint32 * const pwA, const uint32 * const pwB )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_pmulluWMMI
vmp_pmulluWMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // pwA
    lq      t2, 0(a2)       // pwB

        //  2x 32bit multiplies
    pmultuw t0, t1, t2      // {A2*B2 A0*B0}

    pmthl.lw    t1          // HI={# A3 # A1} LO={# A2 # A0}
    pmfhi       t1          // {# A3 # A1}

    pmthl.lw    t2          // HI={# B3 # B1} LO={# B2 # B0}
    pmfhi       t2          // {# B3 # B1}

    pmultuw t1, t1, t2      // {A3*B3 A1*B1}

    ppacw   t1,t1,t0        // {D3 D1 D2 D0}
    pexcw   t0,t1           // {D3 D2 D1 D0}

    sq      t0, 0(a0)       // pwD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_pmulluWMMI0
vmp_pmulluWMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // pwA  {63...0}
    ldr     t1, 0(a1)
    ldl     t2, 15(a1)      //     {127...64}
    ldr     t2, 8(a1)

    ldl     t3, 7(a2)       // pwB  {63...0}
    ldr     t3, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t2,t1        // A  {127...0}
    pcpyld  t2,t4,t3        // B  {127...0}
//  ________________

        //  2x 2x 32bit multiplies
    pmultuw t0, t1, t2      // {A2*B2 A0*B0}

    pmthl.lw    t1          // HI={# A3 # A1} LO={# A2 # A0}
    pmfhi       t1          // {# A3 # A1}

    pmthl.lw    t2          // HI={# B3 # B1} LO={# B2 # B0}
    pmfhi       t2          // {# B3 # B1}

    pmultuw t1, t1, t2      // {A3*B3 A1*B1}

    ppacw   t1,t1,t0        // {D3 D1 D2 D0}
    pexcw   t0,t1           // {D3 D2 D1 D0}
//  ________________
    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // pwD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//*************************************************************************************************



//=================================================================================================
//  Packed signed Multiplication Hi/Lo(2x8bit) = 16x8bit
//
//  void vmp_pmulhlB( int8 * const pbHi, int8 * const pbLo, const int8 * const pbA, const int8 * const pbB )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_pmulhlBMMI
vmp_pmulhlBMMI:

ASSERT_PTR16(a2)
ASSERT_PTR16(a3)
ASSERT_PTR16(a1)
ASSERT_PTR16(a0)

    lq      t1, 0(a2)       // pbA
    lq      t3, 0(a3)       // pbB

        //  Convert 16x8bit to 2x 8x16bit
    pextlb  t0, t1, t1      // {A7  A7  ... A0 A0}
    pextub  t1, t1, t1      // {A15 A15 ... A8 A8}
    pextlb  t2, t3, t3      // {B7  B7  ... B0 B0}
    pextub  t3, t3, t3      // {B15 B15 ... B8 B8}

    psrah   t0,t0,8         // +- {A7...A0}
    psrah   t1,t1,8
    psrah   t2,t2,8
    psrah   t3,t3,8

        //  2x 16bit multiplies
    pmulth  t0, t0, t2      // {D6 D4 D2 D0} A7...A0 * B7...B0
    pmfhl.uw    t2          // {D7 D5 D3 D1} LO:HI

    pmulth  t1, t1, t3      // {D14 D12 D10 D8} A15...A8 * B15...B8
    pmfhl.uw    t3          // {D15 D13 D11 D9} LO:HI

        //  Pack lower 16bits of 32
    pinteh  t2,t2,t0        // +-{D7 ... D0} 16bit
    pinteh  t3,t3,t1        // +-{D15... D8} 16bit

        //  Pack 16x8bit
    ppacb   t0,t3,t2        // {D15...D0} Upper 8bits

        //  Only want upper 8bit results
    psrlh   t2,t2,8         // 0 {D7 ... D0}
    psrlh   t3,t3,8         // 0 {D15... D8}

        //  Pack 16x8bit
    ppacb   t1,t3,t2        // {D15...D0} Upper 8bits

    sq      t0, 0(a1)       // pbLo {D7...D0} Lower 16bits
    sq      t1, 0(a0)       // pbHi {D7...D0} Upper 16bits

    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_pmulhlBMMI0
vmp_pmulhlBMMI0:

ASSERT_PTR4(a2)
ASSERT_PTR4(a3)
ASSERT_PTR4(a1)
ASSERT_PTR4(a0)

    ldl     t1, 7(a2)       // pbA  {63...0}
    ldr     t1, 0(a2)
    ldl     t2, 15(a2)      //     {127...64}
    ldr     t2, 8(a2)

    ldl     t3, 7(a3)       // pbB  {63...0}
    ldr     t3, 0(a3)
    ldl     t4, 15(a3)      //     {127...64}
    ldr     t4, 8(a3)

    pcpyld  t1,t2,t1        // A  {127...0}
    pcpyld  t3,t4,t3        // B  {127...0}
//  ________________

        //  Convert 16x8bit to 2x 8x16bit
    pextlb  t0, t1, t1      // {A7  A7  ... A0 A0}
    pextub  t1, t1, t1      // {A15 A15 ... A8 A8}
    pextlb  t2, t3, t3      // {B7  B7  ... B0 B0}
    pextub  t3, t3, t3      // {B15 B15 ... B8 B8}

    psrah   t0,t0,8         // +- {A7...A0}
    psrah   t1,t1,8
    psrah   t2,t2,8
    psrah   t3,t3,8

        //  2x 16bit multiplies
    pmulth  t0, t0, t2      // {D6 D4 D2 D0} A7...A0 * B7...B0
    pmfhl.uw    t2          // {D7 D5 D3 D1} LO:HI

    pmulth  t1, t1, t3      // {D14 D12 D10 D8} A15...A8 * B15...B8
    pmfhl.uw    t3          // {D15 D13 D11 D9} LO:HI

        //  Pack lower 16bits of 32
    pinteh  t2,t2,t0        // +-{D7 ... D0} 16bit
    pinteh  t3,t3,t1        // +-{D15... D8} 16bit
    
        //  Pack 16x8bit
    ppacb   t0,t3,t2        // {D15...D0} Lower 8bits

        //  Only want upper 8bit results
    psrlh   t2,t2,8         // 0 {D7 ... D0}
    psrlh   t3,t3,8         // 0 {D15... D8}

        //  Pack 16x8bit
    ppacb   t1,t3,t2        // {D15...D0} Upper 8bits

//  ________________
    pcpyud  t2,t0,t0        // D  {63...0  127...64} phLo
    pcpyud  t3,t1,t1        // D  {63...0  127...64} phHi

    sdl     t0, 7(a1)       // pbLo {63...0}
    sdr     t0, 0(a1)
    sdl     t2, 15(a1)      //    {127...64}
    sdr     t2, 8(a1)

    sdl     t1, 7(a0)       // pbHi {63...0}
    sdr     t1, 0(a0)
    sdl     t3, 15(a0)      //    {127...64}
    sdr     t3, 8(a0)

    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//=================================================================================================
//  Packed unsigned Multiplication Hi/Lo(2x8bit) = 16x8bit
//
//  void vmp_pmulhluB( uint8 * const pbHi, uint8 * const pbLo, const uint8 * const pbA, const uint8 * const pbB )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_pmulhluBMMI
vmp_pmulhluBMMI:

ASSERT_PTR16(a2)
ASSERT_PTR16(a3)
ASSERT_PTR16(a1)
ASSERT_PTR16(a0)

    lq      t1, 0(a2)       // pbA
    lq      t3, 0(a3)       // pbB

        //  Convert 16x8bit to 2x 8x16bit
    pextlb  t0, t1, t1      // {A7  A7  ... A0 A0}
    pextub  t1, t1, t1      // {A15 A15 ... A8 A8}
    pextlb  t2, t3, t3      // {B7  B7  ... B0 B0}
    pextub  t3, t3, t3      // {B15 B15 ... B8 B8}

    psrlh   t0,t0,8         // 0 {A7...A0}
    psrlh   t1,t1,8
    psrlh   t2,t2,8
    psrlh   t3,t3,8

        //  2x 16bit multiplies
    pmulth  t0, t0, t2      // {D6 D4 D2 D0} A7...A0 * B7...B0
    pmfhl.uw    t2          // {D7 D5 D3 D1} LO:HI

    pmulth  t1, t1, t3      // {D14 D12 D10 D8} A15...A8 * B15...B8
    pmfhl.uw    t3          // {D15 D13 D11 D9} LO:HI

        //  Pack lower 16bits of 32
    pinteh  t2,t2,t0        // +-{D7 ... D0} 16bit
    pinteh  t3,t3,t1        // +-{D15... D8} 16bit

        //  Pack 16x8bit
    ppacb   t0,t3,t2        // {D15...D0} Upper 8bits

        //  Only want upper 8bit results
    psrlh   t2,t2,8         // 0 {D7 ... D0}
    psrlh   t3,t3,8         // 0 {D15... D8}

        //  Pack 16x8bit
    ppacb   t1,t3,t2        // {D15...D0} Upper 8bits

    sq      t0, 0(a1)       // pbLo {D7...D0} Lower 16bits
    sq      t1, 0(a0)       // pbHi {D7...D0} Upper 16bits

    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_pmulhluBMMI0
vmp_pmulhluBMMI0:

ASSERT_PTR4(a2)
ASSERT_PTR4(a3)
ASSERT_PTR4(a1)
ASSERT_PTR4(a0)

    ldl     t1, 7(a2)       // pbA  {63...0}
    ldr     t1, 0(a2)
    ldl     t2, 15(a2)      //     {127...64}
    ldr     t2, 8(a2)

    ldl     t3, 7(a3)       // pbB  {63...0}
    ldr     t3, 0(a3)
    ldl     t4, 15(a3)      //     {127...64}
    ldr     t4, 8(a3)

    pcpyld  t1,t2,t1        // A  {127...0}
    pcpyld  t3,t4,t3        // B  {127...0}
//  ________________

        //  Convert 16x8bit to 2x 8x16bit
    pextlb  t0, t1, t1      // {A7  A7  ... A0 A0}
    pextub  t1, t1, t1      // {A15 A15 ... A8 A8}
    pextlb  t2, t3, t3      // {B7  B7  ... B0 B0}
    pextub  t3, t3, t3      // {B15 B15 ... B8 B8}

    psrlh   t0,t0,8         // 0 {A7...A0}
    psrlh   t1,t1,8
    psrlh   t2,t2,8
    psrlh   t3,t3,8

        //  2x 16bit multiplies
    pmulth  t0, t0, t2      // {D6 D4 D2 D0} A7...A0 * B7...B0
    pmfhl.uw    t2          // {D7 D5 D3 D1} LO:HI

    pmulth  t1, t1, t3      // {D14 D12 D10 D8} A15...A8 * B15...B8
    pmfhl.uw    t3          // {D15 D13 D11 D9} LO:HI

        //  Pack lower 16bits of 32
    pinteh  t2,t2,t0        // +-{D7 ... D0} 16bit
    pinteh  t3,t3,t1        // +-{D15... D8} 16bit
    
        //  Pack 16x8bit
    ppacb   t0,t3,t2        // {D15...D0} Lower 8bits

        //  Only want upper 8bit results
    psrlh   t2,t2,8         // 0 {D7 ... D0}
    psrlh   t3,t3,8         // 0 {D15... D8}

        //  Pack 16x8bit
    ppacb   t1,t3,t2        // {D15...D0} Upper 8bits

//  ________________
    pcpyud  t2,t0,t0        // D  {63...0  127...64} phLo
    pcpyud  t3,t1,t1        // D  {63...0  127...64} phHi

    sdl     t0, 7(a1)       // pbLo {63...0}
    sdr     t0, 0(a1)
    sdl     t2, 15(a1)      //    {127...64}
    sdr     t2, 8(a1)

    sdl     t1, 7(a0)       // pbHi {63...0}
    sdr     t1, 0(a0)
    sdl     t3, 15(a0)      //    {127...64}
    sdr     t3, 8(a0)

    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//=================================================================================================
//  Packed signed Multiplication Hi/Lo(2x16bit) = 8x16bit
//
//  void vmp_pmulhlH( int16 * const phHi, int16 * const phLo, const int16 * const phA, const int16 * const phB )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_pmulhlHMMI
vmp_pmulhlHMMI:

ASSERT_PTR16(a2)
ASSERT_PTR16(a3)
ASSERT_PTR16(a1)
ASSERT_PTR16(a0)

    lq      t1, 0(a2)       // phA
    lq      t2, 0(a3)       // phB
    LDELAY                  // nop - Load Delay Slot

    pmulth  t1, t1, t2      // Dst = {D6 D4 D2 D0}
    pmfhl.uw    t2          // {D7 D5 D3 D1}

    pinteh  t0,t2,t1

    psrlw   t2,t2,16        // {_D7 _D5 _D3 _D1}
    psrlw   t1,t1,16        // {_D6 _D4 _D2 _D0}
    pinteh  t1,t2,t1

    sq      t0, 0(a1)       // phLo {D7...D0} Lower 16bits
    sq      t1, 0(a0)       // phHi {D7...D0} Upper 16bits

    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_pmulhlHMMI0
vmp_pmulhlHMMI0:

ASSERT_PTR4(a2)
ASSERT_PTR4(a3)
ASSERT_PTR4(a1)
ASSERT_PTR4(a0)

    ldl     t1, 7(a2)       // phA  {63...0}
    ldr     t1, 0(a2)
    ldl     t3, 15(a2)      //     {127...64}
    ldr     t3, 8(a2)

    ldl     t2, 7(a3)       // phB  {63...0}
    ldr     t2, 0(a3)
    ldl     t4, 15(a3)      //     {127...64}
    ldr     t4, 8(a3)

    pcpyld  t1,t3,t1        // A  {127...0}
    pcpyld  t2,t4,t2        // B  {127...0}
//  ________________

    pmulth  t1, t1, t2      // Dst = {D6 D4 D2 D0}
    pmfhl.uw    t2          // {D7 D5 D3 D1}

    pinteh  t0,t2,t1

    psrlw   t2,t2,16        // {_D7 _D5 _D3 _D1}
    psrlw   t1,t1,16        // {_D6 _D4 _D2 _D0}
    pinteh  t1,t2,t1

//  ________________
    pcpyud  t2,t0,t0        // D  {63...0  127...64} phLo
    pcpyud  t3,t1,t1        // D  {63...0  127...64} phHi

    sdl     t0, 7(a1)       // phLo {63...0}
    sdr     t0, 0(a1)
    sdl     t2, 15(a1)      //    {127...64}
    sdr     t2, 8(a1)

    sdl     t1, 7(a0)       // phHi {63...0}
    sdr     t1, 0(a0)
    sdl     t3, 15(a0)      //    {127...64}
    sdr     t3, 8(a0)

    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  Packed unsigned Multiplication Hi/Lo(2x16bit) = 8x16bit
//
//  void vmp_pmulhluH( uint16 * const phHi, uint16 * const phLo, const uint16 * const phA, const uint16 * const phB )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_pmulhluHMMI
vmp_pmulhluHMMI:

ASSERT_PTR16(a2)
ASSERT_PTR16(a3)
ASSERT_PTR16(a1)
ASSERT_PTR16(a0)

    lq      t1, 0(a2)       // phA
    lq      t2, 0(a3)       // phB

    //  Expand unsigned 16bit into 32bit
    pinteh  t0,zero,t1      // {0 A6 0 A4 0 A2 0 A0}
    pinteh  t3,zero,t2      // {0 B6 0 B4 0 B2 0 B0}
    psrlw   t1,t1,16        // {0 A7 0 A5 0 A3 0 A1}
    psrlw   t2,t2,16        // {0 B7 0 B5 0 B3 0 B1}

    //  32bit unsigned multiplication
    pmultuw t4, t0, t3      // {A4*B4  A0*B0} even
    pexeh   t0, t0          // {0 A4 0 A6 0 A0 0 A2}
    pexeh   t3, t3          // {0 B4 0 B6 0 B0 0 B2}
    
    pmultuw t5, t1, t2      // {A5*B5  A1*B1} odd
    pexeh   t1, t1          // {0 A5 0 A7 0 A1 0 A3}
    pexeh   t2, t2          // {0 B5 0 B7 0 B1 0 B3}

    pmultuw t0, t0, t3      // {A6*B6 A2*B2}
    pmultuw t1, t1, t2      // {A7*B7 A3*B3}

    //  Repackage 32bit into upper 16bit
    ppacw   t0, t0, t4      // {D6 D2 D4 D0}
    ppacw   t1, t1, t5      // {D7 D3 D5 D1}

    psrlw   t2,t0,16        // Shift upper 16 Bits into position
    psrlw   t3,t1,16

    pexcw   t0,t0           // {D6 D4 D2 D0}
    pexcw   t1,t1           // {D7 D5 D3 D1}
    pexcw   t2,t2           // {D6 D4 D2 D0}
    pexcw   t3,t3           // {D7 D5 D3 D1}

    pinteh  t0,t1,t0        // {D7 D6 D5 D4 D3 D2 D1 D0}
    pinteh  t1,t3,t2        // {D7 D6 D5 D4 D3 D2 D1 D0}

    sq      t0, 0(a1)       // phLo
    sq      t1, 0(a0)       // phHi

    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_pmulhluHMMI0
vmp_pmulhluHMMI0:

ASSERT_PTR4(a2)
ASSERT_PTR4(a3)
ASSERT_PTR4(a1)
ASSERT_PTR4(a0)

    ldl     t1, 7(a2)       // phA  {63...0}
    ldr     t1, 0(a2)
    ldl     t2, 15(a2)      //     {127...64}
    ldr     t2, 8(a2)

    ldl     t3, 7(a3)       // phB  {63...0}
    ldr     t3, 0(a3)
    ldl     t4, 15(a3)      //     {127...64}
    ldr     t4, 8(a3)

    pcpyld  t1,t2,t1        // A  {127...0}
    pcpyld  t2,t4,t3        // B  {127...0}
//  ________________

    //  Expand unsigned 16bit into 32bit
    pinteh  t0,zero,t1      // {0 A6 0 A4 0 A2 0 A0}
    pinteh  t3,zero,t2      // {0 B6 0 B4 0 B2 0 B0}
    psrlw   t1,t1,16        // {0 A7 0 A5 0 A3 0 A1}
    psrlw   t2,t2,16        // {0 B7 0 B5 0 B3 0 B1}

    //  32bit unsigned multiplication
    pmultuw t4, t0, t3      // {A4*B4  A0*B0} even
    pexeh   t0, t0          // {0 A4 0 A6 0 A0 0 A2}
    pexeh   t3, t3          // {0 B4 0 B6 0 B0 0 B2}
    
    pmultuw t5, t1, t2      // {A5*B5  A1*B1} odd
    pexeh   t1, t1          // {0 A5 0 A7 0 A1 0 A3}
    pexeh   t2, t2          // {0 B5 0 B7 0 B1 0 B3}

    pmultuw t0, t0, t3      // {A6*B6 A2*B2}
    pmultuw t1, t1, t2      // {A7*B7 A3*B3}

    //  Repackage 32bit into upper 16bit
    ppacw   t0, t0, t4      // {D6 D2 D4 D0}
    ppacw   t1, t1, t5      // {D7 D3 D5 D1}

    psrlw   t2,t0,16        // Shift upper 16 Bits into position
    psrlw   t3,t1,16

    pexcw   t0,t0           // {D6 D4 D2 D0}
    pexcw   t1,t1           // {D7 D5 D3 D1}
    pexcw   t2,t2           // {D6 D4 D2 D0}
    pexcw   t3,t3           // {D7 D5 D3 D1}

    pinteh  t0,t1,t0        // {D7 D6 D5 D4 D3 D2 D1 D0}
    pinteh  t1,t3,t2        // {D7 D6 D5 D4 D3 D2 D1 D0}

//  ________________
    pcpyud  t2,t0,t0        // D  {63...0  127...64} phLo
    pcpyud  t3,t1,t1        // D  {63...0  127...64} phHi

    sdl     t0, 7(a1)       // phLo {63...0}
    sdr     t0, 0(a1)
    sdl     t2, 15(a1)      //    {127...64}
    sdr     t2, 8(a1)

    sdl     t1, 7(a0)       // phHi {63...0}
    sdr     t1, 0(a0)
    sdl     t3, 15(a0)      //    {127...64}
    sdr     t3, 8(a0)

    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//*************************************************************************************************



//_________________________________________________________________________________________________
//=================================================================================================
//  Packed unsigned Multiplication 16bit Even = 16x8bit
//
//  void vmp_pmullueB( uint16 * const phD, const uint8 * const pbA, const uint8 * const pbB )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_pmullueBMMI
vmp_pmullueBMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // pbA
    lq      t3, 0(a2)       // pbB

        //  Convert 16x8bit to 2x 8x16bit
    pextlb  t0, t1, t1      // {A7  A7  ... A0 A0}
    pextub  t1, t1, t1      // {A15 A15 ... A8 A8}
    pextlb  t2, t3, t3      // {B7  B7  ... B0 B0}
    pextub  t3, t3, t3      // {B15 B15 ... B8 B8}

    psrlh   t0,t0,8         // 0 {A7...A0}
    psrlh   t1,t1,8
    psrlh   t2,t2,8
    psrlh   t3,t3,8

        //  2x 16bit multiplies
    pmulth  t0, t0, t2      // {D6  D4  D2  D0}
    pmulth  t1, t1, t3      // {D14 D12 D10 D8}

    ppach   t0,t1,t0        // {D15 ... D0}

    sq      t0, 0(a0)       // pbD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_pmullueBMMI0
vmp_pmullueBMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // pbA  {63...0}
    ldr     t1, 0(a1)
    ldl     t2, 15(a1)      //     {127...64}
    ldr     t2, 8(a1)

    ldl     t3, 7(a2)       // pbB  {63...0}
    ldr     t3, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t2,t1        // A  {127...0}
    pcpyld  t3,t4,t3        // B  {127...0}
//  ________________

        //  Convert 16x8bit to 2x 8x16bit
    pextlb  t0, t1, t1      // {A7  A7  ... A0 A0}
    pextub  t1, t1, t1      // {A15 A15 ... A8 A8}
    pextlb  t2, t3, t3      // {B7  B7  ... B0 B0}
    pextub  t3, t3, t3      // {B15 B15 ... B8 B8}

    psrlh   t0,t0,8         // 0 {A7...A0}
    psrlh   t1,t1,8
    psrlh   t2,t2,8
    psrlh   t3,t3,8

        //  2x 16bit multiplies
    pmulth  t0, t0, t2      // {D6  D4  D2  D0}
    pmulth  t1, t1, t3      // {D14 D12 D10 D8}

    ppach   t0,t1,t0        // {D15 ... D0}

//  ________________
    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // pbD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  Packed unsigned Multiplication 16bit Odd = 16x8bit
//
//  void vmp_pmulluoB( uint16 * const phD, const uint8 * const pbA, const uint8 * const pbB )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_pmulluoBMMI
vmp_pmulluoBMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // pbA
    lq      t3, 0(a2)       // pbB

        //  Convert 16x8bit to 2x 8x16bit
    pextlb  t0, t1, t1      // {A7  A7  ... A0 A0}
    pextub  t1, t1, t1      // {A15 A15 ... A8 A8}
    pextlb  t2, t3, t3      // {B7  B7  ... B0 B0}
    pextub  t3, t3, t3      // {B15 B15 ... B8 B8}

    psrlh   t0,t0,8         // 0 {A7...A0}
    psrlh   t1,t1,8
    psrlh   t2,t2,8
    psrlh   t3,t3,8

        //  2x 16bit multiplies
    pmulth  t0, t0, t2      // {D6  D4  D2  D0}
                            // HI  = {D7 D6 D3 D2}
                            // LO  = {D5 D4 D1 D0}
    pmfhl.uw    t0          //       {D7 D5 D3 D1}

    pmulth  t1, t1, t3      // {D14 D12 D10 D8}
                            // HI  = {D15 D14 D11 D10}
                            // LO  = {D13 D12 D9  D8}
    pmfhl.uw    t1          //       {D15 D13 D11 D9}

    ppach   t0,t1,t0        // {D15 ... D0}

    sq      t0, 0(a0)       // pbD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_pmulluoBMMI0
vmp_pmulluoBMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // pbA  {63...0}
    ldr     t1, 0(a1)
    ldl     t2, 15(a1)      //     {127...64}
    ldr     t2, 8(a1)

    ldl     t3, 7(a2)       // pbB  {63...0}
    ldr     t3, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t2,t1        // A  {127...0}
    pcpyld  t3,t4,t3        // B  {127...0}
//  ________________

        //  Convert 16x8bit to 2x 8x16bit
    pextlb  t0, t1, t1      // {A7  A7  ... A0 A0}
    pextub  t1, t1, t1      // {A15 A15 ... A8 A8}
    pextlb  t2, t3, t3      // {B7  B7  ... B0 B0}
    pextub  t3, t3, t3      // {B15 B15 ... B8 B8}

    psrlh   t0,t0,8         // 0 {A7...A0}
    psrlh   t1,t1,8
    psrlh   t2,t2,8
    psrlh   t3,t3,8

        //  2x 16bit multiplies
    pmulth  t0, t0, t2      // {D6  D4  D2  D0}
                            // HI  = {D7 D6 D3 D2}
                            // LO  = {D5 D4 D1 D0}
    pmfhl.uw    t0          //       {D7 D5 D3 D1}

    pmulth  t1, t1, t3      // {D14 D12 D10 D8}
                            // HI  = {D15 D14 D11 D10}
                            // LO  = {D13 D12 D9  D8}
    pmfhl.uw    t1          //       {D15 D13 D11 D9}

    ppach   t0,t1,t0        // {D15 ... D0}

//  ________________
    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // pbD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


//_________________________________________________________________________________________________
//=================================================================================================
//  Packed signed Multiplication 16bit Even = 16x8bit
//
//  void vmp_pmulleB( int16 * const phD, const int8 * const pbA, const int8 * const pbB )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_pmulleBMMI
vmp_pmulleBMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // pbA
    lq      t3, 0(a2)       // pbB

        //  Convert 16x8bit to 2x 8x16bit
    pextlb  t0, t1, t1      // {A7  A7  ... A0 A0}
    pextub  t1, t1, t1      // {A15 A15 ... A8 A8}
    pextlb  t2, t3, t3      // {B7  B7  ... B0 B0}
    pextub  t3, t3, t3      // {B15 B15 ... B8 B8}

    psrah   t0,t0,8         // +- {A7...A0}
    psrah   t1,t1,8
    psrah   t2,t2,8
    psrah   t3,t3,8

        //  2x 16bit multiplies
    pmulth  t0, t0, t2      // {D6  D4  D2  D0}
    pmulth  t1, t1, t3      // {D14 D12 D10 D8}

    ppach   t0,t1,t0        // {D15 ... D0}

    sq      t0, 0(a0)       // phD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_pmulleBMMI0
vmp_pmulleBMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // pbA  {63...0}
    ldr     t1, 0(a1)
    ldl     t2, 15(a1)      //     {127...64}
    ldr     t2, 8(a1)

    ldl     t3, 7(a2)       // pbB  {63...0}
    ldr     t3, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t2,t1        // A  {127...0}
    pcpyld  t3,t4,t3        // B  {127...0}
//  ________________

        //  Convert 16x8bit to 2x 8x16bit
    pextlb  t0, t1, t1      // {A7  A7  ... A0 A0}
    pextub  t1, t1, t1      // {A15 A15 ... A8 A8}
    pextlb  t2, t3, t3      // {B7  B7  ... B0 B0}
    pextub  t3, t3, t3      // {B15 B15 ... B8 B8}

    psrah   t0,t0,8         // +- {A7...A0}
    psrah   t1,t1,8
    psrah   t2,t2,8
    psrah   t3,t3,8

        //  2x 16bit multiplies
    pmulth  t0, t0, t2      // {D6  D4  D2  D0}
    pmulth  t1, t1, t3      // {D14 D12 D10 D8}

    ppach   t0,t1,t0        // {D15 ... D0}

//  ________________
    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // phD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


//_________________________________________________________________________________________________
//=================================================================================================
//  Packed signed Multiplication 16bit Odd = 16x8bit
//
//  void vmp_pmulloB( int16 * const phD, const int8 * const pbA, const int8 * const pbB )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_pmulloBMMI
vmp_pmulloBMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // pbA
    lq      t3, 0(a2)       // pbB

        //  Convert 16x8bit to 2x 8x16bit
    pextlb  t0, t1, t1      // {A7  A7  ... A0 A0}
    pextub  t1, t1, t1      // {A15 A15 ... A8 A8}
    pextlb  t2, t3, t3      // {B7  B7  ... B0 B0}
    pextub  t3, t3, t3      // {B15 B15 ... B8 B8}

    psrah   t0,t0,8         // +- {A7...A0}
    psrah   t1,t1,8
    psrah   t2,t2,8
    psrah   t3,t3,8

        //  2x 16bit multiplies
    pmulth  t0, t0, t2      // {D6  D4  D2  D0}
                            // HI  = {D7 D6 D3 D2}
                            // LO  = {D5 D4 D1 D0}
    pmfhl.uw    t0          //       {D7 D5 D3 D1}

    pmulth  t1, t1, t3      // {D14 D12 D10 D8}
                            // HI  = {D15 D14 D11 D10}
                            // LO  = {D13 D12 D9  D8}
    pmfhl.uw    t1          //       {D15 D13 D11 D9}

    ppach   t0,t1,t0        // {D15 ... D0}

    sq      t0, 0(a0)       // phD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_pmulloBMMI0
vmp_pmulloBMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // pbA  {63...0}
    ldr     t1, 0(a1)
    ldl     t2, 15(a1)      //     {127...64}
    ldr     t2, 8(a1)

    ldl     t3, 7(a2)       // pbB  {63...0}
    ldr     t3, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t2,t1        // A  {127...0}
    pcpyld  t3,t4,t3        // B  {127...0}
//  ________________

        //  Convert 16x8bit to 2x 8x16bit
    pextlb  t0, t1, t1      // {A7  A7  ... A0 A0}
    pextub  t1, t1, t1      // {A15 A15 ... A8 A8}
    pextlb  t2, t3, t3      // {B7  B7  ... B0 B0}
    pextub  t3, t3, t3      // {B15 B15 ... B8 B8}

    psrah   t0,t0,8         // +- {A7...A0}
    psrah   t1,t1,8
    psrah   t2,t2,8
    psrah   t3,t3,8

        //  2x 16bit multiplies
    pmulth  t0, t0, t2      // {D6  D4  D2  D0}
                            // HI  = {D7 D6 D3 D2}
                            // LO  = {D5 D4 D1 D0}
    pmfhl.uw    t0          //       {D7 D5 D3 D1}

    pmulth  t1, t1, t3      // {D14 D12 D10 D8}
                            // HI  = {D15 D14 D11 D10}
                            // LO  = {D13 D12 D9  D8}
    pmfhl.uw    t1          //       {D15 D13 D11 D9}

    ppach   t0,t1,t0        // {D15 ... D0}

//  ________________
    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // phD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//*************************************************************************************************



//_________________________________________________________________________________________________
//=================================================================================================
//  Packed unsigned Multiplication 32bit Even = 8x16bit
//
//  void vmp_pmullueH( uint32 * const pwD, const uint16 * const phA, const uint16 * const phB )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_pmullueHMMI
vmp_pmullueHMMI:

ASSERT_PTR16(a2)
ASSERT_PTR16(a1)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // phA
    lq      t2, 0(a2)       // phB

    //  Expand unsigned 16bit into 32bit
    pinteh  t0,zero,t1      // {0 A6 0 A4 0 A2 0 A0}
    pinteh  t3,zero,t2      // {0 B6 0 B4 0 B2 0 B0}

    //  32bit unsigned multiplication
    pmultuw t4, t0, t3      // {A4*B4  A0*B0}
    pexeh   t0, t0          // {0 A4 0 A6 0 A0 0 A2}
    pexeh   t3, t3          // {0 B4 0 B6 0 B0 0 B2}
    
    pmultuw t0, t0, t3      // {A6*B6 A2*B2}

    //  Repackage 32bit
    ppacw   t0, t0, t4      // {D6 D2 D4 D0}
    pexcw   t0,t0           // {D6 D4 D2 D0}

    sq      t0, 0(a0)       // pwD

    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_pmullueHMMI0
vmp_pmullueHMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // phA  {63...0}
    ldr     t1, 0(a1)
    ldl     t2, 15(a1)      //     {127...64}
    ldr     t2, 8(a1)

    ldl     t3, 7(a2)       // phB  {63...0}
    ldr     t3, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t2,t1        // A  {127...0}
    pcpyld  t2,t4,t3        // B  {127...0}
//  ________________

    //  Expand unsigned 16bit into 32bit
    pinteh  t0,zero,t1      // {0 A6 0 A4 0 A2 0 A0}
    pinteh  t3,zero,t2      // {0 B6 0 B4 0 B2 0 B0}

    //  32bit unsigned multiplication
    pmultuw t4, t0, t3      // {A4*B4  A0*B0}
    pexeh   t0, t0          // {0 A4 0 A6 0 A0 0 A2}
    pexeh   t3, t3          // {0 B4 0 B6 0 B0 0 B2}
    
    pmultuw t0, t0, t3      // {A6*B6 A2*B2}

    //  Repackage 32bit
    ppacw   t0, t0, t4      // {D6 D2 D4 D0}
    pexcw   t0,t0           // {D6 D4 D2 D0}

//  ________________
    pcpyud  t1,t0,t0        // D  {63...0  127...64} pwD

    sdl     t0, 7(a0)       // pwD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)

    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  Packed unsigned Multiplication 32bit Odd = 8x16bit
//
//  void vmp_pmulluoH( uint32 * const pwD, const uint16 * const phA, const uint16 * const phB )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_pmulluoHMMI
vmp_pmulluoHMMI:

ASSERT_PTR16(a2)
ASSERT_PTR16(a1)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // phA
    lq      t2, 0(a2)       // phB

    //  Expand unsigned 16bit into 32bit
    psrlw   t1,t1,16        // {0 A7 0 A5 0 A3 0 A1}
    psrlw   t2,t2,16        // {0 B7 0 B5 0 B3 0 B1}

    //  32bit unsigned multiplication
    pmultuw t0, t1, t2      // {A5*B5 A1*B1}
    pexeh   t1, t1          // {0 A5 0 A7 0 A1 0 A3}
    pexeh   t2, t2          // {0 B5 0 B7 0 B1 0 B3}
    pmultuw t1, t1, t2      // {A7*B7 A3*B3}

    //  Repackage 32bit
    ppacw   t0, t1, t0      // {D7 D3 D5 D1}
    pexcw   t0, t0          // {D7 D5 D3 D1}

    sq      t0, 0(a0)       // pwD

    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_pmulluoHMMI0
vmp_pmulluoHMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // phA  {63...0}
    ldr     t1, 0(a1)
    ldl     t2, 15(a1)      //     {127...64}
    ldr     t2, 8(a1)

    ldl     t3, 7(a2)       // phB  {63...0}
    ldr     t3, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t2,t1        // A  {127...0}
    pcpyld  t2,t4,t3        // B  {127...0}
//  ________________

    //  Expand unsigned 16bit into 32bit
    psrlw   t1,t1,16        // {0 A7 0 A5 0 A3 0 A1}
    psrlw   t2,t2,16        // {0 B7 0 B5 0 B3 0 B1}

    //  32bit unsigned multiplication
    pmultuw t0, t1, t2      // {A5*B5 A1*B1}
    pexeh   t1, t1          // {0 A5 0 A7 0 A1 0 A3}
    pexeh   t2, t2          // {0 B5 0 B7 0 B1 0 B3}
    pmultuw t1, t1, t2      // {A7*B7 A3*B3}

    //  Repackage 32bit
    ppacw   t0, t1, t0      // {D7 D3 D5 D1}
    pexcw   t0, t0          // {D7 D5 D3 D1}

//  ________________
    pcpyud  t1,t0,t0        // D  {63...0  127...64} pwD

    sdl     t0, 7(a0)       // pwD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)

    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  Packed signed Multiplication 32bit Even = 8x16bit
//
//  void vmp_pmulleH( int32 * const pwD, const int16 * const phA, const int16 * const phB )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_pmulleHMMI
vmp_pmulleHMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // phA
    lq      t2, 0(a2)       // phB
    LDELAY                  // nop - Load Delay Slot

    pmulth  t0, t1, t2

    sq      t0, 0(a0)       // pwD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_pmulleHMMI0
vmp_pmulleHMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // phA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

    ldl     t2, 7(a2)       // phB  {63...0}
    ldr     t2, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t3,t1        // A  {127...0}
    pcpyld  t2,t4,t2        // B  {127...0}
//  ________________

    pmulth  t0, t1, t2      // D  {127...0}

//  ________________
    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // pwD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  Packed signed Multiplication 32bit Odd = 8x16bit
//
//  void vmp_pmulloH( int32 * const pwD, const int16 * const phA, const int16 * const phB )
//
//  *** MMI version *** 128bit aligned

    .align 3
    .global vmp_pmulloHMMI
vmp_pmulloHMMI:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    lq      t1, 0(a1)       // phA
    lq      t2, 0(a2)       // phB
    LDELAY                  // nop - Load Delay Slot

    pmulth  t0, t1, t2      // Dst = {D6 D4 D2 D0}
                            // HI  = {D7 D6 D3 D2}
                            // LO  = {D5 D4 D1 D0}
    pmfhl.uw    t0          //       {D7 D5 D3 D1}

    sq      t0, 0(a0)       // pwD
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


//_________________________________________________________________________________________________
//  *** MMI version *** 128bit unaligned

    .align 3
    .global vmp_pmulloHMMI0
vmp_pmulloHMMI0:

ASSERT_PTR4(a1)
ASSERT_PTR4(a2)
ASSERT_PTR4(a0)

    ldl     t1, 7(a1)       // phA  {63...0}
    ldr     t1, 0(a1)
    ldl     t3, 15(a1)      //     {127...64}
    ldr     t3, 8(a1)

    ldl     t2, 7(a2)       // phB  {63...0}
    ldr     t2, 0(a2)
    ldl     t4, 15(a2)      //     {127...64}
    ldr     t4, 8(a2)

    pcpyld  t1,t3,t1        // A  {127...0}
    pcpyld  t2,t4,t2        // B  {127...0}
//  ________________

    pmulth  t0, t1, t2      // Dst = {D6 D4 D2 D0}
                            // HI  = {D7 D6 D3 D2}
                            // LO  = {D5 D4 D1 D0}
    pmfhl.uw    t0          //       {D7 D5 D3 D1}

//  ________________
    pcpyud  t1,t0,t0        // D  {63...0  127...64}

    sdl     t0, 7(a0)       // pwD {63...0}
    sdr     t0, 0(a0)
    sdl     t1, 15(a0)      //    {127...64}
    sdr     t1, 8(a0)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot













