
//
//  Bench.c - (Quad) Vector 3D Multiplication & Division
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 5 Oct 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//

#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries

#include <stdio.h>

    //  Book Development Library Modules

#include "CpuAsm.h"                                     // CPU Module
#include "vmp3D.h"
#include "QVmd3D.h"                                     // (Quad) Vector Multiplication & Division
#include "QVmd3DEx.h"                                   // (Quad)  "   " (Private)



typedef enum
{
    iTEST_DOT_PRODUCT,
    iTEST_CROSS_PRODUCT,
    iTEST_VEC_MUL,
    iTEST_VEC_MADD,
    iTEST_VEC_MSUB,
    iTEST_VEC_SCALE,
    iTEST_VEC_DIV,
    iTEST_VEC_DIV_FAST
} iTEST;



//
//  Test 
//

static void testQVec( bool bAlign, iTEST iFunc )
{
    vmp3DQVector *pvA, *pvB, *pvC, *pvTmp, *pvTmp2;
    int w, x, y, z, dif;
    float f, f2, fPrecision, fScalar;
    float buf[4*4  +3];         // enough space + 12 bytes (3 floats) extra
    bool bF;

        // Force proper alignment (Crude just for this demo)

    bF = false;

    if (bAlign)
      {
        pvA    = (vmp3DQVector*) ALIGN16((int)(buf));
        pvB    = (vmp3DQVector*) ALIGN16((int)(pvA+1));
        pvTmp  = (vmp3DQVector*) ALIGN16((int)(pvB+1));
        pvTmp2 = (vmp3DQVector*) ALIGN16((int)(pvTmp+1));
      }
    else
      {
        pvA    = (vmp3DQVector*) buf;
        if (!(0x000f & (int)pvA))
          {
            pvA = (vmp3DQVector*) (buf+1);          // Make sure misaligned!
          }

        pvB    = pvA + 1;
        pvTmp  = pvB + 1;
        pvTmp2 = pvTmp + 1;
      }

    fPrecision = SINGLE_PRECISION;
    pvC = pvA;
    dif = 4;
    
    for ( w = -dif; w <= dif; w++ )
      {
        pvA->w = 0.123f *(float)w;
        pvB->z = 0.123f *(float)w;
        
        for ( z = -dif; z <= dif; z++ )
          {
            pvA->z = 0.123f *(float)z;
            pvB->y = 0.123f *(float)z;

            for ( y = -dif; y <= dif; y++ )
              {
                pvA->y = 0.123f *(float)y;
                pvB->x = 0.123f *(float)y;

                for ( x = -dif; x <= dif; x++ )
                  { 
                    pvB->w = 0.123f *(float)x;
                    pvA->x = 0.123f *(float)x;

                    fScalar = pvA->x;

                    switch(iFunc)
                      {
                        case iTEST_DOT_PRODUCT:
                            fPrecision = SINGLE_PRECISION;
                            bF = true;
                            vmp_QDotProductGeneric( &f2, pvA, pvB );
                            if (bAlign)
                                vmp_QDotProduct( &f, pvA, pvB );        // Aligned
                            else
                                vmp_QDotProduct0( &f, pvA, pvB );
                            break;

                        case iTEST_CROSS_PRODUCT:
                            fPrecision = SINGLE_PRECISION;
                            bF = false;
                            vmp_QCrossProductGeneric( pvTmp2, pvA, pvB );
                            if (bAlign)
                                vmp_QCrossProduct( pvTmp, pvA, pvB );   // Aligned
                            else
                                vmp_QCrossProduct0( pvTmp, pvA, pvB );
                            break;

                        case iTEST_VEC_MUL:
                            fPrecision = SINGLE_PRECISION;
                            bF = false;
                            vmp_QVecMulGeneric( pvTmp2, pvA, pvB );
                            if (bAlign)
                                vmp_QVecMul( pvTmp, pvA, pvB );         // Aligned
                            else
                                vmp_QVecMul0( pvTmp, pvA, pvB );
                            break;

                        case iTEST_VEC_MADD:
                            fPrecision = SINGLE_PRECISION;
                            bF = false;
                            vmp_QVecMAddGeneric( pvTmp2, pvA, pvB, pvC );
                            if (bAlign)
                                vmp_QVecMAdd( pvTmp, pvA, pvB, pvC );           // Aligned
                            else
                                vmp_QVecMAdd0( pvTmp, pvA, pvB, pvC );
                            break;

                        case iTEST_VEC_MSUB:
                            fPrecision = SINGLE_PRECISION;
                            bF = false;
                            vmp_QVecMSubGeneric( pvTmp2, pvA, pvB, pvC );
                            if (bAlign)
                                vmp_QVecMSub( pvTmp, pvA, pvB, pvC );           // Aligned
                            else
                                vmp_QVecMSub0( pvTmp, pvA, pvB, pvC );
                            break;

                        case iTEST_VEC_SCALE:
                            fPrecision = SINGLE_PRECISION;
                            bF = false;
                            vmp_QVecScaleGeneric( pvTmp2, pvA, fScalar );
                            if (bAlign)
                                vmp_QVecScale( pvTmp, pvA, fScalar );   // Aligned
                            else
                                vmp_QVecScale0( pvTmp, pvA, fScalar );
                            break;

                        case iTEST_VEC_DIV:
                            fPrecision = SINGLE_PRECISION;
                            bF = false;
                            if (0.0==fScalar) continue;

                            vmp_QVecDivGeneric( pvTmp2, pvA, fScalar );
                            if (bAlign)
                                vmp_QVecDiv( pvTmp, pvA, fScalar );         // Aligned
                            else
                                vmp_QVecDiv0( pvTmp, pvA, fScalar );
                            break;

                        case iTEST_VEC_DIV_FAST:                        // Lower Precision
                            fPrecision = FAST_PRECISION;
                            bF = false;
                            if (0.0==fScalar) continue;

                            vmp_QVecDivGeneric( pvTmp2, pvA, fScalar );
                            if (bAlign)
                                vmp_QVecDiv( pvTmp, pvA, fScalar );         // Aligned
                            else
                                vmp_QVecDiv0( pvTmp, pvA, fScalar );
                            break;
                      }

                    if (bF)
                      {
                        if (!vmp_IsFEqual(f, f2, fPrecision))
                          {
                            printf( " vA.x:%f vA.y:%f vA.z:%f\n",
                                pvA->x, pvA->y, pvA->z );
                            printf( " vB.x:%f vB.y:%f vB.z:%f\n", 
                                pvB->x, pvB->y, pvB->z );
                            printf( "    %f Good vs Bad! %f  *** MISMATCH!!! ***\n\n", f2, f );
                          }
                      }
                    else
                      {
                        if (!vmp_IsQVEqual(pvTmp, pvTmp2, fPrecision)) 
                          {
                            printf( " vA.x:%f vA.y:%f vA.z:%f vA.w:%f\n",
                                     pvA->x, pvA->y, pvA->z, pvA->w );
                            printf( " vB.x:%f vB.y:%f vB.z:%f vB.w:%f\n",
                                     pvB->x, pvB->y, pvB->z, pvB->w );
                            printf( "    x:%f y:%f z:%f w:%f  Good vs Bad!!!=\n",
                                     pvTmp2->x, pvTmp2->y, pvTmp2->z, pvTmp2->w );
                            printf( "    x:%f y:%f z:%f w:%f   *** MISMATCH!!! ***\n\n",
                                     pvTmp->x, pvTmp->y, pvTmp->z, pvTmp->w );
                          }
                      }
                  }
              }
          }
      }
}



int main( int argc, char *argv[] )
{
    CpuInfo cinfo;
    char szBuf[ CPU_SZBUF_MAX ];

    CpuDetect( &cinfo );                                // Detect CPU
    
    printf( "\n(Quad) Vector Scaling and Multiplication Test\n\n" );
    printf( "%s\n", CpuInfoStr( szBuf, &cinfo ) );
    
    QVmd3DSetup( &cinfo );                              // Setup Quad (Vector) stuff
    
    printf( "Testing (Quad) Scaling\n" );
    testQVec( false, iTEST_VEC_SCALE );
    
    printf( "Testing (Quad) Scaling (Aligned)\n" );
    testQVec( true,  iTEST_VEC_SCALE );

    printf( "Testing (Quad) Vector Multiplication\n" );
    testQVec( false, iTEST_VEC_MUL );

    printf( "Testing (Quad) Vector Multiplication (Aligned)\n" );
    testQVec( true,  iTEST_VEC_MUL );

    printf( "Testing (Quad) Vector Multiplication-Add\n" );
    testQVec( false, iTEST_VEC_MADD );

    printf( "Testing (Quad) Vector Multiplication-Add (Aligned)\n" );
    testQVec( true,  iTEST_VEC_MADD );

    printf( "Testing (Quad) Vector Multiplication-Sub\n" );
    testQVec( false, iTEST_VEC_MSUB );

    printf( "Testing (Quad) Vector Multiplication-Sub (Aligned)\n" );
    testQVec( true,  iTEST_VEC_MSUB );

    printf( "Testing (Quad) Vector Division\n" );
    testQVec( false, iTEST_VEC_DIV );

    printf( "Testing (Quad) Vector Division (Aligned)\n" );
    testQVec( true,  iTEST_VEC_DIV );

    printf( "Testing (Quad) Vector Division **FAST**\n" );
    testQVec( false, iTEST_VEC_DIV_FAST );

    printf( "Testing (Quad) Vector Division (Aligned) **FAST**\n" );
    testQVec( true,  iTEST_VEC_DIV_FAST );

    printf( "Testing (Quad) Dot Product\n" );
    testQVec( false, iTEST_DOT_PRODUCT );

    printf( "Testing (Quad) Dot Product (Aligned)\n" );
    testQVec( true,  iTEST_DOT_PRODUCT );

    printf( "Testing (Quad) Cross Product\n" );
    testQVec( false, iTEST_CROSS_PRODUCT );

    printf( "Testing (Quad) Cross Product (Aligned)\n" );
    testQVec( true,  iTEST_CROSS_PRODUCT );

    printf( "Fin\n\n" );

    return 0;
}

