
//
//  QVmd3DMips.s - Packed (128bit MIPS) Quad Multiplication
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//      This module is designed for fast 3D math
//
//  Created - 3 Jul, 2002 - James Leiterman
//
//  Modifications
//
//
//  stab = 4
//

    .nolist
#include "Mips.inc"
#include "KariType.inc"
    .list

    .set    noreorder
    .set    nomips16

    .section .text                                                    


//_________________________________________________________________________________________________
//=================================================================================================
//  (Quad) Vector Product (Multiplication)  (MIPS V)
//
//  void vmp_QVecMul(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB)
//

    .align 3
    .global vmp_QVecMulMIPSV
vmp_QVecMulMIPSV:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    ldc1   $f4, 0(a1)     // {Ay Ax} Mem to PS_FP
    ldc1   $f6, 0(a2)     // {By Bx}
    ldc1   $f5, 8(a1)     // {Aw Az}
    ldc1   $f7, 8(a2)     // {Bw Bz}

    mul.ps $f4, $f4, $f6  // {AyBy AxBx}
    mul.ps $f5, $f5, $f7  // {AwBw AzBz}

    sdc1   $f4, 0(a0)     // {Dy Dx}
    sdc1   $f5, 8(a0)     // {Dw Dz}

    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  (Quad) Vector Product Sum (Multiplication-Add) (MIPS-V)
//
//  void vmp_QVecMAdd(vmp3DQVector * const pvD, const vmp3DQVector * const pvA,
//                   const vmp3DQVector * const pvB, const vmp3DQVector * const pvC)
//
    .align 3
    .global vmp_QVecMAddMIPSV
vmp_QVecMAddMIPSV:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a3)
ASSERT_PTR16(a0)

    ldc1   $f4, 0(a1)     // {Ay Ax} Mem to PS_FP
    ldc1   $f6, 0(a2)     // {By Bx}
    ldc1   $f5, 8(a1)     // {Aw Az}
    ldc1   $f7, 8(a2)     // {Bw Bz}
    ldc1   $f8, 0(a3)     // {Cy Cx}
    ldc1   $f9, 8(a3)     // {Cw Cz}
       //    D    C    A    B
    madd.ps $f4, $f8, $f4, $f6  // {AyBy+Cy AxBx+Cx}
    madd.ps $f5, $f9, $f5, $f7  // {AwBw+Cw AzBz+Cz}

    sdc1   $f4, 0(a0)     // {Dy Dx}
    sdc1   $f5, 8(a0)     // {Dw Dz}

    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  (Quad) Scale Vector w/Scalar  (MIPS V)
//
//
//  void vmp_QVecScale(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, float fScalar )
//

    .align 3
    .global vmp_QVecScaleMIPSV
vmp_QVecScaleMIPSV:

ASSERT_PTR16(a1)
ASSERT_PTR16(a0)

#if 0       // Green Hills 
            // fScalar in FP register argument $f12
    ldc1     $f4, 0(a1)    // {Ay Ax}
    ldc1     $f5, 8(a1)    // {Aw Az}

    pll.ps   $f8, $f12, $f12 // {   s    s} Replicate Scalar
#else       // fScalar in general purpose register
    mfc1     $f8, a2       // fScalar (FPU from CPU)
    ldc1     $f4, 0(a1)    // {Ay Ax}
    ldc1     $f5, 8(a1)    // {Aw Az}

    pll.ps   $f8, $f8, $f8 // {   s    s} Replicate Scalar
#endif

    mul.ps   $f4, $f4, $f8 // {Ay*s Ax*s}
    mul.ps   $f5, $f5, $f8 // {Aw*s Az*s}

    sdc1     $f4, 0(a0)    // {Dy Dx}
    sdc1     $f5, 8(a0)    // {Dw Dz}

    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  (Quad) Vector (Division)  (MIPS-3D)
//
//  void vmp_QVecDiv(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, float fScalar)
//

    .align 3
    .global vmp_QVecDivFastMIPS3D
vmp_QVecDivFastMIPS3D:

ASSERT_ZERO(a2)
ASSERT_PTR16(a1)
ASSERT_PTR16(a0)

#if 0       // Green Hills 
            // fScalar in FP register argument $f12
    ldc1     $f4, 0(a1)    // {Ay Ax}
    ldc1     $f5, 8(a1)    // {Aw Az}

    recip1.s $f12, $f12      // {1/s}

    pll.ps   $f8, $f12, $f12 // {1/s 1/s} Replicate Scalar
#else       // fScalar in general purpose register
    mfc1     $f8, a2       // fScalar (FPU from CPU)
    ldc1     $f4, 0(a1)    // {Ay Ax}
    ldc1     $f5, 8(a1)    // {Aw Az}

    recip1.s $f6, $f8      // {1/s}

    pll.ps   $f8, $f8, $f8 // {1/s 1/s} Replicate Scalar
#endif

    mul.ps   $f4, $f4, $f8 // {Ay/s Ax/s}
    mul.ps   $f5, $f5, $f8 // {Aw/s Az/s}

    sdc1     $f4, 0(a0)    // {Dy Dx}
    sdc1     $f5, 8(a0)    // {Dw Dz}

    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



    .align 3
    .global vmp_QVecDivMIPS3D
vmp_QVecDivMIPS3D:

ASSERT_ZERO(a2)
ASSERT_PTR16(a1)
ASSERT_PTR16(a0)

#if 0       // Green Hills 
            // fScalar in FP register argument $f12
    ldc1     $f4, 0(a1)    // {Ay Ax}
    ldc1     $f5, 8(a1)    // {Aw Az}

            // Note: Could use recip.s if MIPS-3D not available!
    recip1.s $f6, $f12      // {1/s}
    recip2.s $f7, $f6, $f12
#else       // fScalar in general purpose register
    mfc1     $f8, a2       // fScalar (FPU from CPU)
    ldc1     $f4, 0(a1)    // {Ay Ax}
    ldc1     $f5, 8(a1)    // {Aw Az}

    // Note: Could use recip.s if MIPS-3D not available!

    recip1.s $f6, $f8      // {1/s}
    recip2.s $f7, $f6, $f8
#endif

    madd.s   $f8, $f6, $f6, $f7 // 24bit {1/s}

    pll.ps   $f8, $f8, $f8 // {1/s 1/s} Replicate Scalar
    mul.ps   $f4, $f4, $f8 // {Ay*1/s Ax*1/s}
    mul.ps   $f5, $f5, $f8 // {Aw*1/s Az*1/s}

    sdc1     $f4, 0(a0)    // {Dy Dx}
    sdc1     $f5, 8(a0)    // {Dw Dz}

    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot
