
//
//  Bench.c - Vector 3D Multiplication & Division
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 5 Oct 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//

#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries

#include <stdio.h>

    //  Book Development Library Modules

#include "CpuAsm.h"                                     // CPU Module
#include "vmp3D.h"
#include "Vmd3D.h"                                      // Vector Multiplication & Division
#include "Vmd3DEx.h"                                    //  "   " (Private)



typedef enum
{
    iTEST_DOT_PRODUCT,
    iTEST_CROSS_PRODUCT,
    iTEST_VEC_MUL,
    iTEST_VEC_MADD,
    iTEST_VEC_MSUB,
    iTEST_VEC_SCALE,
    iTEST_VEC_DIV,
    iTEST_VEC_DIV_FAST
} iTEST;



//
//  Test 
//

static void testVec( bool bAlign, iTEST iFunc )
{
    vmp3DVector *pvA, *pvB, *pvC, *pvTmp, *pvTmp2;
    int x, y, z, dif;
    float f, f2, fPrecision, fScalar, fWTst;
    float buf[4*4  +3];         // enough space + 12 bytes (3 floats) extra
    bool bF;

        // Force proper alignment (Crude just for this demo)

    bF = false;

    if (bAlign)
      {
        pvA    = (vmp3DVector*) ALIGN16((int)(buf));
        pvB    = (vmp3DVector*) ALIGN16((int)(pvA+1));
        pvTmp  = (vmp3DVector*) ALIGN16((int)(pvB+1));
        pvTmp2 = (vmp3DVector*) ALIGN16((int)(pvTmp+1));
      }
    else
      {
        pvA    = (vmp3DVector*) buf;
        if (!(0x000f & (int)pvA))
          {
            pvA = (vmp3DVector*) (buf+1);           // Make sure misaligned!
          }

        pvB    = pvA + 1;
        pvTmp2 = pvB + 1;
        pvTmp  = pvTmp2 + 1;    // no tail, for .W testing
      }

    pvC = pvA;
    dif = 4;
    fPrecision = SINGLE_PRECISION;
    
    for ( z = -dif; z <= dif; z++ )
      {
        pvA->z = 0.123f *(float)z;
        pvB->y = 0.123f *(float)z;

        for ( y = -dif; y <= dif; y++ )
          {
            pvA->y = 0.123f *(float)y;
            pvB->x = 0.123f *(float)y;

            for ( x = -dif; x <= dif; x++ )
              { 
                pvA->x = 0.123f *(float)x;
                pvB->z = 0.123f *(float)y;
                fScalar = pvA->x;

                fWTst = ((vmp3DQVector *) pvTmp)->w = fScalar;  // Get untouchable value!

                switch(iFunc)
                  {
                    case iTEST_DOT_PRODUCT:
                        fPrecision = SINGLE_PRECISION;
                        bF = true;
                        vmp_DotProductGeneric( &f2, pvA, pvB );
                        if (bAlign)
                            vmp_DotProduct( &f, pvA, pvB );     // Aligned
                        else
                            vmp_DotProduct0( &f, pvA, pvB );
                        break;

                    case iTEST_CROSS_PRODUCT:
                        fPrecision = SINGLE_PRECISION;
                        bF = false;
                        vmp_CrossProductGeneric( pvTmp2, pvA, pvB );
                        if (bAlign)
                            vmp_CrossProduct( pvTmp, pvA, pvB );    // Aligned
                        else
                            vmp_CrossProduct0( pvTmp, pvA, pvB );
                        break;

                    case iTEST_VEC_MUL:
                        fPrecision = SINGLE_PRECISION;
                        bF = false;
                        vmp_VecMulGeneric( pvTmp2, pvA, pvB );
                        if (bAlign)
                            vmp_VecMul( pvTmp, pvA, pvB );          // Aligned
                        else
                            vmp_VecMul0( pvTmp, pvA, pvB );
                        break;

                    case iTEST_VEC_MADD:
                        fPrecision = SINGLE_PRECISION;
                        bF = false;
                        vmp_VecMAddGeneric( pvTmp2, pvA, pvB, pvC );
                        if (bAlign)
                            vmp_VecMAdd( pvTmp, pvA, pvB, pvC );            // Aligned
                        else
                            vmp_VecMAdd0( pvTmp, pvA, pvB, pvC );
                        break;

                    case iTEST_VEC_MSUB:
                        fPrecision = SINGLE_PRECISION;
                        bF = false;
                        vmp_VecMSubGeneric( pvTmp2, pvA, pvB, pvC );
                        if (bAlign)
                            vmp_VecMSub( pvTmp, pvA, pvB, pvC );            // Aligned
                        else
                            vmp_VecMSub0( pvTmp, pvA, pvB, pvC );
                        break;

                    case iTEST_VEC_SCALE:
                        fPrecision = SINGLE_PRECISION;
                        bF = false;
                        vmp_VecScaleGeneric( pvTmp2, pvA, fScalar );
                        if (bAlign)
                            vmp_VecScale( pvTmp, pvA, fScalar );    // Aligned
                        else
                            vmp_VecScale0( pvTmp, pvA, fScalar );
                        break;

                    case iTEST_VEC_DIV:
                        fPrecision = SINGLE_PRECISION;
                        bF = false;
                        if (0.0==fScalar) continue;

                        vmp_VecDivGeneric( pvTmp2, pvA, fScalar );
                        if (bAlign)
                            vmp_VecDiv( pvTmp, pvA, fScalar );          // Aligned
                        else
                            vmp_VecDiv0( pvTmp, pvA, fScalar );
                        break;

                    case iTEST_VEC_DIV_FAST:
                        fPrecision = FAST_PRECISION;
                        bF = false;
                        if (0.0==fScalar) continue;

                        vmp_VecDivGeneric( pvTmp2, pvA, fScalar );
                        if (bAlign)
                            vmp_VecDivFast( pvTmp, pvA, fScalar );          // Aligned
                        else
                            vmp_VecDivFast0( pvTmp, pvA, fScalar );
                        break;
                  }

                if (bF)
                  {
                    if (!vmp_IsFEqual(f, f2, fPrecision))
                      {
                        printf( " vA.x:%f vA.y:%f vA.z:%f\n",
                                pvA->x, pvA->y, pvA->z );
                        printf( " vB.x:%f vB.y:%f vB.z:%f\n", 
                                pvB->x, pvB->y, pvB->z );
                        printf( "    %f Good vs Bad! %f  *** MISMATCH!!! ***\n\n", f2, f );
                      }
                  }
                else
                  {
                    if (!vmp_IsVEqual(pvTmp, pvTmp2, fPrecision)) 
                      {
                        printf( " vA.x:%f vA.y:%f vA.z:%f\n",
                                 pvA->x, pvA->y, pvA->z );
                        printf( " vB.x:%f vB.y:%f vB.z:%f\n",
                                 pvB->x, pvB->y, pvB->z );
                        printf( "    x:%f y:%f z:%f  Good vs Bad!!!=\n",
                                 pvTmp2->x, pvTmp2->y, pvTmp2->z );
                        printf( "    x:%f y:%f z:%f   *** MISMATCH!!! ***\n\n",
                                 pvTmp->x, pvTmp->y, pvTmp->z );
                      }
                    
                        //  Check for '.W' contamination
                    if (!vmp_IsFEqual(fWTst, ((vmp3DQVector *) pvTmp)->w, fPrecision ))
                      {
                        printf( ".W is contaminated! should be: %f not %f\n", fWTst, ((vmp3DQVector *) pvTmp)->w );
                      }
                  }
              }
          }
      }
}



int main( int argc, char *argv[] )
{
    CpuInfo cinfo;
    char szBuf[ CPU_SZBUF_MAX ];

    CpuDetect( &cinfo );                                // Detect CPU

    printf( "\nVector Multiplication, Division, Scaling, Dot, and Cross Test\n\n" );
    printf( "%s\n", CpuInfoStr( szBuf, &cinfo ) );
    
    Vmd3DSetup( &cinfo );                               // Setup (Vector) stuff

    printf( "Testing Scaling\n" );
    testVec( false, iTEST_VEC_SCALE );
    
    printf( "Testing Scaling (Aligned)\n" );
    testVec( true,  iTEST_VEC_SCALE );

    printf( "Testing Vector Multiplication\n" );
    testVec( false, iTEST_VEC_MUL );

    printf( "Testing Vector Multiplication (Aligned)\n" );
    testVec( true,  iTEST_VEC_MUL );

    printf( "Testing Vector Multiplication-Add\n" );
    testVec( false, iTEST_VEC_MADD );

    printf( "Testing Vector Multiplication-Add (Aligned)\n" );
    testVec( true,  iTEST_VEC_MADD );

    printf( "Testing Vector Multiplication-Sub\n" );
    testVec( false, iTEST_VEC_MSUB );

    printf( "Testing Vector Multiplication-Sub (Aligned)\n" );
    testVec( true,  iTEST_VEC_MSUB );

    printf( "Testing Vector Division\n" );
    testVec( false, iTEST_VEC_DIV );

    printf( "Testing Vector Division (Aligned)\n" );
    testVec( true,  iTEST_VEC_DIV );

    printf( "Testing Vector Division *14bit FAST*)\n" );
    testVec( false, iTEST_VEC_DIV_FAST );

    printf( "Testing Vector Division (Aligned) *14bit FAST*\n" );
    testVec( true,  iTEST_VEC_DIV_FAST );

    printf( "Testing Dot Product\n" );
    testVec( false, iTEST_DOT_PRODUCT );

    printf( "Testing Dot Product (Aligned)\n" );
    testVec( true,  iTEST_DOT_PRODUCT );

    printf( "Testing Cross Product\n" );
    testVec( false, iTEST_CROSS_PRODUCT );

    printf( "Testing Cross Product (Aligned)\n" );
    testVec( true,  iTEST_CROSS_PRODUCT );

    printf( "Fin\n\n" );

    return 0;
}

