
//
//  Vmd3DMips.s - Packed (128bit MIPS) Multiplication
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//      This module is designed for fast 3D math
//
//  Created - 3 Jul, 2002 - James Leiterman
//
//  Modifications
//
//
//  stab = 4
//

    .nolist
#include "Mips.inc"
#include "KariType.inc"
    .list

    .set    noreorder
    .set    nomips16

    .section .text                                                    


//_________________________________________________________________________________________________
//=================================================================================================
//  Vector Product (Multiplication)  (MIPS V)
//
//  void vmp_VecMul(vmp3DVector * const pvD, const vmp3DVector * const pvA, const vmp3DVector * const pvB)
//

    .align 3
    .global vmp_VecMulMIPSV
vmp_VecMulMIPSV:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR16(a0)

    ldc1   $f4, 0(a1)     // {Ay Ax} Mem to PS_FP
    ldc1   $f6, 0(a2)     // {By Bx}
    lwc1   $f5, 8(a1)     // {   Az}
    lwc1   $f7, 8(a2)     // {   Bz}

    mul.ps $f4, $f4, $f6  // {AyBy AxBx}
    mul.s  $f5, $f5, $f7  // {     AzBz}

    sdc1   $f4, 0(a0)     // {Dy Dx}
    swc1   $f5, 8(a0)     // {   Dz}

    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//  Dot Product (MIPS V & MIPS-3D)
//
//  void vmp_DotProduct(float * const pfD, const vmp3DVector * const pvA, const vmp3DVector * const pvB)
//

    .align 3
    .global vmp_DotProductMIPS3D
vmp_DotProductMIPS3D:

ASSERT_PTR16(a1)
ASSERT_PTR16(a2)
ASSERT_PTR4(a0)

    ldc1    $f4, 0(a1)    // {Ay Ax} Mem to PS_FP
    ldc1    $f6, 0(a2)    // {By Bx}
    lwc1    $f5, 8(a1)    // {   Az}
    lwc1    $f7, 8(a2)    // {   Bz}

    mul.ps  $f4, $f4, $f6 // {AyBy AxBx}
    mul.s   $f5, $f5, $f7 // {     AzBz}
    addr.ps $f4, $f4, $f4 // {AyBy+AxBx AyBy+AxBx}
    add.s   $f5, $f5, $f4 // {     AzBz+AyBy+AxBx}

    swc1    $f5, 0(a0) 

    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot












