
//
//  Bench.c - (Scalar) Special Function (Float)
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 5 Oct 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//

#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include <stdio.h>

    //  Book Development Library Modules

#include "CpuAsm.h"                                     // CPU Module
#include "vmp3D.h"
#include "Fsf.h"
#include "FsfEx.h"



typedef enum
{
    iTEST_FLT_SQRT
} iTEST;



//
//  Test 
//

static void testVal( bool bAlign, iTEST iFunc, bool bFast )
{
    int x, dif;
    float f, f2, fPrecision, fA;

    dif = 4;

    fPrecision = (bFast) ? FAST_PRECISION : SINGLE_PRECISION;
    
    for ( x = -dif; x <= dif; x++ )
      { 
        fA = 0.123f *(float)x;

        switch(iFunc)
          {
            case iTEST_FLT_SQRT:
                if ( 0.0f > fA ) fA = -fA;

                vmp_FSqrtGeneric( &f2, fA );
                if ( bFast )
                  {
                    vmp_FSqrtFast( &f, fA );
                  }
                else
                  {
                    vmp_FSqrt( &f, fA );
                  }
                break;
          }

        if (!vmp_IsFEqual(f, f2, fPrecision))
          {
            printf( " %f vs %f  fA:%f  *** MISMATCH!!! ***\n", f2, f, fA );
          }
      }
}



int main( int argc, char *argv[] )
{
    CpuInfo cinfo;
    char szBuf[ CPU_SZBUF_MAX ];

    CpuDetect( &cinfo );                                // Detect CPU

    printf( "\nTesting (Scalar) Square Root\n\n" );
    printf( "%s\n", CpuInfoStr( szBuf, &cinfo ) );

    FsfSetup( &cinfo );                             // Setup (float) stuff

    printf( "Testing Square Root\n" );
    testVal( false, iTEST_FLT_SQRT, false );

    printf( "Testing Square Root *FAST (Low Precision)*\n" );
    testVal( false, iTEST_FLT_SQRT, true );

    printf( "Fin!\n\n" );

    return 0;
}

