
//
//  FsfMIPS.c - (Scalar) Special Function (Float)
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 3 Oct 2001 - James Leiterman
//
//  Last Modification
//
//  stab=4
//


#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include    <assert.h>
#include    <stdio.h>
#include    <math.h>

    //  Book Development Library Modules

#include    "CpuAsm.h"                                  // CPU
#include    "vmp3D.h"
#include    "vmp3DEx.h"
#include    "Fsf.h"                                     // (float) Math
#include    "FsfEx.h"


typedef float       sceVu0FVECTOR[4] __attribute__((aligned (16)));



//
//  (float) Square Root
//

void vmp_FSqrtVU( float * const pfD, float fA )
{
    sceVu0FVECTOR vD;
    
    ASSERT_PTR4(pfD);
    ASSERT_FNEG(fA);

    asm __volatile__("

    mfc1    $8,%1               #fA
    qmtc2    $8,vf4             #vf4=  xfer FP from EE to VU

    vsqrt Q,vf4x                # q=sqrt(fA)
    vwaitq      
    vaddq.x vf4x, vf0x, Q       #  =0+q

    sqc2    vf4,0x0(%0)         #pfD=

    ":
     : "r" (vD) , "f" (fA) 
     : "$8");
    
    *pfD = vD[0];
}



//
//  Setup Vec3D module
//
//  NOTE: This function sets up the function vectors so the best function
//      is attached to the vector for the processor running.
//

void FsfSetup( const CpuInfo * const pcinfo )
{
    ASSERT_PTR4(pcinfo);

    if ( CPUBITS_VU0 & pcinfo->nBits )
      {
        vmp_FSqrt =             vmp_FSqrtVU;
        vmp_FSqrtFast =         vmp_FSqrtVU;
      }

    else
      {
        vmp_FSqrt =             vmp_FSqrtGeneric;
        vmp_FSqrtFast =         vmp_FSqrtGeneric;
      }
}





