
//
//  Bench.c - Vector 3D Special Function
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 5 Oct 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//

#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include <stdio.h>

    //  Book Development Library Modules
    
#include "CpuAsm.h"                                     // CPU Module
#include "vmp3D.h"

#include "Vsf3D.h"
#include "Vsf3DEx.h"



typedef enum
{
    iTEST_MAGNITUDE,
    iTEST_MAGNITUDE_FAST,
    iTEST_NORMALIZE,
    iTEST_NORMALIZE_FAST,
    iTEST_VEC_SQRT,
    iTEST_VEC_SQRT_FAST,
    iTEST_MIN,
    iTEST_MAX
} iTEST;




//
//  Test 
//

static void testVec( bool bAlign, iTEST iFunc )
{
    vmp3DVector *pvA, *pvB, *pvTmp, *pvTmp2;
    int x, y, z, dif;
    float f, f2, fPrecision, fScalar, fWTst;
    float buf[4 *4  +3];            // enough space + 12 bytes (3 floats) extra
    bool bF;

        // Force proper alignment (Crude just for this demo)

    bF = false;

    if ( bAlign )
      {
        pvA    = (vmp3DVector*) ALIGN16((int)(buf));
        pvB    = (vmp3DVector*) ALIGN16((int)(pvA+1));
        pvTmp  = (vmp3DVector*) ALIGN16((int)(pvB+1));
        pvTmp2 = (vmp3DVector*) ALIGN16((int)(pvTmp+1));
      }
    else
      {
        pvA    = (vmp3DVector*) buf;
        if (!(0x000f & (int)pvA))
          {
            pvA = (vmp3DVector*) (buf+1);           // Make sure misaligned!
          }

        pvB    = pvA + 1;
        pvTmp2 = pvB + 1;
        pvTmp  = pvTmp2 + 1;
      }

    dif = 4;
    fPrecision = SINGLE_PRECISION;

    for ( z = -dif; z <= dif; z++ )
      {
        pvA->z = 0.123f *(float)z;
        pvB->y = 0.123f *(float)z;

        for ( y = -dif; y <= dif; y++ )
          {
            pvA->y = 0.123f *(float)y;
            pvB->x = 0.123f *(float)y;

            for ( x = -dif; x <= dif; x++ )
              { 
                pvA->x = 0.123f *(float)x;
                pvB->z = 0.123f *(float)y;
                fScalar = pvA->x;

                fWTst = ((vmp3DQVector *) pvTmp)->w = fScalar;  // Get untouchable value!

                switch(iFunc)
                  {
                    case iTEST_VEC_SQRT:
                        fPrecision = 0.0000001f;
                        bF = false;
                        if ( pvA->x < 0.0f ) pvA->x = -pvA->x;
                        if ( pvA->y < 0.0f ) pvA->y = -pvA->y;
                        if ( pvA->z < 0.0f ) pvA->z = -pvA->z;

                        vmp_VecSqrtGeneric( pvTmp2, pvA );
                        if (bAlign)
                            vmp_VecSqrt( pvTmp, pvA );      // Aligned
                        else
                            vmp_VecSqrt0( pvTmp, pvA );
                        break;

                    case iTEST_VEC_SQRT_FAST:
                        fPrecision = 0.00001f;
                        bF = false;
                        if ( pvA->x < 0.0f ) pvA->x = -pvA->x;
                        if ( pvA->y < 0.0f ) pvA->y = -pvA->y;
                        if ( pvA->z < 0.0f ) pvA->z = -pvA->z;

                        vmp_VecSqrtGeneric( pvTmp2, pvA );
                        if (bAlign)
                            vmp_VecSqrtFast( pvTmp, pvA );      // Aligned
                        else
                            vmp_VecSqrtFast0( pvTmp, pvA );
                        break;

                    case iTEST_MAGNITUDE:
                        fPrecision = 0.0000001f;
                        bF = true;
                        vmp_VecMagnitudeGeneric( &f2, pvA );
                        if (bAlign)
                            vmp_VecMagnitude( &f, pvA );        // Aligned
                        else
                            vmp_VecMagnitude0( &f, pvA );
                        break;

                    case iTEST_MAGNITUDE_FAST:
                        fPrecision = 0.0001f;
                        bF = true;
                        vmp_VecMagnitudeGeneric( &f2, pvA );
                        if (bAlign)
                            vmp_VecMagnitudeFast( &f, pvA );        // Aligned
                        else
                            vmp_VecMagnitudeFast0( &f, pvA );
                        break;

                    case iTEST_NORMALIZE:
                        fPrecision = 0.0000001f;
                        bF = false;
                        vmp_VecNormalizeGeneric( pvTmp2, pvA );
                        if (bAlign)
                            vmp_VecNormalize( pvTmp, pvA ); // Aligned
                        else
                            vmp_VecNormalize0( pvTmp, pvA );
                        break;

                    case iTEST_NORMALIZE_FAST:
                        fPrecision = 0.00001f;
                        bF = false;
                        vmp_VecNormalizeGeneric( pvTmp2, pvA );
                        if (bAlign)
                            vmp_VecNormalizeFast( pvTmp, pvA ); // Aligned
                        else
                            vmp_VecNormalizeFast0( pvTmp, pvA );
                        break;

                    case iTEST_MIN:
                        fPrecision = 0.0000001f;
                        bF = false;
                        vmp_VecMinGeneric( pvTmp2, pvA, pvB );
                        if (bAlign)
                            vmp_VecMin( pvTmp, pvA, pvB );  // Aligned
                        else
                            vmp_VecMin0( pvTmp, pvA, pvB );
                        break;

                    case iTEST_MAX:
                        fPrecision = 0.0000001f;
                        bF = false;
                        vmp_VecMaxGeneric( pvTmp2, pvA, pvB );
                        if (bAlign)
                            vmp_VecMax( pvTmp, pvA, pvB );  // Aligned
                        else
                            vmp_VecMax0( pvTmp, pvA, pvB );
                        break;
                  }
                if (bF)
                  {
                    if (!vmp_IsFEqual(f, f2, fPrecision))
                      {
                        printf( " vA.x:%f vA.y:%f vA.z:%f\n",
                                pvA->x, pvA->y, pvA->z );
                        printf( " vB.x:%f vB.y:%f vB.z:%f\n", 
                                pvB->x, pvB->y, pvB->z );
                        printf( "    %f Good vs Bad! %f  *** MISMATCH!!! ***\n\n", f2, f );
                      }
                  }
                else
                  {
                    if (!vmp_IsVEqual(pvTmp, pvTmp2, fPrecision)) 
                      {
                        printf( " vA.x:%f vA.y:%f vA.z:%f\n",
                                 pvA->x, pvA->y, pvA->z );
                        printf( " vB.x:%f vB.y:%f vB.z:%f\n",
                                 pvB->x, pvB->y, pvB->z );
                        printf( "    x:%f y:%f z:%f  Good vs Bad!!!=\n",
                                 pvTmp2->x, pvTmp2->y, pvTmp2->z );
                        printf( "    x:%f y:%f z:%f   *** MISMATCH!!! ***\n\n",
                                 pvTmp->x, pvTmp->y, pvTmp->z );
                      }
                    
                        //  Check for '.W' contamination
                    if (!vmp_IsFEqual(fWTst, ((vmp3DQVector *) pvTmp)->w, fPrecision ))
                      {
                        printf( ".W is contaminated! should be: %f not %f\n", fWTst, ((vmp3DQVector *) pvTmp)->w );
                      }
                  }
              }
          }
      }
}



int main( int argc, char *argv[] )
{
    CpuInfo cinfo;
    char szBuf[ CPU_SZBUF_MAX ];

    CpuDetect( &cinfo );                                // Detect CPU
    
    printf( "\nVector Special Function Test\n\n" );
    printf( "%s\n", CpuInfoStr( szBuf, &cinfo ) );

    Vsf3DSetup( &cinfo );                               // Setup our 3D Vector stuff

    printf( "Testing VecSqrt\n" );
    testVec( false, iTEST_VEC_SQRT );

    printf( "Testing VecSqrt\n" );
    testVec( true,  iTEST_VEC_SQRT );

    printf( "Testing VecSqrt (FAST)\n" );
    testVec( false, iTEST_VEC_SQRT_FAST );

    printf( "Testing VecSqrt (FAST)\n" );
    testVec( true,  iTEST_VEC_SQRT_FAST );

    printf( "Testing VecMin\n" );
    testVec( false, iTEST_MIN );

    printf( "Testing VecMin (Aligned)\n" );
    testVec( true,  iTEST_MIN );

    printf( "Testing VecMax\n" );
    testVec( false, iTEST_MAX );

    printf( "Testing VecMax (Aligned)\n" );
    testVec( true,  iTEST_MAX );

    printf( "Testing VecMagnitude\n" );
    testVec( false, iTEST_MAGNITUDE );

    printf( "Testing VecMagnitude (Aligned)\n" );
    testVec( true,  iTEST_MAGNITUDE );

    printf( "Testing VecMagnitude (FAST) possible Low Precision\n" );
    testVec( false, iTEST_MAGNITUDE_FAST );

    printf( "Testing VecMagnitude (Aligned) (FAST) possible Low Precision\n" );
    testVec( true,  iTEST_MAGNITUDE_FAST );

    printf( "Testing VecNormalize\n" );
    testVec( false, iTEST_NORMALIZE );

    printf( "Testing VecNormalize (Aligned)\n" );
    testVec( true,  iTEST_NORMALIZE );

    printf( "Testing VecNormalize (FAST) possible Low Precision\n" );
    testVec( false, iTEST_NORMALIZE_FAST );

    printf( "Testing VecNormalize (Aligned) (FAST) possible Low Precision\n" );
    testVec( true,  iTEST_NORMALIZE_FAST );

    printf( "Fin\n\n" );

    return 0;
}

