
//
//  Vsf3D.c - 3D Vector Special Function
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 3 Oct 2001 - James Leiterman
//
//  Last Modification
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include    <assert.h>
#include    <stdio.h>
#include    <math.h>

    //  Book Development Library Modules

#include    "CpuAsm.h"                                  // CPU
#include    "vmp3D.h"
#include    "vmp3DEx.h"

#include    "Vsf3D.h"                                   // 3D Math
#include    "Vsf3DEx.h"



vmp_VecSqrtProc             vmp_VecSqrt;
vmp_VecSqrtProc             vmp_VecSqrt0;
vmp_VecSqrtProc             vmp_VecSqrtFast;
vmp_VecSqrtProc             vmp_VecSqrtFast0;

vmp_VecMinProc              vmp_VecMin;
vmp_VecMinProc              vmp_VecMin0;
vmp_VecMaxProc              vmp_VecMax;
vmp_VecMaxProc              vmp_VecMax0;

vmp_VecMagnitudeProc        vmp_VecMagnitude;
vmp_VecMagnitudeProc        vmp_VecMagnitude0;
vmp_VecMagnitudeProc        vmp_VecMagnitudeFast;
vmp_VecMagnitudeProc        vmp_VecMagnitudeFast0;

vmp_VecNormalizeProc        vmp_VecNormalize;
vmp_VecNormalizeProc        vmp_VecNormalize0;
vmp_VecNormalizeProc        vmp_VecNormalizeFast;
vmp_VecNormalizeProc        vmp_VecNormalizeFast0;



//
//  Vector Square Root
//

void vmp_VecSqrtGeneric(vmp3DVector * const pvD, const vmp3DVector * const pvA)
{
    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);

    ASSERT_FNEG(pvA->x);
    ASSERT_FNEG(pvA->y);
    ASSERT_FNEG(pvA->z);

    pvD->x = sqrtf( pvA->x );
    pvD->y = sqrtf( pvA->y );
    pvD->z = sqrtf( pvA->z );
}



//
//  Vector Minimum
//

void vmp_VecMinGeneric( vmp3DVector * const pvD, const vmp3DVector * const pvA, const vmp3DVector * const pvB )
{
    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);

    pvD->x = (pvA->x < pvB->x) ? pvA->x : pvB->x;
    pvD->y = (pvA->y < pvB->y) ? pvA->y : pvB->y;
    pvD->z = (pvA->z < pvB->z) ? pvA->z : pvB->z;
}



//
//  Vector Maximum
//

void vmp_VecMaxGeneric( vmp3DVector * const pvD, const vmp3DVector * const pvA, const vmp3DVector * const pvB )
{
    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);

    pvD->x = (pvA->x > pvB->x) ? pvA->x : pvB->x;
    pvD->y = (pvA->y > pvB->y) ? pvA->y : pvB->y;
    pvD->z = (pvA->z > pvB->z) ? pvA->z : pvB->z;
}



//
//  Vector Magnitude
//

void vmp_VecMagnitudeGeneric(float * const pfD, const vmp3DVector * const pvA)
{
    ASSERT_PTR4(pfD);
    ASSERT_PTR4(pvA);

    *pfD = sqrtf( pvA->x * pvA->x + pvA->y * pvA->y + pvA->z * pvA->z );
}



//
//  Vector Normalization
//

void vmp_VecNormalizeGeneric(vmp3DVector * const pvD, const vmp3DVector * const pvA)
{
    float fMag;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);

    fMag = sqrtf( pvA->x * pvA->x + pvA->y * pvA->y + pvA->z * pvA->z );

    if ( fMag < SINGLE_PRECISION )  // If too close to zero...
      {
        pvD->x = pvA->x;
        pvD->y = pvA->y;
        pvD->z = pvA->z;
      }
    else
      {
        fMag = 1.0f / fMag;
        pvD->x = pvA->x * fMag;
        pvD->y = pvA->y * fMag;
        pvD->z = pvA->z * fMag;
      }
}







