
//
//  VTrig3DVU.c - 3D Vector Trignometric Function
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 3 Oct 2001 - James Leiterman
//
//  Last Modification
//
//  stab=4
//


#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include    <assert.h>
#include    <stdio.h>
#include    <math.h>

    //  Book Development Library Modules

#include    "vmp3D.h"
#include    "vmp3DEx.h"
#include    "VTrig3D.h"                                 // 3D Math
#include    "VTrig3DEx.h"



void vmp_FSinCosVU( float * const pfSin, float * const pfCos, float fRads )
{
    ASSERT_PTR( pfSin );
    ASSERT_PTR( pfCos );

    *pfCos = cosf( fRads );
    *pfSin = sinf( fRads );
}


void vmp_FSinVU( float * const pfD, float fRads )
{
    ASSERT_PTR( pfD );

    *pfD = sinf( fRads );
}


void vmp_FCosVU( float * const pfD, float fRads )
{
    ASSERT_PTR( pfD );

    *pfD = cosf( fRads );
}


void vmp_FTanVU( float * const pfD, float fRads )
{
    ASSERT_PTR( pfD );

    *pfD = tanf( fRads );
}



//  Setup Vec3D module
//
//  NOTE: This function sets up the function vectors so the best function
//      is attached to the vector for the processor running.
//

void VTrig3DSetup( const CpuInfo * const pcinfo )
{
    if ( CPUBITS_VU0 & pcinfo->nBits )
      {
        vmp_FSinCos =               vmp_FSinCosVU;
        vmp_FSin =                  vmp_FSinVU;
        vmp_FCos =                  vmp_FCosVU;
        vmp_FTan =                  vmp_FTanVU;
      }

    else
      {
        vmp_FSinCos =               vmp_FSinCosGeneric;
        vmp_FSin =                  vmp_FSinGeneric;
        vmp_FCos =                  vmp_FCosGeneric;
        vmp_FTan =                  vmp_FTanGeneric;
      }
}





