
//
//  Bench.c - Packed Motion Compensation
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 21 Feb, 2002 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include <stdio.h>
#include <assert.h>
#include <string.h>

    //  Book Development Library Modules
    
#include "CpuAsm.h"										// CPU Module
#include "vmp3D.h"
#include "MComp.h"
#include "MCompEx.h"



typedef enum
{
	iTEST_MCOMP_HZ_8R,			// 8x8   Horizontal Motion Compensation w/Rounding
	iTEST_MCOMP_HZ_16R,			// 16x16 Horizontal Motion Compensation w/Rounding
	iTEST_MCOMP_VT_8R,			// 8x8   Vertical Motion Compensation w/Rounding
	iTEST_MCOMP_VT_16R			// 16x16 Vertical Motion Compensation w/Rounding

} iTEST;


//
//	16x16 Motion Compensation Data Block within 32x17 image buffer
//
//	NOTE: A test sample extracted from an actual film clip.
//		ALSO we need this 16byte aligned!
//

#define MCTestDataWidth		(16+16)
#define MCTestDataHeight	(16+1) 

#define MCTestDataSize		(MCTestDataWidth*MCTestDataHeight)

uint8 MCTestData[ MCTestDataSize ] =	// 32x17
{	//                                                                                                 |
	0xdc, 0xdc, 0xdc, 0xdc,  0xdc, 0xdd, 0xdd, 0xdd,  0xd5, 0xd8, 0xdb, 0xde,  0xde, 0xdc, 0xd9, 0xd7,   0x80,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0xd5, 0xd7, 0xd9, 0xdb,  0xdd, 0xde, 0xdd, 0xdd,  0xda, 0xdc, 0xe0, 0xe2,  0xe1, 0xde, 0xda, 0xd8,   0x80,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0xc5, 0xcb, 0xd3, 0xda,  0xdc, 0xda, 0xd5, 0xd1,  0xdb, 0xdc, 0xdd, 0xdf,  0xe1, 0xe3, 0xe5, 0xe5,   0x80,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0xad, 0xb1, 0xb7, 0xc0,  0xca, 0xd5, 0xdd, 0xe2,  0xe0, 0xdf, 0xdd, 0xdd,  0xde, 0xe0, 0xe3, 0xe5,   0x80,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,

	0xc6, 0xbd, 0xb2, 0xac,  0xaf, 0xbb, 0xca, 0xd5,  0xd7, 0xd9, 0xdd, 0xdf,  0xdf, 0xdb, 0xd6, 0xd3,   0x80,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0xe5, 0xdd, 0xd1, 0xc3,  0xb6, 0xae, 0xa9, 0xa6,  0xb2, 0xbb, 0xca, 0xd7,  0xde, 0xde, 0xd9, 0xd5,   0x80,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0xd8, 0xdc, 0xdf, 0xdf,  0xd8, 0xcc, 0xbf, 0xb6,  0xa4, 0xa8, 0xb1, 0xbc,  0xc9, 0xd5, 0xdf, 0xe5,   0x80,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0xd9, 0xdb, 0xdf, 0xe2,  0xe3, 0xe3, 0xe2, 0xe1,  0xbc, 0xb3, 0xa7, 0xa2,  0xab, 0xbf, 0xd6, 0xe6,   0x80,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,

	0xac, 0xaf, 0xb4, 0xba,  0xc1, 0xc8, 0xcd, 0xcf,  0xd7, 0xd7, 0xd2, 0xc5,  0xb6, 0xad, 0xae, 0xb3,   0x80,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0x88, 0x8a, 0x8e, 0x93,  0x98, 0x9d, 0xa1, 0xa3,  0xac, 0xb1, 0xb5, 0xb2,  0xa8, 0xa0, 0x9e, 0x9f,   0x80,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0x84, 0x85, 0x87, 0x8a,  0x8d, 0x90, 0x92, 0x93,  0x88, 0x8e, 0x94, 0x97,  0x95, 0x93, 0x93, 0x95,   0x80,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0x85, 0x85, 0x86, 0x87,  0x88, 0x89, 0x8a, 0x8a,  0x82, 0x82, 0x83, 0x84,  0x87, 0x8c, 0x93, 0x98,   0x80,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,

	0x81, 0x81, 0x81, 0x81,  0x81, 0x81, 0x82, 0x82,  0x82, 0x80, 0x7e, 0x7f,  0x84, 0x89, 0x8e, 0x90,   0x80,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0x7e, 0x7e, 0x7e, 0x7f,  0x7f, 0x80, 0x80, 0x81,  0x7a, 0x7a, 0x7b, 0x80,  0x85, 0x85, 0x81, 0x7c,   0x80,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0x71, 0x71, 0x72, 0x74,  0x76, 0x77, 0x78, 0x79,  0x78, 0x76, 0x76, 0x7b,  0x82, 0x83, 0x7d, 0x76,   0x80,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0x6e, 0x6f, 0x71, 0x74,  0x76, 0x79, 0x7a, 0x7b,  0x7f, 0x78, 0x72, 0x74,  0x7c, 0x83, 0x83, 0x80,   0x80,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	//-------------------------------------------------------------------------------------------------|
	0x80, 0x80, 0x80, 0x80,  0x80, 0x80, 0x80, 0x80,  0x80, 0x80, 0x80, 0x80,  0x80, 0x80, 0x80, 0x80,   0x80,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
};



//
//	Compare blocks
//
static bool vmp_IsMBlockEqual( const uint8 * const pbTblA, const uint8 * const pbTblB, uint wDiag, int iStride )
{
	uint dx;
	uint32 *pA, *pB;
	uint w;

	ASSERT_PTR(pbTblA);
	ASSERT_PTR(pbTblB);
	ASSERT_ZERO(wDiag);
	ASSERT(0 == (wDiag&7));

	w = wDiag >> 2;						// # of 32bit values
	pA = (uint32*) pbTblA;
	pB = (uint32*) pbTblB;
	iStride = (iStride >> 2) - w;		// # of 32bit blocks to skip

	do {
		for (dx = 0; dx < w; dx++)
		  {
			if ( *pA++ != *pB++ )
			  {
				return false;
			  }
		  }

		pA += iStride;		// +- to beginning next scanline!
		pB += iStride;
	} while (--wDiag);

	return true;
}



//
//	Test 
//
static void test_MComp( iTEST iFunc )
{
	uint	v, wDiag;
    uint8	buf[3*MCTestDataSize  +15+16], *bt, *bu, *bv;

		//	Verify Master Data has 128bit aligned width
	ASSERT(MCTestDataWidth == ALIGN16(MCTestDataWidth));

	wDiag = 0;
	
		//	Destination needs to be aligned
		
	bu = (uint8 *) ALIGN16((uint)buf);
	bv = bu + MCTestDataSize;

	for ( v = 0; v < 16; v++ )
	  {
		bt = bv + MCTestDataSize + v;					// Source needs to be misaligned
		
		memcpy( bt, MCTestData, sizeof(MCTestData));	// Copy to guarantee alignment
		memset( bu, 0x55, MCTestDataSize );
		memset( bv, 0xaa, MCTestDataSize );

		switch (iFunc)
		  {
			case iTEST_MCOMP_HZ_8R:
				wDiag = 8;
#ifdef CC_VMP_MIPS
				vmp_MCompHz8RGenericSLOW( bu, bt, MCTestDataWidth );
#else
				vmp_MCompHz8RGeneric( bu, bt, MCTestDataWidth );
#endif
				vmp_MCompHz8R(        bv, bt, MCTestDataWidth );
				break;

			case iTEST_MCOMP_HZ_16R:
				wDiag = 16;
#ifdef CC_VMP_MIPS
				vmp_MCompHz16RGenericSLOW( bu, bt, MCTestDataWidth );
#else
				vmp_MCompHz16RGeneric( bu, bt, MCTestDataWidth );
#endif
				vmp_MCompHz16R(        bv, bt, MCTestDataWidth );
				break;

			case iTEST_MCOMP_VT_8R:
				wDiag = 8;
#ifdef CC_VMP_MIPS
				vmp_MCompVt8RGenericSLOW( bu, bt, MCTestDataWidth );
#else
				vmp_MCompVt8RGeneric( bu, bt, MCTestDataWidth );
#endif
				vmp_MCompVt8R(        bv, bt, MCTestDataWidth );
				break;

			case iTEST_MCOMP_VT_16R:
				wDiag = 16;
#ifdef CC_VMP_MIPS
				vmp_MCompVt16RGenericSLOW( bu, bt, MCTestDataWidth );
#else
				vmp_MCompVt16RGeneric( bu, bt, MCTestDataWidth );
#endif
				vmp_MCompVt16R(        bv, bt, MCTestDataWidth );
				break;
		  }

		if ( !vmp_IsMBlockEqual( bu, bv, wDiag, MCTestDataWidth ) )
		  {
			int offset = ((uint)bt)&15;
			printf( "Mismatch on %u\n", offset );
		  }
	  } 
}



int main( int argc, char *argv[] )
{
    CpuInfo cinfo;
	char szBuf[ CPU_SZBUF_MAX ];

    CpuDetect( &cinfo );								// Detect CPU
    
    printf( "\nPacked Motion Compensation Test\n\n" );
	printf( "%s\n", CpuInfoStr( szBuf, &cinfo ));
    
	MCompSetup( &cinfo );								// Setup our Motion Compensation

	printf( "Testing Horizontal 8x8 Motion Compensation w/Rounding\n" );
	test_MComp( iTEST_MCOMP_HZ_8R );

	printf( "Testing Horizontal 16x16 Motion Compensation w/Rounding\n" );
	test_MComp( iTEST_MCOMP_HZ_16R );

	printf( "Testing Vertical 8x8 Motion Compensation w/Rounding\n" );
	test_MComp( iTEST_MCOMP_VT_8R );

	printf( "Testing Vertical 16x16 Motion Compensation w/Rounding\n" );
	test_MComp( iTEST_MCOMP_VT_16R );

	printf( "Fin!\n\n" );

    return 0;
}

