
//
//  MCompMMI.s - Packed (128bit Mips) Motion Compensation
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//		This module is designed for fast 3D math
//
//  Created - 21 Feb, 2002 - James Leiterman
//
//  Modifications
//
//
//  stab = 4
//

	.nolist
#include "Mips.inc"
#include "KariType.inc"
	.list

	.set	noreorder
	.set	nomips16

	.rdata

	.align	4

OneHO:
	.dword	0x0001000100010001,0x0001000100010001

	.section .text                                                    



//_________________________________________________________________________________________________
//=================================================================================================
//	 8x8 Horizontal Motion Compensation w/Rounding
//
//	NOTE: Stride (+) is top down and (-) is bottom up orientation
//		Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
//
//	8bit values are unpacked into 16bit elements, rounding average calculated
//	then two 16bit elements are packed and interlaced into single 8bit element.
//
//	[X][X]
//	[ ][ ]
//
//	void vmp_MCompHz8RMMI( uint8 * pbDst, const uint8 * pbSrc, int iStride )
//
//	*** MMI version *** 64bit (un)aligned

	.align 3
	.global vmp_MCompHz8RMMI
vmp_MCompHz8RMMI:

ASSERT_PTR(a0)
ASSERT_PTR(a1)
ASSERT_ZERO(a2)

	la		t4,OneHO
	lq		t4,0(t4)		// 0001000100010001 0001000100010001
	li		t5,8			// # of scanlines

		//	Load 1st & 2nd 64bits 'S' & 'R' from (un)aligned memory
$H0:
	ldl		t0,	7(a1)		// S  {63...0}
	ldl		t1, 8(a1)		// R  {63...0}
 	ldr		t0, 0(a1)
	ldr		t1, 1(a1)
	add		a1,a1,a2		//  pbSrc += iStride

	pextlb	t0,zero,t0		//  [ _7 _6 _5 _4 _3 _2 _1 _0 ]
	pextlb	t1,zero,t1		//  [ _8 _7 _6 _5 _4 _3 _2 _1 ]

		//	_________________
		//	Now average t1:t0  (A + B + 1)>>1

	paddh	t0,t0,t1		// S+R
	paddh	t0,t0,t4		// [...]+1
	psrlh	t0,t0,1			// ()>>1

	ppacb	t0,t0,t0		// [7 6 5 4 3 2 1 0 7 6 5 4 3 2 1 0]

		//	_________________
		//	Save result to (unaligned) memory

	sdl		t0,	7(a0)		// pwD {63...0}
	sdr		t0, 0(a0)
	add		a0,a0,a2		// pDst += iStride

	addi	t5,t5,-1
	bne		zero,t5,$H0		// Loop for 8 scanlines
	BDELAY					// nop = Branch Delay Slot

	j		ra   			// ret
	BDELAY					// nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//	 16x16 Horizontal Motion Compensation w/Rounding
//
//	NOTE: Stride (+) is top down and (-) is bottom up orientation
//		Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
//
//	8bit values are unpacked into 16bit elements, rounding average calculated
//	then two 16bit elements are packed and interlaced into single 8bit element.
//
//	[X][X]
//	[ ][ ]
//
//	void vmp_MCompHz16RMMI( uint8 * pbDst, const uint8 * pbSrc, int iStride )
//
//	*** MMI version *** 128bit (un)aligned

	.align 3
	.global vmp_MCompHz16RMMI
vmp_MCompHz16RMMI:

ASSERT_PTR(a0)
ASSERT_PTR(a1)
ASSERT_ZERO(a2)

	la		t4,OneHO
	lq		t4,0(t4)		// 0001000100010001 0001000100010001
	li		t5,16			// # of scanlines

		//	Load 1st 128bits 'S' from (un)aligned memory
$H1:
	ldl		t0,	7(a1)		// S  {63...0}
	ldl		t1,	8(a1)		// R  {63...0}
	ldl		t2, 15(a1)		//     {127...64}
	ldl		t3, 16(a1)		// R  {127...64}

 	ldr		t0, 0(a1)
 	ldr		t1, 1(a1)
	ldr		t2, 8(a1)
	ldr		t3, 9(a1)

	add		a1,a1,a2		//  pbSrc += iStride

	pcpyld	t0,t2,t0		// S  {127...0}
	pcpyld	t1,t3,t1		// R  {127...0}

	pextub	t2,zero,t0		// S[ _F _E _D _C _B _A _9 _8 ]
	pextub	t3,zero,t1		// R[_10 _F _E _D _C _B _A _9 ]
	pextlb	t0,zero,t0		//  [ _7 _6 _5 _4 _3 _2 _1 _0 ]
	pextlb	t1,zero,t1		//  [ _8 _7 _6 _5 _4 _3 _2 _1 ]

		//	_________________
		//	Now average t2:t0 and t3:t1 : (A + B + 1)>>1

	paddh	t2,t2,t3		// S+R
	paddh	t0,t0,t1		// S+R
	paddh	t2,t2,t4		// [...]+1
	paddh	t0,t0,t4		// [...]+1
	psrlh	t2,t2,1			// ()>>1
	psrlh	t0,t0,1			// ()>>1

	ppacb	t0,t2,t0		// Interleave 16 halfwords to 16 bytes

		//	_________________
		//	Save result to (unaligned) memory

	pcpyud	t2,t0,t0		// D  {63...0  127...64}
	sdl		t0,	7(a0)		// pwD {63...0}
	sdr		t0, 0(a0)
	sdl		t2, 15(a0)		//    {127...64}
	sdr		t2, 8(a0)

	add		a0,a0,a2		// pDst += iStride

	addi	t5,t5,-1
	bne		zero,t5,$H1		// Loop for 16 scanlines
	BDELAY					// nop = Branch Delay Slot

	j		ra   			// ret
	BDELAY					// nop = Branch Delay Slot

					 	

//_________________________________________________________________________________________________
//=================================================================================================
//	 8x8 Vertical Motion Compensation w/Rounding
//
//	NOTE: Stride (+) is top down and (-) is bottom up orientation
//		Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
//
//	8bit values are unpacked into 16bit elements, rounding average calculated
//	then two 16bit elements are packed and interlaced into single 8bit element.
//
//	[X][ ]
//	[X][ ]
//
//	void vmp_MCompVt8RMMI( uint8 * pbDst, const uint8 * pbSrc, int iStride )
//
//	*** MMI version *** 64bit (un)aligned

	.align 3
	.global vmp_MCompVt8RMMI
vmp_MCompVt8RMMI:

ASSERT_PTR(a0)
ASSERT_PTR(a1)
ASSERT_ZERO(a2)

	la		t4,OneHO
	lq		t4,0(t4)		// 0001000100010001 0001000100010001
	li		t5,8			// # of scanlines

		//	Load 1st 64bits 'S' from (un)aligned memory

	ldl		t0,	7(a1)		// S  {63...0}
 	ldr		t0, 0(a1)
	add		a1,a1,a2		//  pbSrc += iStride

	pextlb	t0,zero,t0		//  [ _7 _6 _5 _4 _3 _2 _1 _0 ]

		//	_________________
$V0:	//	Load 64bits from (un)aligned memory

	ldl		t1,	7(a1)		// R  {63...0}
 	ldr		t1, 0(a1)
	add		a1,a1,a2		//  pbSrc += iStride

	pextlb	t1,zero,t1		//  [ _7 _6 _5 _4 _3 _2 _1 _0 ]

		//	_________________
		//	Now average t2:t0 : (A + B + 1)>>1

	paddh	t0,t0,t1		// S+R
	paddh	t0,t0,t4		// [...]+1
	psrlh	t0,t0,1			// ()>>1

	ppacb	t0,t0,t0		// [7 6 5 4 3 2 1 0 7 6 5 4 3 2 1 0]

		//	_________________
		//	Save result to (unaligned) memory

	sdl		t0,	7(a0)		// pwD {63...0}
	sdr		t0, 0(a0)

	add		a0,a0,a2		// pDst += iStride

	por		t0,t1,t1		// R = S Copy register(s)

	addi	t5,t5,-1
	bne		zero,t5,$V0		// Loop for 8 scanlines
	BDELAY					// nop = Branch Delay Slot

	j		ra   			// ret
	BDELAY					// nop = Branch Delay Slot


		
//_________________________________________________________________________________________________
//=================================================================================================
//	 16x16 Vertical Motion Compensation w/Rounding
//
//	NOTE: Stride (+) is top down and (-) is bottom up orientation
//		Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
//
//	8bit values are unpacked into 16bit elements, rounding average calculated
//	then two 16bit elements are packed and interlaced into single 8bit element.
//
//	[X][ ]
//	[X][ ]
//
//	void vmp_MCompVt16RMMI( uint8 * pbDst, const uint8 * pbSrc, int iStride )
//
//	*** MMI version *** 128bit (un)aligned

	.align 3
	.global vmp_MCompVt16RMMI
vmp_MCompVt16RMMI:

ASSERT_PTR(a0)
ASSERT_PTR(a1)
ASSERT_ZERO(a2)

	la		t4,OneHO
	lq		t4,0(t4)		// 0001000100010001 0001000100010001
	li		t5,16			// # of scanlines

		//	Load 1st 128bits 'S' from (un)aligned memory

	ldl		t0,	7(a1)		// S  {63...0}
	ldl		t2, 15(a1)		//     {127...64}
 	ldr		t0, 0(a1)
	ldr		t2, 8(a1)

	add		a1,a1,a2		//  pbSrc += iStride

	pcpyld	t0,t2,t0		// S  {127...0}
	pextub	t2,zero,t0		// S[ _F _E _D _C _B _A _9 _8 ]
	pextlb	t0,zero,t0		//  [ _7 _6 _5 _4 _3 _2 _1 _0 ]

		//	_________________
$V1:	//	Load 128bits from (un)aligned memory

	ldl		t1,	7(a1)		// R  {63...0}
	ldl		t3, 15(a1)		// R  {127...64}
 	ldr		t1, 0(a1)
	ldr		t3, 8(a1)
	add		a1,a1,a2		//  pbSrc += iStride

	pcpyld	t1,t3,t1		// R  {127...0}
	pextub	t3,zero,t1		// R[ _F _E _D _C _B _A _9 _8 ]
	pextlb	t1,zero,t1		//  [ _7 _6 _5 _4 _3 _2 _1 _0 ]

		//	_________________
		//	Now average t2:t0 and t3:t1 : (A + B + 1)>>1

	paddh	t2,t2,t3		// S+R
	paddh	t0,t0,t1		// S+R
	paddh	t2,t2,t4		// [...]+1
	paddh	t0,t0,t4		// [...]+1
	psrlh	t2,t2,1			// ()>>1
	psrlh	t0,t0,1			// ()>>1

	ppacb	t0,t2,t0		// Interleave 16 halfwords to 16 bytes

		//	_________________
		//	Save result to (unaligned) memory

	pcpyud	t2,t0,t0		// D  {63...0  127...64}

	sdr		t0, 0(a0)
	sdl		t0,	7(a0)		// pwD {63...0}
	sdr		t2, 8(a0)
	sdl		t2, 15(a0)		//    {127...64}

	add		a0,a0,a2		// pDst += iStride

	por		t0,t1,t1		// R = S Copy register(s)
	por		t2,t3,t3		// R = S

	addi	t5,t5,-1
	bne		zero,t5,$V1		// Loop for 16 scanlines
	BDELAY					// nop = Branch Delay Slot

	j		ra   			// ret
	BDELAY					// nop = Branch Delay Slot

					 	
