
//
//  yuv12.c - YUV12 bit to RGB Conversion
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 25 Feb, 2002 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include    <assert.h>
#include    <stdio.h>

    //  Book Development Library Modules
    
#include "yuv12.h"
#include "yuv12Ex.h"



vmp_yuv12toRGBProc  vmp_yuv12toRGB32;


    //  8bit saturation
#define SAT_RGB8(a)     (a)>255 ? 255 : (a)<0 ? 0 : (a)


//
//  YUV12 to RGB Conversion
//
//  NOTE: Optimized from original code found in ProjectMayo Open
//       DivX;-) Source Code! yuv2rgb.c

void vmp_yuv12toRGB32Generic(
    void *pImage, uint nWidth, uint nHeight, int iStride,
    const uint8 * pBaseY, uint nStrideY, 
    const uint8 * pBaseU, const uint8 * pBaseV, uint nStrideUV )
{
    uint w, suv;
    byte *pImg;

    ASSERT_PTR(pBaseY);
    ASSERT_PTR(pBaseU);
    ASSERT_PTR(pBaseV);
    ASSERT_PTR(pImage);
    ASSERT_ZERO(nStrideY);
    ASSERT_ZERO(nStrideUV);
    ASSERT_ZERO(nWidth);
    ASSERT_ZERO(nHeight);
    ASSERT_ZERO(iStride);

    pImg = (uint8*)pImage;
    iStride -= (nWidth<<2);               // Remaining bytes to skip
    nStrideY -= nWidth;                 // remaining Y table skip
    nWidth >>= 1;
    suv = 0;

    do {
        uint8 const *pu, *pv;
                
        pu = pBaseU;
        pv = pBaseV;
        w = nWidth;

        do {
            signed int u,v, y0,y1, ur,ug,vg,vb, r,g,b;

            u = *pu++ - 128;
            v = *pv++ - 128;
            y0 =  0x2568 * (*pBaseY++ - 16);
            y1 =  0x2568 * (*pBaseY++ - 16);
            ur =  0x3343 * u;
            ug = -0x1a1e * u;       // 0xe5e2
            vg = -0x0c92 * v;       // 0xf36e
            vb =  0x40cf * v;           

                //#define SAT_RGB8(a)       (a)>255 ? 255 : (a)<0 ? 0 : (a)

#ifdef VMP_BIG_ENDIAN
            r = (y0+   ur) >> 13; pImg[3] = SAT_RGB8(r);  // Saturate RGB
            g = (y0+vg+ug) >> 13; pImg[2] = SAT_RGB8(g);
            b = (y0+vb   ) >> 13; pImg[1] = SAT_RGB8(b);
            pImg[0] = 0;
            r = (y1+   ur) >> 13; pImg[7] = SAT_RGB8(r);
            g = (y1+vg+ug) >> 13; pImg[6] = SAT_RGB8(g);
            b = (y1+vb   ) >> 13; pImg[5] = SAT_RGB8(b);
            pImg[4] = 0;
#else
            r = (y0+   ur) >> 13; pImg[0] = SAT_RGB8(r);  // Saturate RGB
            g = (y0+vg+ug) >> 13; pImg[1] = SAT_RGB8(g);
            b = (y0+vb   ) >> 13; pImg[2] = SAT_RGB8(b);
            pImg[3] = 0;
            r = (y1+   ur) >> 13; pImg[4] = SAT_RGB8(r);
            g = (y1+vg+ug) >> 13; pImg[5] = SAT_RGB8(g);
            b = (y1+vb   ) >> 13; pImg[6] = SAT_RGB8(b);
            pImg[7] = 0;
#endif
            pImg += 8;
        } while( --w );

        pBaseY += nStrideY;
        pBaseU += suv;
        pBaseV += suv;
        pImg   += iStride;
        suv ^= nStrideUV;       // every other scanline
    } while (--nHeight);
}





