
//
//  yuv12MMI.s - Packed (128bit Mips) YUV12bit to RGB
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//      This module is designed for fast 3D math
//
//  Created - 21 Feb, 2002 - James Leiterman
//
//  Modifications
//
//
//  stab = 4
//

    .nolist
#include "Mips.inc"
#include "KariType.inc"
    .list

    .set    noreorder
    .set    nomips16

    .rdata

    .align  4

OneHO:
    .dword  0x0001000100010001,0x0001000100010001

    .section .text                                                    



//_________________________________________________________________________________________________
//=================================================================================================
//   YUV12 to RGB32
//
//  NOTE: Stride (+) is top down and (-) is bottom up orientation
//      Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
//
//  8bit values are unpacked into 16bit elements, rounding average calculated
//  then two 16bit elements are packed and interlaced into single 8bit element.
//
//  [X][X]
//  [ ][ ]
//
//void vmp_yuv12toRGB32Generic(
//  void *pImage,       // a0
//  uint nWidth,        // a1
//  uint nHeight,       // a2
//  int iStride,        // a3
//  const uint8 * pBaseY,
//  uint nStrideY, 
//  const uint8 * pBaseU,
//  const uint8 * pBaseV,
//  uint nStrideUV )

//$4,$16  pImage
//$5,64  width
//$6,64  height
//$7,256 iStride

//  *** MMI version *** 128bit (un)aligned

    .align 3
    .global vmp_yuv12toRGB32MMI
vmp_yuv12toRGB32MMI:

ASSERT_PTR(a0)
ASSERT_ZERO(a1)
ASSERT_ZERO(a2)
ASSERT_ZERO(a3)

    move    t9,sp
    subu    t9,t9,64

    ldl     t0, 0x07(t9)        // S  {63...0}
    ldr     t0, 0x00(t9)
    ldl     t2, 0x0f(t9)        //     {127...64}
    ldr     t2, 0x08(t9)
    pcpyld  t0,t2,t0        // S  {127...0}

    ldl     t1, 0x17(t9)        // R  {63...0}
    ldr     t1, 0x10(t9)
    ldl     t3, 0x1f(t9)        // R  {127...64}
    ldr     t3, 0x18(t9)
    pcpyld  t1,t3,t1        // R  {127...0}

sq  t0,0x00(a1)
sq  t1,0x10(a1)
    ldl     t0, 0x27(t9)        // S  {63...0}
    ldr     t0, 0x20(t9)
    ldl     t2, 0x2f(t9)        //     {127...64}
    ldr     t2, 0x28(t9)
    pcpyld  t0,t2,t0        // S  {127...0}

    ldl     t1, 0x37(t9)        // R  {63...0}
    ldr     t1, 0x30(t9)
    ldl     t3, 0x3f(t9)        // R  {127...64}
    ldr     t3, 0x38(t9)
    pcpyld  t1,t3,t1        // R  {127...0}

sq  t0,0x20(a1)
sq  t1,0x30(a1)

    ldl     t0, 0x47(t9)        // S  {63...0}
    ldr     t0, 0x40(t9)
    ldl     t2, 0x4f(t9)        //     {127...64}
    ldr     t2, 0x48(t9)
    pcpyld  t0,t2,t0        // S  {127...0}

    ldl     t1, 0x57(t9)        // R  {63...0}
    ldr     t1, 0x50(t9)
    ldl     t3, 0x5f(t9)        // R  {127...64}
    ldr     t3, 0x58(t9)
    pcpyld  t1,t3,t1        // R  {127...0}

sq  t0,0x40(a1)
sq  t1,0x50(a1)
    ldl     t0, 0x67(t9)        // S  {63...0}
    ldr     t0, 0x60(t9)
    ldl     t2, 0x6f(t9)        //     {127...64}
    ldr     t2, 0x68(t9)
    pcpyld  t0,t2,t0        // S  {127...0}

    ldl     t1, 0x77(t9)        // R  {63...0}
    ldr     t1, 0x70(t9)
    ldl     t3, 0x7f(t9)        // R  {127...64}
    ldr     t3, 0x78(t9)
    pcpyld  t1,t3,t1        // R  {127...0}

sq  t0,0x60(a1)
sq  t1,0x70(a1)
    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



    la      t4,OneHO
    lq      t4,0(t4)        // 0001000100010001 0001000100010001
    li      t5,8            // # of scanlines

        //  Load 1st & 2nd 64bits 'S' & 'R' from (un)aligned memory
$H0:
    ldl     t0, 7(a1)       // S  {63...0}
    ldl     t1, 8(a1)       // R  {63...0}
    ldr     t0, 0(a1)
    ldr     t1, 1(a1)
    add     a1,a1,a2        //  pbSrc += iStride

    pextlb  t0,zero,t0      //  [ _7 _6 _5 _4 _3 _2 _1 _0 ]
    pextlb  t1,zero,t1      //  [ _8 _7 _6 _5 _4 _3 _2 _1 ]

        //  _________________
        //  Now average t1:t0  (A + B + 1)>>1

    paddh   t0,t0,t1        // S+R
    paddh   t0,t0,t4        // [...]+1
    psrlh   t0,t0,1         // ()>>1

    ppacb   t0,t0,t0        // [7 6 5 4 3 2 1 0 7 6 5 4 3 2 1 0]

        //  _________________
        //  Save result to (unaligned) memory

    sdl     t0, 7(a0)       // pwD {63...0}
    sdr     t0, 0(a0)
    add     a0,a0,a2        // pDst += iStride

    addi    t5,t5,-1
    bne     zero,t5,$H0     // Loop for 8 scanlines
    BDELAY                  // nop = Branch Delay Slot

    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot



//_________________________________________________________________________________________________
//=================================================================================================
//   16x16 Horizontal Motion Compensation w/Rounding
//
//  NOTE: Stride (+) is top down and (-) is bottom up orientation
//      Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
//
//  8bit values are unpacked into 16bit elements, rounding average calculated
//  then two 16bit elements are packed and interlaced into single 8bit element.
//
//  [X][X]
//  [ ][ ]
//
//  void vmp_MCompHz16RMMI( uint8 * pbDst, const uint8 * pbSrc, int iStride )
//
//  *** MMI version *** 128bit (un)aligned

    .align 3
    .global vmp_MCompHz16RMMI
vmp_MCompHz16RMMI:

ASSERT_PTR(a1)
ASSERT_PTR(a2)
ASSERT_PTR(a0)

    la      t4,OneHO
    lq      t4,0(t4)        // 0001000100010001 0001000100010001
    li      t5,16           // # of scanlines

        //  Load 1st 128bits 'S' from (un)aligned memory
$H1:
    ldl     t0, 7(a1)       // S  {63...0}
    ldl     t1, 8(a1)       // R  {63...0}
    ldl     t2, 15(a1)      //     {127...64}
    ldl     t3, 16(a1)      // R  {127...64}

    ldr     t0, 0(a1)
    ldr     t1, 1(a1)
    ldr     t2, 8(a1)
    ldr     t3, 9(a1)

    add     a1,a1,a2        //  pbSrc += iStride

    pcpyld  t0,t2,t0        // S  {127...0}
    pcpyld  t1,t3,t1        // R  {127...0}

    pextub  t2,zero,t0      // S[ _F _E _D _C _B _A _9 _8 ]
    pextub  t3,zero,t1      // R[_10 _F _E _D _C _B _A _9 ]
    pextlb  t0,zero,t0      //  [ _7 _6 _5 _4 _3 _2 _1 _0 ]
    pextlb  t1,zero,t1      //  [ _8 _7 _6 _5 _4 _3 _2 _1 ]

        //  _________________
        //  Now average t2:t0 and t3:t1 : (A + B + 1)>>1

    paddh   t2,t2,t3        // S+R
    paddh   t0,t0,t1        // S+R
    paddh   t2,t2,t4        // [...]+1
    paddh   t0,t0,t4        // [...]+1
    psrlh   t2,t2,1         // ()>>1
    psrlh   t0,t0,1         // ()>>1

    ppacb   t0,t2,t0        // Interleave 16 halfwords to 16 bytes

        //  _________________
        //  Save result to (unaligned) memory

    pcpyud  t2,t0,t0        // D  {63...0  127...64}
    sdl     t0, 7(a0)       // pwD {63...0}
    sdr     t0, 0(a0)
    sdl     t2, 15(a0)      //    {127...64}
    sdr     t2, 8(a0)

    add     a0,a0,a2        // pDst += iStride

    addi    t5,t5,-1
    bne     zero,t5,$H1     // Loop for 16 scanlines
    BDELAY                  // nop = Branch Delay Slot

    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot

                        

//_________________________________________________________________________________________________
//=================================================================================================
//   8x8 Vertical Motion Compensation w/Rounding
//
//  NOTE: Stride (+) is top down and (-) is bottom up orientation
//      Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
//
//  8bit values are unpacked into 16bit elements, rounding average calculated
//  then two 16bit elements are packed and interlaced into single 8bit element.
//
//  [X][ ]
//  [X][ ]
//
//  void vmp_MCompVt8RMMI( uint8 * pbDst, const uint8 * pbSrc, int iStride )
//
//  *** MMI version *** 64bit (un)aligned

    .align 3
    .global vmp_MCompVt8RMMI
vmp_MCompVt8RMMI:

ASSERT_PTR(a1)
ASSERT_PTR(a2)
ASSERT_PTR(a0)

    la      t4,OneHO
    lq      t4,0(t4)        // 0001000100010001 0001000100010001
    li      t5,8            // # of scanlines

        //  Load 1st 64bits 'S' from (un)aligned memory

    ldl     t0, 7(a1)       // S  {63...0}
    ldr     t0, 0(a1)
    add     a1,a1,a2        //  pbSrc += iStride

    pextlb  t0,zero,t0      //  [ _7 _6 _5 _4 _3 _2 _1 _0 ]

        //  _________________
$V0:    //  Load 64bits from (un)aligned memory

    ldl     t1, 7(a1)       // R  {63...0}
    ldr     t1, 0(a1)
    add     a1,a1,a2        //  pbSrc += iStride

    pextlb  t1,zero,t1      //  [ _7 _6 _5 _4 _3 _2 _1 _0 ]

        //  _________________
        //  Now average t2:t0 : (A + B + 1)>>1

    paddh   t0,t0,t1        // S+R
    paddh   t0,t0,t4        // [...]+1
    psrlh   t0,t0,1         // ()>>1

    ppacb   t0,t0,t0        // [7 6 5 4 3 2 1 0 7 6 5 4 3 2 1 0]

        //  _________________
        //  Save result to (unaligned) memory

    sdl     t0, 7(a0)       // pwD {63...0}
    sdr     t0, 0(a0)

    add     a0,a0,a2        // pDst += iStride

    por     t0,t1,t1        // R = S Copy register(s)

    addi    t5,t5,-1
    bne     zero,t5,$V0     // Loop for 8 scanlines
    BDELAY                  // nop = Branch Delay Slot

    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot


        
//_________________________________________________________________________________________________
//=================================================================================================
//   16x16 Vertical Motion Compensation w/Rounding
//
//  NOTE: Stride (+) is top down and (-) is bottom up orientation
//      Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
//
//  8bit values are unpacked into 16bit elements, rounding average calculated
//  then two 16bit elements are packed and interlaced into single 8bit element.
//
//  [X][ ]
//  [X][ ]
//
//  void vmp_MCompVt16RMMI( uint8 * pbDst, const uint8 * pbSrc, int iStride )
//
//  *** MMI version *** 128bit (un)aligned

    .align 3
    .global vmp_MCompVt16RMMI
vmp_MCompVt16RMMI:

ASSERT_PTR(a1)
ASSERT_PTR(a2)
ASSERT_PTR(a0)

    la      t4,OneHO
    lq      t4,0(t4)        // 0001000100010001 0001000100010001
    li      t5,16           // # of scanlines

        //  Load 1st 128bits 'S' from (un)aligned memory

    ldl     t0, 7(a1)       // S  {63...0}
    ldl     t2, 15(a1)      //     {127...64}
    ldr     t0, 0(a1)
    ldr     t2, 8(a1)

    add     a1,a1,a2        //  pbSrc += iStride

    pcpyld  t0,t2,t0        // S  {127...0}
    pextub  t2,zero,t0      // S[ _F _E _D _C _B _A _9 _8 ]
    pextlb  t0,zero,t0      //  [ _7 _6 _5 _4 _3 _2 _1 _0 ]

        //  _________________
$V1:    //  Load 128bits from (un)aligned memory

    ldl     t1, 7(a1)       // R  {63...0}
    ldl     t3, 15(a1)      // R  {127...64}
    ldr     t1, 0(a1)
    ldr     t3, 8(a1)
    add     a1,a1,a2        //  pbSrc += iStride

    pcpyld  t1,t3,t1        // R  {127...0}
    pextub  t3,zero,t1      // R[ _F _E _D _C _B _A _9 _8 ]
    pextlb  t1,zero,t1      //  [ _7 _6 _5 _4 _3 _2 _1 _0 ]

        //  _________________
        //  Now average t2:t0 and t3:t1 : (A + B + 1)>>1

    paddh   t2,t2,t3        // S+R
    paddh   t0,t0,t1        // S+R
    paddh   t2,t2,t4        // [...]+1
    paddh   t0,t0,t4        // [...]+1
    psrlh   t2,t2,1         // ()>>1
    psrlh   t0,t0,1         // ()>>1

    ppacb   t0,t2,t0        // Interleave 16 halfwords to 16 bytes

        //  _________________
        //  Save result to (unaligned) memory

    pcpyud  t2,t0,t0        // D  {63...0  127...64}

    sdr     t0, 0(a0)
    sdl     t0, 7(a0)       // pwD {63...0}
    sdr     t2, 8(a0)
    sdl     t2, 15(a0)      //    {127...64}

    add     a0,a0,a2        // pDst += iStride

    por     t0,t1,t1        // R = S Copy register(s)
    por     t2,t3,t3        // R = S

    addi    t5,t5,-1
    bne     zero,t5,$V1     // Loop for 16 scanlines
    BDELAY                  // nop = Branch Delay Slot

    j       ra              // ret
    BDELAY                  // nop = Branch Delay Slot

                        
