
//
//  KariType.h - Master Definitions
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 2 Jun 98 - James Leiterman
//
//  Last Modification
//
//  stab=4
//



#ifndef __KariType_H
#define __KariType_H

#ifdef __cplusplus
extern "C" {
#endif


#define CC_VMP_MIPS
#define CC_VMP_PS2                                      // MAC
//#define CC_VMP_MAC                                        // MAC



#ifndef byte
typedef unsigned char  byte;
#endif

typedef unsigned char  uint8;
typedef unsigned short uint16;
typedef unsigned int   uint32;
typedef unsigned long  uint64;

#ifndef uint
typedef unsigned int   uint;
#endif

typedef signed char    int8;
typedef signed short   int16;
typedef signed int     int32;
typedef signed long    int64;

#ifdef HOMEBREW
typedef unsigned char       u8;
typedef unsigned short      u16;
typedef unsigned int        u32;
typedef unsigned long int   u64;

#ifdef _EE
typedef unsigned int        u128 __attribute__(( mode(TI) ));
#endif

typedef signed char         s8;
typedef signed short        s16;
typedef signed int      s32;
typedef signed long int     s64;

#ifdef _EE
typedef int         s128 __attribute__(( mode(TI) ));
#endif
#endif



union vmp_qword {
    byte    b[8];
    uint16  h[4];
    uint32  w[2];

    char    c[8];
    int16   s[4];
    int32   i[2];

    float   f[2];
    double  d[1];
};



union vmp_oword {
    byte    b[16];
    uint16  h[8];
    uint32  w[4];

    char    c[16];
    int16   s[8];
    int32   i[4];

    float   f[4];
    double  d[2];
};


#ifndef bool
typedef enum
{
    vmp_false = 0,
    vmp_true,
} vmp_bool;

    //  Don't use vmp_false, vmp_true... instead use { bool, false, true } !!!
    //  Makes bool true : false compatible with C and C++

#define false   vmp_false
#define true    vmp_true
#define bool    vmp_bool
#endif



#define ALIGN2( len ) (( len + 1 ) & ~1 )     // round up to 16bits
#define ALIGN4( len ) (( len + 3 ) & ~3 )     // round up to 32bits
#define ALIGN8( len ) (( len + 7 ) & ~7 )     // round up to 64bits
#define ALIGN16( len ) (( len + 15 ) & ~15 )  // round up to 128bits

#define ALIGN2048( len ) (( len + 2047 ) & ~2047 )  // round up to 1 CD Sector
#define ALIGN2336( len ) (( len + 2335 ) % 2336 )   // round up to 1 CD Sector



#define USE_ASSERT


#ifdef USE_ASSERT
#define ASSERT( arg1 ) assert( (arg1) )
#define ASSERT_PTR( arg1 )   assert(NULL!=(arg1))
#define ASSERT_PTR4( arg1 )   assert((NULL!=(arg1)) && (0==(((uint)(arg1))&3)))
#define ASSERT_PTR16( arg1 ) assert((NULL!=(arg1)) && (0==(((uint)(arg1))&15)))
#define ASSERT_FNEG( arg1 ) assert(0.0f<=(arg1));
#define ASSERT_ZERO( arg1 ) assert(0.0f!=(arg1));

#else
#define ASSERT( arg1 )
#define ASSERT_PTR( arg1 )
#define ASSERT_PTR4( arg1 )
#define ASSERT_PTR16( arg1 )
#define ASSERT_FNEG( arg1 )
#define ASSERT_ZERO( arg1 )
#endif



#ifdef __cplusplus
}
#endif

#endif //__KariType_H



